
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "polis.h"
#include "cache.h"

#ifdef UNIX 
int is_SW = 1;
int csimulflag = 1;
int tracegenflag = 0;
#endif


/* Global variables used for both os_sim and PTOLEMY co-simulation. */

#ifdef UNIX
FILE * fptraces;
#endif /* UNIX */
task_list_t task[NUM_MAX_TASKS];
array_t *history_cache;
st_table *layout_table;
st_table *confl_inter_table;

int cache_size;
int associativity;
int line_size;
int history_length;
float threshold;
int csimulstyle;        /* cache simulation style */
int i_fetchsize;
int miss_penalty;

int mem_ptr = 0;        /* current position in main memory (task loading) */
float min_hit_rate;
int path_found_in_history;
int path_found_in_layout_table;
int history_pos;        /* insert position in history */
int n_instr_fetch;      /* total number of instruction fetches */ 
float n_instr_hit;      /* total number of instruction hits */
int n_instr_fetch_path; /* number of instruction fetches in the current path */
float n_instr_hit_path; /* number of instruction hits in the current path */
int n_data_fetch;       /* total number of data fetches */ 
int n_data_hit;         /* total number of data hits */
int tmp_i;
float tmp_f;


#include "full_cache.c"
#include "fast_cache.c"

/* Functions used to instrument the z_*.c file */

int
cache_path_lookup( int task_id, int path_id,
                   c_sg_node_t *sgraph, header_path_t **curr_header_path,
                   path_cache_layout_t **curr_layout,
                   int is_SW, int csimulstyle, int csimulflag )
{
   layout_path_pair_t *new_layout_path_pair;
   int path_found = FALSE;

   path_found_in_history = FALSE;
   path_found_in_layout_table = FALSE;
   n_instr_fetch_path = 0;
   n_instr_hit_path = 0.0;

   if ( is_SW && csimulflag ) {
      switch ( csimulstyle ) {
         case FULLFAST:
            if ((new_layout_path_pair = lookup_layout_table( task_id, path_id))!=NULL ) {
#ifdef DEBUG
               printf("Found in layout_table\n");
#endif
               path_found = TRUE;
               path_found_in_layout_table = TRUE;
               *curr_header_path = get_path_layout_path_pair( new_layout_path_pair );
               *curr_layout = get_layout_layout_path_pair( new_layout_path_pair );
               _AddInstrHit( update_intra_conflicts( *curr_header_path, sgraph ));
            }
            break;
         case FAST:
            if ((new_layout_path_pair = lookup_history( task_id,path_id ))!=NULL
 ) {
#ifdef DEBUG
               printf("Found in history_cache\n");
#endif
               path_found = TRUE;
               path_found_in_history = TRUE;
               free_path( *curr_header_path );
               *curr_header_path = get_path_layout_path_pair( new_layout_path_pair );
               *curr_layout = get_layout_layout_path_pair( new_layout_path_pair );
               _AddInstrHit( update_intra_conflicts( *curr_header_path, sgraph ));
            }
            else if ((new_layout_path_pair = lookup_layout_table( task_id, path_id))!=NULL ) {
#ifdef DEBUG
               printf("Found in layout_table\n");
#endif
               path_found = TRUE;
               path_found_in_layout_table = TRUE;
               free_path( *curr_header_path );
               *curr_header_path = get_path_layout_path_pair( new_layout_path_pair );
               *curr_layout = get_layout_layout_path_pair( new_layout_path_pair );
               _AddInstrHit( update_intra_conflicts( *curr_header_path, sgraph ));
            }
            break;
      }
   }
   return path_found;
}

void
cache( header_path_t *curr_header_path, c_sg_node_t *sgraph, int node,
       int path_found, path_cache_layout_t **curr_layout,
       int is_SW, int csimulstyle, int csimulflag, int tracegenflag,
       FILE *fptraces )
{
   c_sg_node_t *curr_node;
   float curr_node_hit_rate;
   int curr_node_data_hit;
   int curr_node_n_i_fetches;
   path_cache_layout_t *temp_layout = NULL;

   if ( is_SW && csimulflag ) {
      switch( csimulstyle ) {
         case FULLFAST:
         case FAST:
            if ( ! path_found ) {
#ifdef DEBUG
               printf("CACHE: hit_rate(%d)= %.2f\n",
                       node, curr_node_hit_rate);
#endif
               curr_node = get_node_sgraph( sgraph, node );
               curr_node_hit_rate = get_hit_rate( curr_node );
               curr_node_n_i_fetches = get_n_i_fetches( curr_node );
#ifdef DEBUG
               printf("CACHE: s_mem= %d(0x%x), e_mem= %d(0x%x), s_cache= %d(0x%x), e_cache= %d(0x%x)\n",
                      get_s_mem( curr_node ), get_s_mem( curr_node ),
                      get_e_mem( curr_node ), get_e_mem( curr_node ),
                      get_s_cache( curr_node ), get_s_cache( curr_node ),
                      get_e_cache( curr_node ), get_e_cache( curr_node ));
#endif
               /* update layout */
               temp_layout = create_path_cache_layout( curr_node );
               *curr_layout = update_cache_layout( *curr_layout, temp_layout );
#ifdef DEBUG
               print_path_cache_layout( *curr_layout );
#endif
               /* add_node_in_path( curr_header_path, node ); */ /* update path */
               set_path_vect_pos( curr_header_path, node ); 
               compute_intra_conflicts( curr_header_path, sgraph, node );
               _AddInstrHit((float) (curr_node_n_i_fetches * curr_node_hit_rate));
               set_hit_rate( curr_node, 1.0 ); /* the node goes in cache */
               add_n_i_fetches_path( curr_header_path, curr_node_n_i_fetches );
            }
            break;
         case FULL:
            curr_node = get_node_sgraph( sgraph, node );
            curr_node_n_i_fetches = get_n_i_fetches( curr_node );
            curr_node_hit_rate = put_in_cache( cache_tags, get_i_fetch_list( curr_node ) );
            _AddInstrHit( curr_node_hit_rate );
            _AddInstrFetch( curr_node_n_i_fetches );

            /* Begin Mauro */
	    /*curr_node_n_i_fetches = get_n_d_fetches( curr_node ); */
            /* curr_node_data_hit = put_in_cache( data_cache, get_d_fetch_list( curr_node ) );
             */
	    _AddDataHit( curr_node_data_hit );
	    _AddDataFetch( curr_node_n_i_fetches );
            /* End Mauro */
            break;
      }
      if ( tracegenflag ) { 
         print_traces( sgraph, node, is_SW, csimulflag, tracegenflag, fptraces );
      }
   }
}

void
cache_layout( int task_id, int path_id, path_cache_layout_t *curr_layout,
              header_path_t *curr_header_path, c_sg_node_t *sgraph,
              int path_found, int is_SW, int csimulstyle, int csimulflag )
{
   layout_path_pair_t *new_layout_path_pair;

   if ( is_SW && csimulflag ) {
      switch ( csimulstyle ) {
         case FULLFAST:
            if ( ! path_found ) {
               new_layout_path_pair = create_layout_path_pair( task_id,path_id,curr_layout,curr_header_path );
               insert_in_layout_table( new_layout_path_pair );
            }
            break;
         case FAST:
            new_layout_path_pair = create_layout_path_pair( task_id,path_id,curr_layout,curr_header_path );
            update_inter_conflicts( new_layout_path_pair, _GetHistoryPos );
            _AddInstrFetch( get_n_i_fetches_path( curr_header_path ));
            if ( ! path_found ) {
#ifdef DEBUG
               printf("CACHE_LAYOUT: task_id= %d, path_id= %d\n", task_id, path_id);
#endif
               record_path_hit_rate( curr_header_path, sgraph );
               _SetHistoryPos( insert_in_history( new_layout_path_pair, _GetHistoryPos ));
            }
#ifdef DEBUG
            print_history();
            print_path_hit_rate( curr_header_path, sgraph );
#endif
         break;
      }
#ifdef STAT
      printf( "\n" );
      printf( "=========== \n" );
      printf( "Statistics: \n" );
      printf( "=========== \n" );
      printf( "n_instr_fetch= %d\n", _GetInstrFetch );
      printf( "n_instr_hit= %.0f\n", _GetInstrHit );
      printf( "instr_hit_rate= %.2f\n", (float) (_GetInstrHit/_GetInstrFetch*100) );
      printf( "n_instr_miss= %.0f\n", (float) (_GetInstrFetch-_GetInstrHit) );
      printf( "instr_miss_rate= %.2f\n", (float) (_GetInstrFetch-_GetInstrHit)/_GetInstrFetch*100 );
      /* DATA STATISTICS */
/*
      printf( "n_data_fetch= %d\n", _GetDataFetch );
      printf( "n_data_hit= %d\n", _GetDataHit );
      printf( "data_hit_rate= %.2f\n", (float)(_GetDataHit/_GetDataFetch*100) );
      printf( "n_data_miss= %d\n", _GetDataFetch-_GetDataHit );
      printf( "data_miss_rate= %.2f\n", (float)(_GetDataFetch-_GetDataHit)/_GetDataFetch*100 );
*/
#endif
   }
}

void 
cache_plot_hit_rate( int task_id, c_sg_node_t *sgraph, 
                     header_path_t *curr_header_path, int max_label,
                     int is_SW, int csimulflag, int csimulstyle )
{
   int curr_node_id;
   c_sg_node_t *curr_node;

   if ( is_SW && csimulflag ) {
      if ( csimulstyle==FAST ) {
         curr_node_id = 0;
#ifdef DEBUG
         printf("CACHE_PLOT_HIT_RATE:\n");
         printf("\ttask_id= %d\n", task_id);
#endif
         while( curr_node_id <= max_label ) {
            curr_node = get_node_sgraph( sgraph, curr_node_id );
#ifdef DEBUG
            printf("\thit_rate(%d)= %.2f\tpath_vect(%d)= %d\n",
                   curr_node_id, get_hit_rate( curr_node ), curr_node_id,
                   get_path_vect_pos(curr_header_path, curr_node_id));
#endif
#ifdef PTOLEMY
            /* xyplot->addPoint( curr_node_id, get_hit_rate(curr_node), 1 );*/
            /* addPoint( curr_node_id, get_hit_rate(curr_node), 1 );*/
#endif
            curr_node_id++;
         }
      }
   }
}

void 
compute_abstract_cache_state( int task_id, int path_id,
                              c_sg_node_t *sgraph, 
                              header_path_t **header_path,
                              path_cache_layout_t **curr_layout,
                              int is_SW, int csimulstyle, int csimulflag,
                              int tracegenflag, FILE *fptraces )
{
   int node;
   int path_found; /* in history or in layout table */
   path_t *curr_path;

   if ( is_SW && csimulflag ) {

      fprintf( stderr, "compute_abstract_cache_state: task_id= %d, sgraph= %d\n", task_id, sgraph );
  
      path_found = cache_path_lookup( task_id, path_id, sgraph,
               header_path, curr_layout, is_SW, csimulstyle, csimulflag );

      if ( !path_found || tracegenflag ) {
         /* compute_cache_info( ); */
         curr_path = get_head_path( *header_path );
         while( curr_path ) {
            node = get_node_path( curr_path );
            cache( *header_path, sgraph, node, path_found, curr_layout, 
                   is_SW, csimulstyle, csimulflag, tracegenflag, fptraces );
            curr_path = get_next_node_path( curr_path );
         }
      }
      cache_layout( task_id, path_id, *curr_layout, *header_path,
                    sgraph, path_found, is_SW, csimulstyle, csimulflag );
   }
}

void
print_traces( c_sg_node_t *sgraph, int node, int is_SW, 
              int csimulflag, int tracegenflag, FILE *fptraces ) {
   c_sg_node_t *curr_node;
   i_fetch_list_t *instr_fetch_list;

   if ( is_SW && csimulflag && tracegenflag ) {
      curr_node = get_node_sgraph( sgraph, node );
      instr_fetch_list = get_i_fetch_list( curr_node );
      while ( instr_fetch_list != NULL ) {
         PrintTraces( fptraces, get_i_fetch_addr( instr_fetch_list ));
         instr_fetch_list = get_next_i_fetch( instr_fetch_list );
      }
   }
}

/***********************/

void reset_cache() {

   history_pos = 0;
   n_instr_fetch = 0;
   n_instr_hit = 0.0;
}

void
init_cache( int cache_size, int line_size, int associativity,
            int history_length, float threshold, int csimulstyle,
            int i_fetch_size, int miss_penalty ) {
   int i;

   /* Reset Cache */
   history_pos = 0;
   n_instr_fetch = 0;
   n_instr_hit = 0.0;
   mem_ptr = 0;

   /* Set cache parameters (they can change at a new simulation run) */
   set_cache_size( cache_size );
   set_line_size( line_size );
   set_associativity( associativity );
   set_history_length( history_length );
   set_threshold( threshold );
   set_csimulstyle( csimulstyle );
   set_i_fetchsize( i_fetch_size );
   set_miss_penalty( miss_penalty );

#ifdef DEBUG
   printf("init_cache:\n");
   printf("CACHESIZE= %d\n", CACHESIZE);
   printf("LINESIZE= %d\n", LINESIZE);
   printf("ASSOCIATIVITY= %d\n", ASSOCIATIVITY);
   printf("NUM_CACHE_LINES= %d\n", NUM_CACHE_LINES);
   printf("NUM_CACHE_SETS= %d\n", NUM_CACHE_SETS);
   printf("HISTORY_LENGTH= %d\n", HISTORY_LENGTH);
   printf("THRESHOLD= %.2f\n", THRESHOLD);
   printf("MISS_PENALTY= %d\n", MISS_PENALTY);
#endif /*DEBUG*/

   min_hit_rate = 1.0 - (float) (2/LINESIZE);

   switch( csimulstyle ) {
      case FAST:
         free_history_cache();
         alloc_history_cache();
         /* free_layout_table(); */
         alloc_layout_table();
         /* free_confl_inter_table(); */
         alloc_confl_inter_table();
         break;
      case FULL:
	/* begin Mauro */
         cache_tags = create_cache(NUM_CACHE_LINES);
	 data_cache = create_cache(NUM_CACHE_LINES);
        /* removed by Mauro */ 
/*         cache_tags = (int *)malloc(sizeof(int) * (NUM_CACHE_LINES));

         if ( cache_tags != NULL ) {
            for (i=0; i < NUM_CACHE_LINES; i++) {
               set_cache_tag( i, -1 );
            }
            i = time(NULL);
            srand(i);	
            }
         else {
            fprintf(stderr, "Error in init_cache\n");
            exit(1);
         } */
/* end Mauro */
         break;
      default:
         break;
   }
}

/******************/
/* TASK OPERATORS */
/******************/

task_list_t * 
get_task( int task_id ) {
   return &task[task_id];
}

int 
get_task_num_c_sg_nodes( task_list_t *task ) {
   if( task == NULL ) {
      fprintf(stderr, "get_task_num_c_sg_nodes: task NULL\n");
      exit(1);
   }
   return task->num_c_sg_nodes;
}

void 
set_task_num_c_sg_nodes( task_list_t *task, int value ) {
   if( task == NULL ) {
      fprintf(stderr, "set_task_num_c_sg_nodes: task NULL\n");
      exit(1);
   }
   task->num_c_sg_nodes = value;
}

int 
get_task_start_addr( task_list_t *task ) {
   if( task == NULL ) {
      fprintf(stderr, "get_task_start_address: task NULL\n");
      exit(1);
   }
   return task->start_addr;
}

void 
set_task_start_addr( task_list_t *task, int value ) {
   if( task == NULL ) {
      fprintf(stderr, "set_task_start_address: task NULL\n");
      exit(1);
   }
   task->start_addr = value;
}

c_sg_node_t * 
get_task_sgraph( task_list_t *task ) {
   if( task == NULL ) {
      fprintf(stderr, "get_task_sgraph: task NULL\n");
      exit(1);
   }
   return task->sgraph;
}

void 
set_task_sgraph( task_list_t *task, c_sg_node_t *node ) {
   if( task == NULL ) {
      fprintf(stderr, "set_task_sgraph: task NULL\n");
      exit(1);
   }
   task->sgraph = node;
}

void
update_addresses( c_sg_node_t *sgraph, int max_label, int offset ) {
   int node_id, i;
   c_sg_node_t *curr_node;
   int s_addr, e_addr;
   int curr_addr = -1;
   char *fuzzy_expr;

   if ( offset == 0 ) offset = mem_ptr;
#ifdef DEBUG
   printf("update_addresses: \n");
   printf("max_label= %d\n", max_label);
#endif

   /* Set addresses in main memory */

   for ( node_id=1; node_id<max_label; node_id++ ) {
      curr_node = get_node_sgraph( sgraph, node_id );
      fuzzy_expr = get_fuzzy_expr( curr_node );
      s_addr = curr_addr + 1;
      e_addr = curr_addr + fuzzy_size( fuzzy_expr );
      curr_addr = e_addr;
      set_addr_mem( offset, sgraph, node_id, s_addr, e_addr );
#ifdef DEBUG
      printf( "update_addresses: fuzzy_expr(%d) %s\n", node_id, fuzzy_expr );
      printf( "update_addresses: set_addr_mem(%d)= %d, %d\n", node_id, s_addr, e_addr );
#endif

       /* Clean I-fetch-list allocated in previuous running */
       free_i_fetch_list( sgraph, node_id );
       
      /* Fuzzy Instruction Fetches computation */
      for ( i=s_addr; i<=e_addr; i+=get_i_fetchsize() ) {
          add_i_fetch( sgraph, node_id, i + offset );
      }
   }

   /* update current position in main memory (task loading) */
   mem_ptr += (curr_addr + 1);
#ifdef DEBUG
   printf( "update_addresses: mem_ptr= %d\n", mem_ptr );
#endif

   /* Set addresses in cache */

#ifdef DEBUG
   printf("node\ts_mem\t\te_mem\t\ts_cache\t\te_cache\t\thit_rate\n");
#endif

   node_id = 0;

   while( node_id < max_label ) {
      curr_node = get_node_sgraph( sgraph, node_id );
      set_addr_cache( curr_node );
      set_hit_rate( curr_node, 0 );

#ifdef DEBUG
      printf("%d\t%d(0x%x)\t%d(0x%x)\t%d\t\t%d\t\t%.2f\n",
              node_id,
              get_s_mem( curr_node ), get_s_mem( curr_node ),
              get_e_mem( curr_node ), get_e_mem( curr_node ),
              get_s_cache( curr_node ), get_e_cache( curr_node ),
              get_hit_rate( curr_node ));
#endif

      node_id++;
   }
}

/*****************************/
/*   SGRAPH CONSTRUCTORS     */
/*****************************/

/* 
 * Read the sgraph info file "sgraphinfofile"
 * and the cache info file "cacheinfofile" and build the sgraph.
 */

void
build_sgraph( int task_id, int custom_layout, int offset_address,
              char *basenamefile ) {

   c_sg_node_t *sgraph = NULL;
   task_list_t *task = NULL;
   char buff[BUFSIZE];
   int numnodes, node_id, child;
   int fetch, s_addr, e_addr;
   int i;
   int dummy;
   int offset;
   int size;
   FILE *ftemp;
   char sgraphinfofile[256];
   i_fetch_list_t *list;
   c_sg_node_t *node;
   char fuzzy_expr[1024];
   int curr_addr = -1;

   if ( custom_layout ) {
      offset = offset_address;
   }
   else {
      offset = mem_ptr;
   }

   sprintf( sgraphinfofile, "%s.sgraph", basenamefile );

#ifdef DEBUG
   printf("build_sgraph: Building s-graph for task %d with sgraphinfofile %s\n",
           task_id, sgraphinfofile);
   printf("build_sgraph: mem_ptr= %d\n", mem_ptr );
#endif

/*************************/
/* Read sgraph info file */
/*************************/

   ftemp = fopen( sgraphinfofile, "r" );
   if ( ftemp ) {
      while ( fscanf( ftemp, "%s\n", &buff[0] ) != EOF) {
         switch( buff[0] ) {
            case ' ':
            case '#':
            case '\n':
               break;
            case 'D' :  /* DIM */
               fscanf( ftemp,"%d\n", &numnodes );
#ifdef DEBUG
               printf("build_sgraph: D: buff= %s\n", buff);
               printf("build_sgraph: numnodes= %d\n", numnodes);
#endif
               alloc_sgraph( task_id, numnodes );
               task = get_task( task_id );
               sgraph = get_task_sgraph( task );
               set_task_start_addr( task, offset );
               set_task_num_c_sg_nodes( task, numnodes );
               break;
            case 'N' :  /* NODE */
               fscanf( ftemp,"%d\n", &node_id );
#ifdef DEBUG
               printf("build_sgraph: N: buff= %s\n", buff);
               printf("build_sgraph: node_id= %d\n", node_id);
#endif
               break;
            case 'C' :  /* CHILD */
#ifdef DEBUG
               printf("build_sgraph: C: buff= %s\n", buff);
#endif
               while (fscanf( ftemp,"%d", &child ) != '\0' ) {
#ifdef DEBUG
                  printf("add_child= %d, %d\n", node_id,child);
#endif
                  add_child(sgraph,node_id,child);
               }
               fscanf( ftemp, "\n");
               break;
            case 'F' :  /* FUZZY */
#ifdef DEBUG
               printf( "build_sgraph: FUZZY\n" );
               fscanf( ftemp,"%s\n", fuzzy_expr );
               printf( "build_sgraph: F: fuzzy_expr= %s\n", fuzzy_expr );
               printf( "build_sgraph: F: fuzzy_size= %d\n", fuzzy_size(fuzzy_expr));
#endif
               node = get_node_sgraph( sgraph, node_id );
               set_fuzzy_expr( node, fuzzy_expr );
               break;
/* Mauro */
	    case 'V' : /* DATA OR VARIABLE ADDR */
	       dummy = fscanf( ftemp,"%d", &fetch );
	       while ((dummy != '\0' ) && (dummy != EOF)) {
#ifdef DEBUG
                  printf("build_sgraph: VAR_ADDR= %d\n", fetch);
#endif
                  /* fetch = lookup_varname(var_name); */
                  add_d_fetch( sgraph, node_id, fetch /* + offset */ );
                  dummy = fscanf( ftemp,"%d", &fetch );
               }
/* fine Mauro */

            default:
               exit;
         }
      }
   }
   else {
      fprintf(stderr,"Can't open file %s\n", sgraphinfofile);
      exit(1);
   }

   fclose( ftemp );

   update_addresses( sgraph, numnodes, offset );
}

c_sg_node_t * 
create_sgraph( int task_id, int num_c_sg_nodes ) {
   c_sg_node_t *temp, *temp1;
   int i;
 
   temp = (c_sg_node_t *)malloc(sizeof(c_sg_node_t)*num_c_sg_nodes);
   if ( temp ) {
      for ( i=0; i<num_c_sg_nodes; i++ ) {
         temp1 = get_node_sgraph( temp, i );
         set_task_id( temp1, task_id );
         set_s_mem( temp1, 0 );
         set_e_mem( temp1, 0 );
         set_s_cache( temp1, 0 );
         set_e_cache( temp1, 0 );
         set_hit_rate( temp1, 0 );
         set_n_i_fetches( temp1, 0 );
         set_i_fetch_list( temp1, NULL );
         set_last_i_fetch( temp1, NULL );
         set_child_list( temp1, NULL );
         set_last_child( temp1, NULL ); 
      }
   }
   else {
      fprintf(stderr, "Error in create_c_sg_node_struct\n");
      exit(1);
   }
   return temp;
}

void 
alloc_sgraph( int task_id, int num_c_sg_nodes ) {
   c_sg_node_t *temp;
 
   temp = create_sgraph( task_id, num_c_sg_nodes );

   set_task_num_c_sg_nodes(get_task(task_id),num_c_sg_nodes);
   set_task_sgraph(get_task(task_id), temp);
}

/******************************/
/*   NODE_SGRAPH OPERATORS    */
/******************************/

c_sg_node_t * 
get_node_sgraph( c_sg_node_t *sgraph, int node ) {
   if( sgraph ) {
      return &sgraph[node];
   }
   else {
      fprintf(stderr, "get_node_sgraph: sgraph NULL\n");
      exit(1);
   }
}

int 
get_task_id( c_sg_node_t *c_sg_node ) {
   return c_sg_node->task_id;
}

void 
set_task_id( c_sg_node_t *c_sg_node, int value ) {
   c_sg_node->task_id = value;
}
 
char * 
get_fuzzy_expr( const c_sg_node_t *c_sg_node ) {
   if( c_sg_node == NULL ) {
      fprintf(stderr, "get_fuzzy_expr: c_sg_node NULL\n");
      exit(1);
   }
   return c_sg_node->fuzzy_expr;
}

void 
set_fuzzy_expr( c_sg_node_t *c_sg_node, char *value ) {
   if( c_sg_node == NULL ) {
      fprintf(stderr, "set_fuzzy_expr: c_sg_node NULL\n");
      exit(1);
   }
   c_sg_node->fuzzy_expr = strdup( value );
}
int 
get_s_mem( const c_sg_node_t *c_sg_node ) {
   if( c_sg_node == NULL ) {
      fprintf(stderr, "get_s_mem: c_sg_node NULL\n");
      exit(1);
   }
   return c_sg_node->s_mem;
}

void
set_s_mem( c_sg_node_t *c_sg_node, int value ) {
   if( c_sg_node == NULL ) {
      fprintf(stderr, "set_s_mem: c_sg_node NULL\n");
      exit(1);
   }
   c_sg_node->s_mem = value;
}

int
get_e_mem( c_sg_node_t *c_sg_node ) {
   if( c_sg_node == NULL ) {
      fprintf(stderr, "get_e_mem: c_sg_node NULL\n");
      exit(1);
   }
   return c_sg_node->e_mem;
}

void 
set_e_mem( c_sg_node_t *c_sg_node, int value ) {
   if( c_sg_node == NULL ) {
      fprintf(stderr, "set_e_mem: c_sg_node NULL\n");
      exit(1);
   }
   c_sg_node->e_mem = value;
}

int 
get_s_cache( const c_sg_node_t *c_sg_node ) {
   if( c_sg_node == NULL ) {
      fprintf(stderr, "get_s_cache: c_sg_node NULL\n");
      exit(1);
   }
   return c_sg_node->s_cache;
}

void 
set_s_cache( c_sg_node_t *c_sg_node, int value ) {
   if( c_sg_node == NULL ) {
      fprintf(stderr, "set_s_cache: c_sg_node NULL\n");
      exit(1);
   }
   c_sg_node->s_cache = value;
}

int 
get_e_cache( const c_sg_node_t *c_sg_node ) {
   if( c_sg_node == NULL ) {
      fprintf(stderr, "get_e_cache: c_sg_node NULL\n");
      exit(1);
   }
   return c_sg_node->e_cache;
}

void 
set_e_cache( c_sg_node_t *c_sg_node, int value ) {
   if( c_sg_node == NULL ) {
      fprintf(stderr, "set_e_cache: c_sg_node NULL\n");
      exit(1);
   }
   c_sg_node->e_cache = value;
}

float 
get_hit_rate( c_sg_node_t *c_sg_node ) {
   if( c_sg_node == NULL ) {
      fprintf(stderr, "get_hit_rate: c_sg_node NULL\n");
      exit(1);
   }
   return c_sg_node->hit_rate;
}

void 
set_hit_rate( c_sg_node_t *c_sg_node, float value ) {
   if( c_sg_node == NULL ) {
      fprintf(stderr, "set_hit_rate: c_sg_node NULL\n");
      exit(1);
   }
   c_sg_node->hit_rate = value;
}

int 
get_n_i_fetches( c_sg_node_t *c_sg_node ) {
   return c_sg_node->n_i_fetches;
}

void 
set_n_i_fetches( c_sg_node_t *c_sg_node, int value ) {
   c_sg_node->n_i_fetches = value;
}

/**************************/
/*   I_FETCH SELECTORS    */
/**************************/

i_fetch_list_t * 
get_i_fetch_list( c_sg_node_t *c_sg_node ) {
   return c_sg_node->i_fetch_list;
}

void 
set_i_fetch_list(c_sg_node_t *c_sg_node, i_fetch_list_t *list) {
   c_sg_node->i_fetch_list = list;
}

i_fetch_list_t *  
get_last_i_fetch( c_sg_node_t *c_sg_node ) {
   return c_sg_node->last_i_fetch;
}

void 
set_last_i_fetch( c_sg_node_t * c_sg_node, i_fetch_list_t *list) {
   c_sg_node->last_i_fetch = list;
}
i_fetch_list_t * 
get_d_fetch_list( c_sg_node_t *c_sg_node ) {
   return c_sg_node->d_fetch_list;
}

int 
get_n_d_fetches( c_sg_node_t *c_sg_node ) {
   return c_sg_node->n_d_fetches;
}

void 
set_n_d_fetches( c_sg_node_t *c_sg_node, int value ) {
   c_sg_node->n_d_fetches = value;
}

void 
set_d_fetch_list(c_sg_node_t *c_sg_node, i_fetch_list_t *list) {
   c_sg_node->d_fetch_list = list;
}

i_fetch_list_t *  
get_last_d_fetch( c_sg_node_t *c_sg_node ) {
   return c_sg_node->last_d_fetch;
}

void 
set_last_d_fetch( c_sg_node_t * c_sg_node, i_fetch_list_t *list) {
   c_sg_node->last_d_fetch = list;
}

/**************************/
/*   I_FETCH OPERATORS    */
/**************************/

int 
get_i_fetch_addr( i_fetch_list_t *list ) {
   return list->addr;
} 

void 
set_i_fetch_addr( i_fetch_list_t *list, int addr ) {
   list->addr = addr;
}

void 
set_next_i_fetch( i_fetch_list_t *list, i_fetch_list_t *list2) {
   list->next = list2;
}
 
i_fetch_list_t * 
get_next_i_fetch(i_fetch_list_t *list) {
   return list->next;
}

/*****************************/
/*   I_FETCH CONSTRUCTORS    */
/*****************************/

i_fetch_list_t * 
alloc_i_fetch_list( void ) {
   i_fetch_list_t *temp;

   temp = (i_fetch_list_t *)malloc(sizeof(i_fetch_list_t));
   if (temp==NULL) {
      printf("Error in alloc_i_fetch_list\n");
      exit(1);
   }
   set_i_fetch_addr( temp, 0 );
   set_next_i_fetch( temp, NULL );
   return temp;
}
 
void 
del_i_fetch( i_fetch_list_t *list ) {
   if( list != NULL )
      free( list );
}

void 
free_i_fetch_list( c_sg_node_t *sgraph, int node ) {
   i_fetch_list_t *list, *list2;
   c_sg_node_t *tmp;

   tmp = get_node_sgraph( sgraph, node );
   list = get_i_fetch_list( tmp );
   while ( list != NULL ) {
      list2 = get_next_i_fetch(list);     
      del_i_fetch( list );
      list = list2;
   }
   set_i_fetch_list( tmp, NULL );
   set_last_i_fetch( tmp, NULL );
   set_n_i_fetches( tmp, 0 );
}

i_fetch_list_t * 
create_i_fetch( int addr ) {
   i_fetch_list_t *temp;
 
   temp = alloc_i_fetch_list();
   set_i_fetch_addr( temp, addr );
   return temp;
}

void 
add_i_fetch( c_sg_node_t *sgraph, int node, int addr ) {
   i_fetch_list_t *list, *list2;
   c_sg_node_t *tmp;

   tmp = get_node_sgraph( sgraph, node );
   list = get_last_i_fetch( tmp );

   if ( list==NULL ) {
      list2 = create_i_fetch( addr );
      set_next_i_fetch( list2, NULL );
      set_i_fetch_list( tmp, list2 );
      set_last_i_fetch( tmp, list2 );
   }
   else {
      list2 = create_i_fetch( addr );
      set_next_i_fetch( list, list2 );
      set_next_i_fetch( list2, NULL );
      set_last_i_fetch( tmp, list2 );
   }
   set_n_i_fetches( tmp, get_n_i_fetches( tmp )+1 );
}

void 
add_d_fetch( c_sg_node_t *sgraph, int node, int addr ) {
   i_fetch_list_t *list, *list2;
   c_sg_node_t *tmp;

   tmp = get_node_sgraph( sgraph, node );
   list = get_last_d_fetch( tmp );

   if ( list==NULL ) {
      list2 = create_i_fetch( addr );
      set_next_i_fetch( list2, NULL );
      set_d_fetch_list( tmp, list2 );
      set_last_d_fetch( tmp, list2 );
   }
   else {
      list2 = create_i_fetch( addr );
      set_next_i_fetch( list, list2 );
      set_next_i_fetch( list2, NULL );
      set_last_d_fetch( tmp, list2 );
   }
}
/***********************/
/*   CHILD SELECTORS   */
/***********************/

void 
set_child_list( c_sg_node_t * c_sg_node, child_list_t *list) {
   c_sg_node->child_list = list;
}
 
child_list_t * 
get_child_list( c_sg_node_t * c_sg_node ) {
   return c_sg_node->child_list;
}

child_list_t *  
get_last_child( c_sg_node_t *c_sg_node ) {
   return c_sg_node->last_child;
}

void 
set_last_child( c_sg_node_t *c_sg_node, child_list_t *list) {
   c_sg_node->last_child = list;
}

void 
set_child_id( child_list_t *list, int child_id ) {
   list->child_id = child_id;
}
 
int 
get_child_id( child_list_t *list ) {
   return list->child_id;
}

void 
set_next_child( child_list_t *list, child_list_t *list2 ) {
   list->child_list = list2;
}
 
child_list_t * 
get_next_child( child_list_t *list ) {
   return list->child_list;
}

float 
get_child_confl_instr( child_list_t *list ) {
   return list->confl_instr;
}

void 
set_child_confl_instr( child_list_t *list, float value ) {
   list->confl_instr = value;
}

/**************************/
/*   CHILD CONSTRUCTORS   */
/**************************/

child_list_t * 
alloc_child_list( void ) {
   child_list_t *temp;

   temp = (child_list_t *)malloc(sizeof(child_list_t));
   if (temp==NULL) {
      printf("Error in alloc_child_list\n");
      exit(1);
   }
   set_child_id( temp, 0 );
   set_next_child( temp, NULL );
   set_child_confl_instr( temp, 0 );
   return temp;
}
 
void 
free_child_list( child_list_t *list ) {
   if( list != NULL) 
      free( list );
}

child_list_t * 
create_child( int child_id ) {
   child_list_t *temp;
 
   temp = alloc_child_list();
   set_child_id( temp, child_id );
   return temp;
}

void 
add_child( c_sg_node_t *sgraph, int node1, int node2 ) {
   child_list_t *list, *list2;
   c_sg_node_t *temp;

   temp = get_node_sgraph( sgraph, node1 );
   list = get_last_child( temp );

   if ( list==NULL ) {
      list2 = create_child( node2 );
      set_next_child( list2, NULL );
      set_child_list( temp, list2 );
      set_last_child( temp, list2 );
   }
   else {
      list2 = create_child( node2 );
      set_next_child( list, list2 );
      set_next_child( list2, NULL );
      set_last_child( temp, list2 );
   }
}

/**********************/
/*   CHILD OPERATORS  */
/**********************/

void 
set_cache_confl_instr( c_sg_node_t *sgraph, int start_node, int end_node, 
                       float value ) {
   child_list_t *temp;
   c_sg_node_t * temp1;
   int found;

   found = FALSE; 
   temp1 = get_node_sgraph( sgraph, start_node );
   temp = get_child_list( temp1 );

   while ((temp!=NULL) && (found==FALSE)){
      if (get_child_id(temp)==end_node) {
         set_child_confl_instr( temp, value);
         found = TRUE;
      }
      else {
         temp=get_next_child(temp);
      }
   }
}

/******************************************************* 
 *  HACK functions to allow setting of constants also  *
 *  during a  PTOLEMY co-simulation session.           * 
 *******************************************************
 */

int 
get_cache_size( ) {
   return cache_size;
}

void 
set_cache_size( int value ) {
   cache_size = value;
}

int 
get_line_size( ) {
   return line_size;
}

void 
set_line_size( int value ) {
   line_size = value;
}

int 
get_associativity( ) {
   return associativity;
}

void 
set_associativity( int value ) {
   associativity = value;
}

int 
get_history_length( ) {
   return history_length;
}

void 
set_history_length( int value ) {
   history_length = value;
}

float 
get_threshold( ) {
   return threshold;
}

void 
set_threshold( float value ) {
   threshold = value;
}

int
get_csimulstyle( ) {
   return csimulstyle;
}

void
set_csimulstyle( int value ) {
   csimulstyle = value;
}

int
get_i_fetchsize( ) {
   return i_fetchsize;
}

void
set_i_fetchsize( int value ) {
   i_fetchsize = value;
}

int 
get_miss_penalty( ) {
   return miss_penalty;
}

void 
set_miss_penalty( int value ) {
   miss_penalty = value;
}

/*****************************************************/
/* Miscellaneous of functions for debugging purposes */
/*****************************************************/

void 
testnode( c_sg_node_t *sgraph, int i ) {
/* Only for test purposes.
 * It writes out information for a specified node. 
 */ 
   child_list_t *temp;
   c_sg_node_t *temp1;

   temp1 = get_node_sgraph( sgraph, i );
   temp = get_child_list ( temp1 );

   while( temp!=NULL ) {
      printf("node= %d child= %d  s_mem= %xh(%d), e_mem= %xh(%d), s_cache= %d, e_cache= %d\n", 
              i, get_child_id( temp ),
              get_s_mem( temp1 ), get_s_mem( temp1 ),
              get_e_mem( temp1 ), get_e_mem( temp1 ),
              get_s_cache( temp1 ), get_e_cache( temp1 ));
      temp = get_next_child( temp );
   }
}

void 
test( int task_id ) {
/* Only for test purposes.
 * It writes out all the information contained in the s-graph
 * structure for the specified task_id.
 */
   int i;
   int num_c_sg_nodes;
   c_sg_node_t *sgraph;

   sgraph = get_task_sgraph( get_task( task_id ));
   num_c_sg_nodes = get_task_num_c_sg_nodes(get_task(task_id));

   printf("TASK n. %d\n", task_id);

   for (i=1; i<num_c_sg_nodes; i++) {
      testnode(sgraph,i);
   }

   printf("node= 0 s_mem= %d(0x%x), e_mem= %d(0x%x), s_cache= %d(0x%x), e_cache= %d(0x%x)\n",
           get_s_mem( sgraph ), get_s_mem( sgraph ),
           get_e_mem( sgraph ), get_e_mem( sgraph ),
           get_s_cache( sgraph ), get_s_cache( sgraph ),
           get_e_cache( sgraph ), get_e_cache( sgraph ));
} 

void 
PrintTraces( FILE *fptraces, int addr  ) 
{
   fprintf( fptraces, "2 %x\n", addr );
   fflush( fptraces );
}

void 
clean_hit_rate( c_sg_node_t *sgraph, int max_label ) {
   int curr_node_id;
   c_sg_node_t *curr_node;

   curr_node_id = 0;
   while( curr_node_id <= max_label ) {
      curr_node = get_node_sgraph( sgraph, curr_node_id );
      set_hit_rate( curr_node, 0.0 );
      curr_node_id++;
   }
}
