/*
 * Analyze command line arguments.
 */

#include "cache.h"

static int atoiKMG( char * );

int
getcmdargs( int argc, char **argv)
/* 
 * returns: OK if command line arguments are valid
 *	    ERR otherwise
 */
{
register int i;
register int nextnum;
int returnflag;
char nextchar;
char nextstring[5];
int tmp_cache_size = DEFAULT_CACHESIZE;
int tmp_line_size = DEFAULT_LINESIZE;
int tmp_associativity = DEFAULT_ASSOCIATIVITY;
int tmp_history_length = DEFAULT_HISTORY_LENGTH;
float tmp_threshold = DEFAULT_THRESHOLD;
int tmp_csimulstyle = DEFAULT_CSIMULSTYLE;
int tmp_i_fetchsize = DEFAULT_I_FETCHSIZE;
int tmp_miss_penalty = DEFAULT_MISS_PENALTY;

   setbuf( stdout, NULL );

   returnflag = OK;


   /* 
    **   Echo command line 
    */
   printf("\n");
   printf("CMDLINE: ");
   for (i=0; i<argc; i++) {
      printf("%s%c", argv[i], (i<argc-1) ? ' ' : '\n');
   }

   /* 
    **   Get arguments 
    */
    for( i=1; i<argc; i++ ) {
       if (argv[i][0]!='-') {
          printf("Argument %d (%s) must be preceded by '-'.\n",i,argv[i]);
          returnflag = ERR;
       }
       else {
          switch( argv[i][1] )  {   /* switch on second char. */
                                    /* allow: "-m100" or "-m 100" */

             case 'b' :  /* BLOCK SIZE */
                nextnum = atoiKMG(argv[i][2]=='\0' ? argv[++i] : &argv[i][2]);
                if (nextnum >= 1 )
                   tmp_line_size = nextnum;  
                else {
                   returnflag = ERR;
                 }
                 break;

             case 'u' : /* UNIFIED INSTRUCTION/DATA CACHE SIZE */
                nextnum = atoiKMG(argv[i][2]=='\0' ? argv[++i] : &argv[i][2]);
                if (nextnum >= 0 )  /* size 0 implies no unified cache. */
                   tmp_cache_size = nextnum;  
                else {
                   returnflag = ERR;
                }
                break;

             case 'i' : /* INSTRUCTION CACHE SIZE */
                nextnum = atoiKMG(argv[i][2]=='\0' ? argv[++i] : &argv[i][2]);
                if (nextnum >= 0 )
                   tmp_cache_size = nextnum;  
                else {
                   returnflag = ERR;
                }
                break;

             case 'd' : /* DATA CACHE SIZE */
                fprintf(stderr, "Error: data cache is not available yet..... stay tuned!!!!");
                break;

             case 'a' :  /* ASSOCIATIVITY */
                nextnum = atoiKMG(argv[i][2]=='\0' ? argv[++i] : &argv[i][2]);
                if (nextnum >= 1 )
                   tmp_associativity = nextnum;
                else {
                   returnflag = ERR;
                }
                break;

             case 'm' :  /* MISS_PENALTY */
                nextnum = atoiKMG(argv[i][2]=='\0' ? argv[++i] : &argv[i][2]);
                if (nextnum >= 1 )
                   tmp_miss_penalty = nextnum;
                else {
                   returnflag = ERR;
                }
                break;

             case 'r' :  /* REPLACEMENT POLICY */
                nextchar = (argv[i][2]=='\0' ? argv[++i][0] : argv[i][2]);
                if (nextchar==LRU_POLICY || nextchar==FIFO_POLICY || nextchar==RANDOM_POLICY) {
                   if (nextchar != RANDOM_POLICY) {
                      fprintf(stderr, 
                          "Only random replacement is currently supported\n");
                   }
                }
                else {
                   returnflag = ERR;
                }
                break;

             case 's' :   /* STYLE */
                nextstring[0] = argv[i][2];
                nextstring[1] = argv[i][3];
                nextstring[2] = argv[i][4];
                nextstring[3] = argv[i][5];
                nextstring[4] = '\0';
                fprintf(stderr,"CACHE_SIMUL_STYLE= %s\n",nextstring);
                if (!strcmp( nextstring, "FAST" )) {
                   tmp_csimulstyle = FAST;
                }
                else if (!strcmp( nextstring, "FULL" )) {
                   tmp_csimulstyle = FULL;
                }
                else {
                   returnflag = ERR;
                }
                break;

             case 'g' :   /* TRACE generation */
                tracegenflag = 1;
                break;

             case 'h' :  /* HISTORY LENGHT */
                nextnum = atoi(argv[i][2]=='\0' ? argv[++i] : &argv[i][2]);
                if (nextnum >= 1 )
                   tmp_history_length = nextnum;
                else {
                   returnflag = ERR;
                }
                break;

             case 't' :  /* THRESHOLD */
                nextnum = atof(argv[i][2]=='\0' ? argv[++i] : &argv[i][2]);
                if (nextnum >= 1 )
                   tmp_threshold = nextnum;
                else {
                   returnflag = ERR;
                }
                break;
             }

          }
       }

   if ( returnflag == OK ) {

      printf("tracegenflag= %d\n", tracegenflag);

      init_cache( tmp_cache_size, tmp_line_size, tmp_associativity,
                  tmp_history_length, tmp_threshold,
                  tmp_csimulstyle, tmp_i_fetchsize, tmp_miss_penalty );
   }

   return returnflag;

}

static int 
atoiKMG( char *str )
/*
**  Return integer equal to value of ascii string that is sequence
**  of digits possibly followed by one of the following suffixes:
**
**	K, k	1024
**	M, m	1024**2 = 1048576
**	G, g	1024**3 = 1073741824
**
**  E.g., "64K" evaluated to integer 65536.
**  ILLEGALNUM is returned for illegal suffixes or on overflow.
*/
{
char str_buffer[256];
int length;
char suffix;
int root, multiplier, value;

   strcpy( str_buffer, str );

   length = strlen(str_buffer);
   suffix = str_buffer[length-1];

   switch ( suffix ) {

      case 'K' :
      case 'k' :
         multiplier = 1024;
         str_buffer[length-1] = '\0';
         break;

      case 'M' :
      case 'm' :
         multiplier = 1048576;    /* 1024**2 */
         str_buffer[length-1] = '\0';
         break;

      case 'G' :
      case 'g' :
         multiplier = 1073741824;	/* 1024**3 */
         str_buffer[length-1] = '\0';
         break;

      case '0' :
      case '1' :
      case '2' :
      case '3' :
      case '4' :
      case '5' :
      case '6' :
      case '7' :
      case '8' :
      case '9' :
         multiplier = 1;
         break;

      default :
         multiplier = ILLEGALNUM;
         break;
   }

   if ( multiplier!=ILLEGALNUM ) {
      root = atoi( str_buffer );
      value = root*multiplier;

      if ( root != ( value/multiplier )) {
         value = ILLEGALNUM;
      }
   }
   else {
      value = ILLEGALNUM;
   }

   return(value);
}
