/************************/
/*   HEADER OPERATORS   */
/************************/
 
header_path_t *
alloc_header_path( int numnodes ) {
   header_path_t *tmp;
   path_vect_t *tmp_path_vect;
 
   tmp = (header_path_t *)malloc(sizeof(header_path_t));
   tmp_path_vect = alloc_path_vect( numnodes );
   if ( tmp && tmp_path_vect ) {
      set_head_path( tmp, NULL );
      set_tail_path( tmp, NULL );
      set_path_vect( tmp, tmp_path_vect );
      set_n_i_fetches_path( tmp, 0 );
   }
   else {
      fprintf(stderr, "Error in alloc_header_path\n");
      exit(1);
   }
   return tmp;
}
 
path_t *
get_head_path( header_path_t *header ) {
   if ( header)
      return header->head;
   else {
      fprintf(stderr, "Error in get_head_path\n");
      exit(1);
   }
}
 
void
set_head_path( header_path_t *header, path_t *path ) {
   header->head = path;
}
 
path_t *
get_tail_path( header_path_t *header ) {
   if ( header )
      return header->tail;
   else
      return NULL;
}
 
void
set_tail_path( header_path_t *header, path_t *path ) {
   header->tail = path;
}
 
int
get_n_i_fetches_path( header_path_t *header ) {
   return header->n_i_fetches;
}
 
void
set_n_i_fetches_path( header_path_t *header, int n_i_fetches ) {
   header->n_i_fetches = n_i_fetches;
}

void
add_n_i_fetches_path( header_path_t *header, int n_i_fetches ) {
   header->n_i_fetches += n_i_fetches;
}

path_t *
alloc_node_path( void ) {
   path_t *tmp;
 
   tmp = (path_t *)malloc(sizeof(path_t));
   if ( tmp ) {
      set_node_path( tmp, 0 );
      set_next_node_path( tmp, NULL );
      set_hit_rate_after_path_exec( tmp, 0.0 );
      set_correct_percen_path( tmp, 1.0 );
   }
   else {
      fprintf(stderr, "Error in alloc_node_path\n");
      exit(1);
   }
   return tmp;
}

void free_node_path( path_t *list ) {
   if( list )
      free( list );
}
 
void
free_path( header_path_t *header ) {
   path_t *tmp1, *tmp2;

   if ( header ) { 
      fprintf( stderr, "get_head_path(free_path)\n" );
      tmp1 = get_head_path( header );
      while ( tmp1 ) {
         tmp2 = get_next_node_path( tmp1 );
         free( tmp1 );
         tmp1 = tmp2;
      }
      set_head_path( header, NULL );
      set_tail_path( header, NULL );
   }
}
 
path_t *
create_node_path( int node ) {
   path_t *tmp;
 
   tmp = alloc_node_path();
   set_node_path( tmp, node );
   set_next_node_path( tmp,NULL );
   return tmp;
}
 
void
add_node_in_path( header_path_t *header_path, int node,
               int is_SW, int csimulflag ) {
   path_t *list, *list2;

   if ( is_SW && csimulflag ) { 
      if ( header_path ) {
         list = get_tail_path( header_path );
         list2 = create_node_path( node );
         if ( list==NULL ) {
            set_head_path( header_path, list2 );
            set_tail_path( header_path, list2 );
         }
         else {
            set_next_node_path( list, list2 );
            set_tail_path( header_path, list2 );
         }
      }
   }
}

/***********************************/
/*  BASIC CACHE INFO COMPUTATIONS  */
/***********************************/
 
void
set_addr_cache( c_sg_node_t *curr_node ) {
   int k;
 
   k = ( get_e_mem(curr_node)>>1)-( get_s_mem(curr_node)>>1 );

   if ( k >= NUM_CACHE_SETS ) {
      set_s_cache( curr_node, 0 );
      set_e_cache( curr_node, NUM_CACHE_SETS -1 );
   }
   else {
      set_s_cache( curr_node, (get_s_mem(curr_node)>>1) % NUM_CACHE_SETS );
      set_e_cache( curr_node, (get_e_mem(curr_node)>>1) % NUM_CACHE_SETS );
   }
 
}
 
void
set_addr_mem( int task_start_addr, c_sg_node_t *sgraph, 
              int node, int s_mem, int e_mem ) {
   c_sg_node_t *tmp;
   int start_mem, end_mem;
 
   tmp  = get_node_sgraph( sgraph, node );
   start_mem = s_mem + task_start_addr;
   end_mem   = e_mem + task_start_addr;
 
   set_s_mem ( tmp, start_mem );
   set_e_mem ( tmp, end_mem );
}
 
float
calc_confl( int s11, int s12, int s21, int s22 ) {
/* compute the number of cache lines in which
 * segments (s11,s12) and (s21,s22) overlap
 */
   float confl;
   int l1, l2;
 
   confl = 0;
 
   l1 = s12-s11;
   l2 = s22-s21;
 
   if ((s12>=s21)&&(s21>=s11)&&(s22>=s12)) {      /*  |---------|    */
      confl = (float) (s12-s21)+1;                /*     |---------| */
      }
   else if ((s21<=s11)&&(s22>=s11)&&(s22<=s12)) { /*     |---------| */
      confl = (float) (s22-s11)+1;                /*  |---------|    */
      }
   else if ((s21<=s11)&&(s22>=s12)) {             /*    |-----|      */
      confl = (float) l1+1;                       /* |----------|    */
      }
   else if ((s21>=s11)&&(s22<=s12)) {             /* |----------|    */
      confl = (float) l2+1;                       /*    |-----|      */
      }
 
   return confl;
}
 
float
calc_confl_instr( c_sg_node_t *sgraph, int node1, int node2 ) {
/* compute the percentage of instruction references
 * that node 2 removes to node 1
 */
   float confl;
   int s11, s12, s21, s22, l1;
   c_sg_node_t *tmp;
 
   tmp = get_node_sgraph( sgraph, node1 );
   s11 = get_s_cache( tmp );
   s12 = get_e_cache( tmp );
 
   tmp = get_node_sgraph( sgraph, node2 );
   s21 = get_s_cache( tmp );
   s22 = get_e_cache( tmp );
 
 
   confl = 0;
 
   if (s12<s11) {
      l1 = s12 - s11 + NUM_CACHE_LINES;
      if (s22<s21) {
         confl  = calc_confl(  0, s12, 0, s22 );
         confl += calc_confl(  0, s12, s21, NUM_CACHE_LINES-1 );
         confl += calc_confl( s11, NUM_CACHE_LINES-1, 0, s22 );
         confl += calc_confl( s11, NUM_CACHE_LINES-1, s21, NUM_CACHE_LINES-1 );
      }
      else {
         confl  = calc_confl(  0, s12, s21, s22 );
         confl += calc_confl( s11, NUM_CACHE_LINES-1, s21, s22 );
      }
   }
   else {
      l1 = s12 - s11;
      if (s22<s21) {
         confl  = calc_confl( s11, s12, 0, s22 );
         confl += calc_confl( s11, s12, s21, NUM_CACHE_LINES-1 );
      }
      else {
         confl  = calc_confl( s11, s12, s21, s22 );
      }
   }
 
   confl = confl/(l1+1); /* approximation....
                          * the correct number of instruction
                          * fetches should be considered
                          */

   confl = confl/ASSOCIATIVITY;

   return confl;
}
 
void
set_confl_instr( c_sg_node_t *sgraph, int node1, int node2 ) {
   float confl_glob;
 
   confl_glob = calc_confl_instr( sgraph, node1, node2 );
   printf("confl( %d, %d ) = %.2f\n", node1, node2, confl_glob);
   set_cache_confl_instr( sgraph, node1, node2, confl_glob );
}
 
/*****************************/
/*  HISTORY CACHE OPERATORS  */
/*****************************/
 
void
alloc_history_cache( ) {
   if ( history_cache==NULL ) {
      printf("alloc_history_cache: HISTORY_LENGTH=%d\n", HISTORY_LENGTH);
      history_cache = array_alloc( layout_path_pair_t *, HISTORY_LENGTH );
   }
}
 
void 
free_history_cache( ) {
   int i; 
   layout_path_pair_t *tmp;

   if ( history_cache ) {
      for ( i=0; i < array_n( history_cache ); i++ ) {
         tmp = (layout_path_pair_t *)get_history_elem( i );
         if ( tmp ) {
            free_layout_path_pair( tmp );
         }
      }
      array_free( history_cache );
      history_cache = NULL;
   }
}
 
layout_path_pair_t *
get_history_elem( int history_pos ) {
   if ( history_pos > HISTORY_LENGTH ) {
      fprintf(stderr, "Error in get_history_elem\n");
      exit(1);
   }
   if ( history_pos < array_n( history_cache )) { 
      return (array_fetch( layout_path_pair_t *, history_cache, history_pos ));
   }
   else {
      return NULL;
   }
}
 
void
set_history_elem( layout_path_pair_t *elem, int history_pos ) {
   if ( history_pos > HISTORY_LENGTH ) {
      fprintf(stderr, "Error in set_history_elem\n");
      exit(1);
   }
   array_insert( layout_path_pair_t *, history_cache, history_pos, elem );
}
 
int
insert_in_history( layout_path_pair_t *layout_path, int history_pos ) {
   layout_path_pair_t *elem_history;
 
   elem_history = get_history_elem( history_pos );
   if ( elem_history ) {
      insert_in_layout_table( elem_history );
   }
   set_history_elem( layout_path, history_pos );
 
#ifdef DEBUG
   elem_history = get_history_elem( history_pos );
   printf("insert_in_history: task_id= %d, path_id= %d  (history_pos= %d)\n",
      get_task_id_layout_path_pair( elem_history ),
      get_path_id_layout_path_pair( elem_history ), history_pos);
#endif
 
   history_pos = (history_pos+1)%(HISTORY_LENGTH);
 
   return history_pos;
}
 
layout_path_pair_t *
lookup_history( int task_id, int path_id ) {
   layout_path_pair_t *tmp;
   int curr_pos=0;
 
#ifdef DEBUG
   fprintf( stderr, "Begin lookup_history( task_id= %d, path_id= %d )\n",
                    task_id, path_id );
#endif
   tmp = get_history_elem( curr_pos );
   while(( tmp ) && ( curr_pos<HISTORY_LENGTH )) {
      if ((get_task_id_layout_path_pair(tmp)==task_id) &&
          (get_path_id_layout_path_pair(tmp)==path_id)) {
#ifdef DEBUG
         fprintf(stderr, "lookup_history: Found(%d,%d) in position %d\n",
              get_task_id_layout_path_pair(tmp),
              get_path_id_layout_path_pair(tmp),
              curr_pos);
         fprintf( stderr, "End lookup_history\n" );
#endif
         return tmp;   /* found path in history */
      }
      curr_pos++;
      tmp = get_history_elem( curr_pos );
   }
#ifdef DEBUG
   fprintf( stderr, "End lookup_history\n" );
#endif
   return NULL;
}
 
void
print_history( void ) {
   layout_path_pair_t *tmp;
   int curr_pos=0;
 
   curr_pos=0;
   tmp = get_history_elem( curr_pos );
   printf("print_history: HISTORY_LENGTH= %d\n", HISTORY_LENGTH);
   while(( tmp ) && (curr_pos<HISTORY_LENGTH)) {
#ifdef DEBUG
      printf("print_history: curr_pos= %d, task_id= %d, path_id= %d\n",
        curr_pos, get_task_id_layout_path_pair(tmp),
        get_path_id_layout_path_pair(tmp));
#endif
      curr_pos += 1;
      tmp = get_history_elem( curr_pos );
  }
}
 
/***************************/
/*   NODE_PATH SELECTORS   */
/***************************/
 
int
get_node_path( path_t *list ) {
   if ( list )
      return list->node;
   else {
      fprintf(stderr, "Error in get_node_path\n");
      exit(1);
   }
}
 
void
set_node_path( path_t *list, int value ) {
   list->node = value;
}
 
path_t *
get_next_node_path( path_t *list ) {
   if ( list )
      return list->next;
   else
      return NULL;
}
 
void
set_next_node_path( path_t *list, path_t *list2 ) {
   list->next = list2;
}
 
/*********************************/
/* Hit Rate after path execution */
/*********************************/
 
float
get_hit_rate_after_path_exec( path_t *list ) {
   return list->hit_rate_after_path_exec;
}
 
void
set_hit_rate_after_path_exec( path_t *list, float value ) {
   list->hit_rate_after_path_exec = value;
}
 
/***************************/
/* Correct percentage path */
/***************************/
 
float
get_correct_percen_path( path_t *list ) {
   return list->correct_percen;
}
 
void
set_correct_percen_path( path_t *list, float value ) {
   list->correct_percen = value;
}
 
void
add_correct_percen_path( path_t *list, float value ) {
   list->correct_percen = (list->correct_percen)*(1-value);
}
 
/***************************************************************
 *   path_vect COSTRUCTORS                                     *
 *                                                             *
 *   The path_vect structure is used to store the information  *
 *   about the presence or not of a node in the path.          *
 *   This information is then used during the computation      *
 *   of inter-conflicts between paths of the same task.        *
 ***************************************************************/
 
path_vect_t *
alloc_path_vect( int num_nodes ) {
   path_vect_t *tmp;
   int i;
 
   tmp = (int *)malloc(sizeof(path_vect_t)*num_nodes);
   if ( tmp ) {
      for ( i=0; i<num_nodes; i++ ) {
         tmp[i] = 0;
      }
   }
   else {
      fprintf(stderr, "Error in alloc_path_vect\n");
      exit(1);
   }
   return tmp;
}
 
void
free_path_vect( path_vect_t *path_vect ) {
   if( path_vect ) free( path_vect );
}
 
path_vect_t
get_path_vect_pos( header_path_t *header, int node ) {
   path_vect_t *tmp;
 
   if ( header ) {
      tmp = get_path_vect( header );
      if ( tmp )
         return tmp[node];
   }
   fprintf(stderr, "Error in get_path_vect_pos\n");
   exit(1);
}
 
void
set_path_vect_pos( header_path_t *header, int node ) {
   path_vect_t *tmp;
 
   if ( header ) {
      tmp = get_path_vect( header );
      tmp[node] = TRUE;
   }
   else {
      fprintf(stderr, "Error in set_path_vect_pos\n");
      exit(1);
   }
}
 
path_vect_t *
get_path_vect( header_path_t *header ) {
   if ( header )
      return header->path_vect;
   else
      return NULL;
}
 
void
set_path_vect( header_path_t *header, path_vect_t *path_vect ) {
   header->path_vect = path_vect;
}
 
/*******************************/
/* path_cache_layout SELECTORS */
/*******************************/
 
int
get_start_cache_segment(path_cache_layout_t *layout) {
   return layout->start;
}
 
void
set_start_cache_segment(path_cache_layout_t *layout, int value) {
   layout->start = value;
}
 
int
get_end_cache_segment(path_cache_layout_t *layout) {
   return layout->end;
}
 
void
set_end_cache_segment(path_cache_layout_t *layout, int value) {
   layout->end = value;
}
 
path_cache_layout_t *
get_next_cache_segment(path_cache_layout_t *layout) {
   return layout->next_segment;
}
 
void
set_next_cache_segment(path_cache_layout_t *layout, path_cache_layout_t *next) {
   layout->next_segment = next;
}
 
/**********************************/
/* path_cache_layout CONSTRUCTORS */
/**********************************/
 
path_cache_layout_t *
create_path_cache_layout( c_sg_node_t *node ) {
   path_cache_layout_t *tmp, *tmp2;
   int start, end;
 
   start = get_s_cache( node );
   end = get_e_cache( node );
 
   if (start<=end) {
      tmp = (path_cache_layout_t *)malloc(sizeof(path_cache_layout_t));
      if (tmp==NULL) {
         fprintf(stderr, "Error in create_path_cache_layout\n");
         exit(1);
      }
      else {
         set_start_cache_segment( tmp, start );
         set_end_cache_segment( tmp, end );
         set_next_cache_segment( tmp, NULL );
      }
   }
   else {
      tmp = (path_cache_layout_t *)malloc(sizeof(path_cache_layout_t));
      tmp2 = (path_cache_layout_t *)malloc(sizeof(path_cache_layout_t));
      if ((tmp==NULL)||(tmp2==NULL)) {
         fprintf(stderr, "Error in create_path_cache_layout\n");
         exit(1);
      }
      else {
         set_start_cache_segment( tmp, 0 );
         set_end_cache_segment( tmp,end  );
         set_next_cache_segment( tmp, tmp2 );
         set_start_cache_segment( tmp2, start );
         set_end_cache_segment( tmp2, NUM_CACHE_LINES-1  );
         set_next_cache_segment( tmp2, NULL );
      }
   }
   return tmp;
}
 
void
free_cache_segment( path_cache_layout_t *segment ) {
   if( segment )
      free( segment );
}
 
void
free_cache_layout( path_cache_layout_t *layout ) {
   path_cache_layout_t *tmp1, *tmp2;
 
   if ( layout ) {
      tmp1 = layout;
      tmp2 = get_next_cache_segment( tmp1 );
 
      while ( tmp2 ) {
         free_cache_segment( tmp1 );
         tmp1 = tmp2;
         tmp2 = get_next_cache_segment( tmp1 );
      }
      free_cache_segment( tmp1 );
   }
}
 
/*******************************/
/* path_cache_layout OPERATORS */
/*******************************/
 
path_cache_layout_t *
update_cache_segment( path_cache_layout_t *seg1, path_cache_layout_t *seg2 ) {
   int s11, s12, s21, s22;
 
   if ((seg1 == NULL) || (seg2 == NULL)) return NULL;
 
   s11 = get_start_cache_segment( seg1 );  /* start pos. in cache of seg 1 */
   s12 = get_end_cache_segment( seg1 )+1;  /* end   pos. in cache of seg 1 */
   s21 = get_start_cache_segment( seg2 );  /* start pos. in cache of seg 2 */
   s22 = get_end_cache_segment( seg2 )+1;  /* end   pos. in cache of seg 2 */
 
   if (s12<s11)
      s12 += NUM_CACHE_LINES;
 
   if (s22<s21)
      s22 += NUM_CACHE_LINES;
 
   if (s11<=s21) {
      if (s12>=s22) {                        /*  |-------------------| */
                                             /*            |------|    */
         set_start_cache_segment( seg2, get_start_cache_segment( seg1 ));
         set_end_cache_segment( seg2, get_end_cache_segment( seg1 ));
         return seg2;
      }
      else if (s21<=s12) {
         set_start_cache_segment( seg2, get_start_cache_segment( seg1 ));
                                             /*    |------|            */
         return seg2;                        /*       |--------------| */
      }
   }
   else /* if (s11>=s21) { */
      if (s12<=s22) {                        /*            |------|    */
        return seg2;                         /*  |-------------------| */
      }
      else if (s11<=s22) {
         set_end_cache_segment( seg2, get_end_cache_segment( seg1 ));
         return seg2;                        /*       |--------------| */
                                             /*    |------|            */
      }
 
   return NULL;
}
 
path_cache_layout_t *
add_node_to_layout( path_cache_layout_t *layout, c_sg_node_t *sgraph, int node_id ) {
   path_cache_layout_t *tmp;
   path_cache_layout_t *new_layout;
   c_sg_node_t *node;
 
   node = get_node_sgraph( sgraph, node_id );
   tmp = create_path_cache_layout( node );
   new_layout = update_cache_layout( layout, tmp );
/*
   printf("add_node_to_layout\n");
   print_path_cache_layout( layout );
*/
   return new_layout;
}
 
path_cache_layout_t *
update_cache_layout( path_cache_layout_t *layout, path_cache_layout_t *new_seg ) {
   int s11, s21, s22/*, s12*/;
   path_cache_layout_t dummy, *tmp1, *tmp2, *found;
 
   if ( (tmp1 = get_next_cache_segment( new_seg )) != NULL) {
       update_cache_layout( layout, tmp1 );
       set_next_cache_segment( new_seg, NULL);
   }
 
   set_next_cache_segment( &dummy, layout);
 
   s21 = get_start_cache_segment( new_seg );
   s22 = get_end_cache_segment( new_seg );
 
   tmp1 = &dummy;
   tmp2 = layout;
 
   while( tmp2 ) {
      s11 = get_start_cache_segment( tmp2 );
      /* s12 = get_end_cache_segment( tmp2 ); */
 
      found = update_cache_segment( tmp2, new_seg );
 
      if (( !found ) && ( s22 < s11 )) {
         set_next_cache_segment( new_seg, tmp2);
         set_next_cache_segment( tmp1, new_seg);
         return get_next_cache_segment( &dummy );
      }
      else if ( found ) {
         while ( found ) {
            tmp2 = get_next_cache_segment( tmp2 );
            found = update_cache_segment( tmp2, new_seg );
         }
 
         set_next_cache_segment( tmp1, new_seg );
         set_next_cache_segment( new_seg, tmp2 );
         return get_next_cache_segment( &dummy );
      }
 
      tmp1 = tmp2;
      tmp2 = get_next_cache_segment( tmp2 );
   }
 
   set_next_cache_segment( tmp1, new_seg);
 
   return get_next_cache_segment( &dummy );
}
 
void
print_path_cache_layout( path_cache_layout_t *layout ) {
   path_cache_layout_t *tmp;
 
   tmp = layout;
   printf ("- layout -   ");
   while( tmp ) {
      printf( "(%d,%d) ",
              get_start_cache_segment( tmp ), get_end_cache_segment( tmp ));
      tmp = get_next_cache_segment( tmp );
   };
   printf ("\n");
}
 
/*********************************/
/* layout_path_pair_t  SELECTORS */
/*********************************/
 
int
get_task_id_layout_path_pair(layout_path_pair_t *element) {
   return element->task_id;
}
 
void
set_task_id_layout_path_pair(layout_path_pair_t *element, int id) {
   element->task_id = id;
}
 
int
get_path_id_layout_path_pair(layout_path_pair_t *element) {
   return element->path_id;
}
 
void
set_path_id_layout_path_pair(layout_path_pair_t *element, int id ) {
   element->path_id = id;
}
 
path_cache_layout_t *
get_layout_layout_path_pair(layout_path_pair_t *element) {
   return element->layout;
}
 
void
set_layout_layout_path_pair(layout_path_pair_t *element,
                            path_cache_layout_t *layout) {
   element->layout = layout;
}
 
header_path_t *
get_path_layout_path_pair(layout_path_pair_t *element) {
   return element->path;
}
 
void
set_path_layout_path_pair(layout_path_pair_t *element, header_path_t *path_head) {
   element->path = path_head;
}
 
/***********************************/
/* layout_path_pair_t CONSTRUCTORS */
/***********************************/
 
layout_path_pair_t *
create_layout_path_pair( int task_id, int path_id,
                         path_cache_layout_t *layout, header_path_t *path ) {
   layout_path_pair_t *tmp;

   tmp = (layout_path_pair_t *)malloc(sizeof(layout_path_pair_t));
   if (tmp==NULL) {
      fprintf(stderr, "Error in create_layout_path_pair\n");
      exit(1);
   }
   else {
      set_task_id_layout_path_pair( tmp, task_id );
      set_path_id_layout_path_pair( tmp, path_id );
      set_layout_layout_path_pair( tmp, layout );
      set_path_layout_path_pair( tmp, path );
   }
   return tmp;
}
 
void
free_layout_path_pair( layout_path_pair_t *segment ) {
   if( segment ) {
      free_cache_layout( get_layout_layout_path_pair( segment ) );
      free_path( get_path_layout_path_pair( segment ) );
      free( segment );
   }
}
 
/**************************/
/* layout_table OPERATORS */
/**************************/
 
void
alloc_layout_table( ) {
   printf("alloc_layout_table\n");
   layout_table = st_init_table( st_ptrcmp, st_ptrhash );
}
 
void 
free_layout_table( ) {
   layout_path_pair_t *tmp;
   layout_path_pair_t *tmp1;
   st_generator *gen; 

   st_foreach_item( layout_table, gen, (char**)&tmp, (char**)&tmp1 ) {
      free_layout_path_pair( tmp1 );
   }
   st_free_table( layout_table );
}

int
make_key ( int task, int path ) {
    return ((task&KEY_TASK_MASK)<<KEY_PATH_BITS )|(path&KEY_PATH_MASK);
}
 
layout_path_pair_t *
lookup_layout_table( int task, int path ) {
    int key;
    char skey[80]; 
    layout_path_pair_t *tmp;

    key = make_key ( task, path );
    sprintf( skey, "%d", key ); 
#ifdef DEBUG
    fprintf( stderr, "Begin lookup_layout_table\n" );
#endif
/* 
    tmp = get_layout_from_table( key );
*/
    if ( st_lookup( layout_table, (char *) skey, (char **) &tmp )) {
       if ((get_task_id_layout_path_pair(tmp) == task) &&
           (get_path_id_layout_path_pair(tmp) == path))
          return tmp;  /* found path in layout table */
    }
    return NULL;
}
 
void
insert_in_layout_table( layout_path_pair_t *layout ) {
    int task, path, key;
    char skey[80]; 
 
    task = get_task_id_layout_path_pair( layout );
    path = get_path_id_layout_path_pair( layout );
 
    key = make_key (task, path );
    sprintf( skey, "%d", key ); 
    /* set_layout_in_table( key, layout ); */
    st_insert( layout_table, (char *) skey, (char *) layout ); 
}

/*************************************************************/
/* INTRA-CONFLICTS analysis:                                 */
/* conflicts between basic blocks belonging to the same task */
/*************************************************************/
 
void
compute_intra_conflicts( header_path_t *header_path, 
                         c_sg_node_t *sgraph, int curr_node_id ) {
   int node1, node2;
   path_t *curr_path;
   path_t *next_path;
   c_sg_node_t *curr_node;
   float confl;
   float new_hit_rate;
 
   curr_path = get_tail_path( header_path );

   node1 = curr_node_id;
   fprintf( stderr, "get_head_path(2)\n" );
   next_path = get_head_path( header_path );
   node2 = get_node_path( next_path );
 
   while( node2 != curr_node_id ) {

      /* forward conflict (what node2 removes of node1) */ 

      confl = calc_confl_instr( sgraph, node1, node2 );
      curr_node = get_node_sgraph( sgraph, node1);
      if ( confl >= THRESHOLD ) {
#ifdef DEBUG
         printf("compute_intra_conflicts: ");
         printf("confl(%d,%d)= %.2f (TAKEN into account)\n",
                 node1, node2, confl);
#endif
         new_hit_rate = get_hit_rate( curr_node )*( 1-confl );
         set_hit_rate( curr_node, new_hit_rate );
         add_correct_percen_path( curr_path, confl );
#ifdef DEBUG
         printf("compute_intra_conflicts: add_correct_percen_path(%d,%.2f)\n",
               get_node_path(curr_path), get_correct_percen_path( curr_path ));
#endif
      }
#ifdef DEBUG
      else {
         printf("compute_intra_conflicts: ");
         printf("confl(%d,%d)= %.2f (NOT taken into account)\n",
                 node1, node2, confl);
      }
#endif

#ifdef DEBUG 
      if ( confl ) {
         printf("compute_intra_conflicts: ");
         printf("node1= %d, node2= %d, remove %.2f, new_hit_rate(%d)= %.2f\n\n",
                 node1, node2, confl, node1, new_hit_rate);
         printf("get_hit_rate(%d)= %.2f\n", node1, get_hit_rate( curr_node ));
      }
#endif
 
      /* backword conflict (what node1 removes of node2) */ 

      confl = calc_confl_instr( sgraph, node2, node1 );
      curr_node = get_node_sgraph( sgraph, node2 );
      if ( confl >= THRESHOLD ) {
#ifdef DEBUG
         printf("compute_intra_conflicts: ");
         printf("confl(%d,%d)= %.2f (TAKEN into account)\n",
                 node2, node1, confl);
#endif
         new_hit_rate = get_hit_rate( curr_node )*( 1-confl );
         set_hit_rate( curr_node, new_hit_rate );
      } 
#ifdef DEBUG
      else {
         printf("compute_intra_conflicts: ");
         printf("confl(%d,%d)= %.2f (NOT taken into account)\n",
                 node2, node1, confl);
      }
#endif

#ifdef DEBUG 
      if ( confl ) {
         printf("compute_intra_conflicts: ");
         printf("node2= %d, node1= %d, remove %.2f, new_hit_rate(%d)= %.2f\n\n",
                 node2, node1, confl, node2, new_hit_rate);
         printf("get_hit_rate(%d)= %.2f\n", node2, get_hit_rate( curr_node ));
      }
#endif
 
      next_path = get_next_node_path( next_path );
      node2 = get_node_path( next_path );
   }
}
 
float
update_intra_conflicts( header_path_t *header_path, c_sg_node_t *sgraph ) {
   int node;
   path_t *curr_path;
   c_sg_node_t *curr_node;
   float curr_node_hit_rate;
   int curr_node_n_i_fetches;
   float n_instr_hit_tmp;

#ifdef DEBUG 
   printf("update_intra_conflicts:\n");
#endif
   fprintf( stderr, "get_head_path(3)\n" );
   curr_path = get_head_path( header_path );
   n_instr_hit_tmp = 0.0;
 
   while( curr_path ) {
      node = get_node_path( curr_path );
      curr_node = get_node_sgraph( sgraph, node);
      curr_node_n_i_fetches = get_n_i_fetches( curr_node );

#ifdef DEBUG
      printf("\thit_rate(%d) (before)= %.2f\n",
              node, get_hit_rate( curr_node ));
      printf("\tcorrect_percen_path(%d)= %.2f\n",
              node, get_correct_percen_path( curr_path ));
#endif

      curr_node_hit_rate = get_hit_rate( curr_node ) *
                           get_correct_percen_path( curr_path );

      /* correction for LINESIZE > 2 */
      if ( curr_node_hit_rate < min_hit_rate ) {
         printf("min_hit_rate_correction (orig= %.2f)\n", curr_node_hit_rate);
         curr_node_hit_rate = min_hit_rate;
      }

#ifdef DEBUG
      printf("\tcurr_node_hit_rate= %.2f\n", curr_node_hit_rate);
#endif

      n_instr_hit_tmp += (float) (curr_node_n_i_fetches * curr_node_hit_rate);
      set_hit_rate( curr_node, get_hit_rate_after_path_exec( curr_path ));

#ifdef DEBUG
      printf("\thit_rate(%d) (after)= %.2f\n\n",
              node, get_hit_rate( curr_node ));
#endif
 
      curr_path = get_next_node_path( curr_path );
   }
   return n_instr_hit_tmp;
}
 
void
record_path_hit_rate( header_path_t *header_path, c_sg_node_t *sgraph )
{
/* This is the actual caching of the intra-conflicts information
 * just computed.
 */
   path_t *curr_path;
   int node;
   c_sg_node_t *curr_node;

   fprintf( stderr, "get_head_path(4)\n" ); 
   curr_path = get_head_path( header_path );

#ifdef DEBUG 
   printf("record_path_hit_rate:\n");
#endif

   while( curr_path ) {
      node = get_node_path( curr_path );
      curr_node = get_node_sgraph( sgraph, node );
      set_hit_rate_after_path_exec( curr_path, get_hit_rate( curr_node ));

#ifdef DEBUG
      printf("\tnode= %d, hit_rate_after_path= %.2f\n",
         get_node_path( curr_path ), get_hit_rate_after_path_exec( curr_path ));
#endif

      curr_path = get_next_node_path( curr_path );
   }
}
 
void
print_path_hit_rate( header_path_t *header_path, c_sg_node_t *sgraph )
{
   path_t *curr_path;
   int node;
   c_sg_node_t *curr_node;

#ifdef DEBUG 
   printf("print_path_hit_rate:\n");
#endif

   fprintf( stderr, "get_head_path(5)\n" ); 
   curr_path = get_head_path( header_path );
 
   while( curr_path ) {
      node = get_node_path( curr_path );
      curr_node = get_node_sgraph( sgraph, node );
#ifdef DEBUG
      printf("\tnode= %d, hit_rate=%f\n", 
             get_node_path( curr_path ), get_hit_rate( curr_node ));
#endif
      curr_path = get_next_node_path( curr_path );
   }
}

/***********************************************/
/* INTER-CONFLICTS analysis:                   */
/* conflicts between different path executions */
/***********************************************/
 
float
compute_inter_conflicts( path_cache_layout_t *layout1, 
                         path_cache_layout_t *layout2  )
{
   path_cache_layout_t *curr_seg1;
   path_cache_layout_t *curr_seg2;
   int s11, s12, s21, s22;
   int l1;
   float confl;
 
   l1 = 0;
   confl = 0;
   curr_seg1 = layout1;
   curr_seg2 = layout2;
 
   while ( curr_seg1 && curr_seg2 && (curr_seg1!=curr_seg2) ) {
      s11 = get_start_cache_segment( curr_seg1 );
      s12 = get_end_cache_segment( curr_seg1 );
      s21 = get_start_cache_segment( curr_seg2 );
      s22 = get_end_cache_segment( curr_seg2 );
      confl += calc_confl( s11, s12, s21, s22 );
      if (s12>s22)
         curr_seg2 = get_next_cache_segment( curr_seg2 );
      else if (s12<s22) {
         l1 += s12-s11+1;
         curr_seg1 = get_next_cache_segment( curr_seg1 );
      }
      else {  /* s12==s22 */
         l1 += s12-s11+1;
         curr_seg1 = get_next_cache_segment( curr_seg1 );
         curr_seg2 = get_next_cache_segment( curr_seg2 );
      }
   }
 
   while( curr_seg1 ) {
      s11 = get_start_cache_segment( curr_seg1 );
      s12 = get_end_cache_segment( curr_seg1 );
      l1 += s12-s11+1;
      curr_seg1 = get_next_cache_segment( curr_seg1 );
   }
   confl = confl/((float) l1);

#ifdef DEBUG
   printf("compute_inter_conflict: confl= %.2f\n", confl);
#endif

   return confl;
}
 
void
update_inter_conflicts( layout_path_pair_t *new_layout_path_pair, int history_pos )
{
   layout_path_pair_t *tmp_layout_path_pair = NULL;
   path_cache_layout_t *new_layout;
   path_cache_layout_t *tmp_layout = NULL;
   header_path_t *new_path;
   header_path_t *tmp_path = NULL;
   header_path_t *diff_path = NULL;
   path_vect_t *path_vect = NULL;
   c_sg_node_t *sgraph;
   int curr_pos;
   float confl;
   int new_task_id, new_path_id;
   int tmp_task_id, tmp_path_id;
   int end_history = FALSE;
 
   curr_pos = (history_pos-1+HISTORY_LENGTH)%(HISTORY_LENGTH);
   new_layout = get_layout_layout_path_pair( new_layout_path_pair );
   new_path = get_path_layout_path_pair( new_layout_path_pair );
   new_task_id = get_task_id_layout_path_pair( new_layout_path_pair );
   new_path_id = get_path_id_layout_path_pair( new_layout_path_pair );
   tmp_layout_path_pair = get_history_elem( curr_pos );
 
   while(( tmp_layout_path_pair ) && ( !end_history )) {

#ifdef DEBUG
      printf("\nupdate_inter_conflicts: history_pos= %d, curr_pos= %d\n",
              history_pos, curr_pos);
#endif

      tmp_layout = get_layout_layout_path_pair( tmp_layout_path_pair );
      tmp_path = get_path_layout_path_pair( tmp_layout_path_pair );
      tmp_task_id = get_task_id_layout_path_pair( tmp_layout_path_pair );
      tmp_path_id = get_path_id_layout_path_pair( tmp_layout_path_pair );

#ifdef DEBUG
      printf("new_(task_id,path_id) = (%d,%d), tmp_(task_id,path_id)= (%d,%d)\n
", new_task_id, new_path_id, tmp_task_id, tmp_path_id);
#endif

      if ( new_task_id == tmp_task_id ) {
         if ( new_path_id == tmp_path_id ) {
#ifdef DEBUG
            printf("update_inter_conflicts: break inter_conflicts\n");
#endif
            return;  /* when I find the same (task_id,path_id)
                      * I stop the update_inter_conflicts
                      */
         }

#ifdef DEBUG            
         printf("new_task_id == tmp_task_id !!!!\n");
         printf("new_path_id = %d, tmp_path_id= %d\n",
                 new_path_id, tmp_path_id);
#endif
 
         diff_path = alloc_header_path( 
                           get_task_num_c_sg_nodes( get_task( new_task_id )));
         path_vect = get_path_vect( diff_path );
         tmp_layout = create_diff_path_cache_layout(
                           tmp_path, new_path, new_task_id, diff_path);
         tmp_path = diff_path;
      }

      if ( tmp_layout ) {

         confl = lookup_confl_inter_table( new_task_id, new_path_id,
                                           tmp_task_id, tmp_path_id );
#ifdef DEBUG
         printf("confl = %.2f\n", confl);
#endif
         if (( confl < 0.0 ) || ( confl > 1.0 )) {

            confl = compute_inter_conflicts( tmp_layout, new_layout );
#ifdef DEBUG
            printf("**** compute_inter_conflicts\n");
#endif

            insert_in_confl_inter_table (new_task_id, new_path_id,
                                     tmp_task_id, tmp_path_id, confl );
         }
         else {
#ifdef DEBUG
            printf("**** used confl from confl_inter_table\n");
#endif
         }

#ifdef DEBUG 
         printf("update_inter_conflicts: task_id= %d, path_id= %d\n",
                 get_task_id_layout_path_pair( tmp_layout_path_pair ),
                 get_path_id_layout_path_pair( tmp_layout_path_pair ));
#endif
         sgraph = get_task_sgraph( get_task( tmp_task_id ));
         propagate_inter_in_intra_conflicts( tmp_path, sgraph, confl );
#ifdef DEBUG
         print_path_hit_rate( tmp_path, sgraph );
#endif
         if ( diff_path ) {
            free_path( diff_path );
            diff_path = NULL;  /* without this it crashes on linux! */
            free_path_vect( path_vect );
         }
      }
      curr_pos = (curr_pos-1+HISTORY_LENGTH)%HISTORY_LENGTH;
      if ( curr_pos==((history_pos-1+HISTORY_LENGTH)%HISTORY_LENGTH )) {
         end_history = TRUE;
      }
      tmp_layout_path_pair = get_history_elem( curr_pos );
   }
}
 
void
propagate_inter_in_intra_conflicts( header_path_t *header_path, 
                                 c_sg_node_t *sgraph, float confl )
{
   int node;
   path_t *curr_path;
   c_sg_node_t *curr_node;
   float curr_hit_rate;
   float new_hit_rate;
 
   if ( confl > THRESHOLD ) {
      fprintf( stderr, "get_head_path(6)\n" );
      curr_path = get_head_path( header_path );
 
      while( curr_path ) {
         node = get_node_path( curr_path );
         curr_node = get_node_sgraph( sgraph, node );
         curr_hit_rate = get_hit_rate( curr_node );
         new_hit_rate = curr_hit_rate - confl*curr_hit_rate;
         if ( new_hit_rate < 0 ) new_hit_rate = 0;
         set_hit_rate( curr_node, new_hit_rate );
         curr_path = get_next_node_path( curr_path );
      }
   }
}
 
path_cache_layout_t *
create_diff_path_cache_layout( header_path_t *header_old,
         header_path_t *header_new, int task_id, header_path_t *diff_path ) {
/* This function is used to update inter-conflicts in paths belonging
 * to the same task.
 * It first identifies the nodes in which the two paths differ.
 * The correction (inter_conflicts) will be applied only to those nodes.
 */
   int i, num_nodes, pos1, pos2;
   c_sg_node_t *curr_node;
   c_sg_node_t *sgraph;
   path_cache_layout_t *curr_layout = NULL;
   path_cache_layout_t *tmp_layout = NULL;
   int found;

#ifdef DEBUG 
   printf("create_diff_path_cache_layout\n");
#endif
   found = 0;
   sgraph = get_task_sgraph( get_task( task_id ));
   num_nodes = get_task_num_c_sg_nodes( get_task( task_id ));
   set_head_path( diff_path, NULL );
   set_tail_path( diff_path, NULL );
 
   for( i=0; i<num_nodes; i++ ) {
      pos1 = get_path_vect_pos( header_old, i );
      pos2 = get_path_vect_pos( header_new, i );
      if ( pos1 ) {
         if ( pos1 != pos2 ) {
            found = 1;
            curr_node = get_node_sgraph( sgraph,i );
#ifdef DEBUG
            printf("create_diff_path_cache_layout: %d\n", i);
            printf("hit_rate(%d)= %.2f\n", i, get_hit_rate( curr_node ));
#endif 
            add_node_in_path( diff_path, i, 1, 1 );
 
            tmp_layout = create_path_cache_layout( curr_node );
            curr_layout = update_cache_layout( curr_layout, tmp_layout );
#ifdef DEBUG
            print_path_cache_layout( curr_layout );
#endif
         }
      }
   }
 
   if ( !found ) {
#ifdef DEBUG
      printf("create_diff_path_cache_layout: NOT FOUND\n");
#endif
      return NULL;
   }
   else {
      return curr_layout;
   }
}

/***************************/
/* INTER-CONFLICTS hashing */
/***************************/


/*******************/
/* confl SELECTORS */
/*******************/

int
get_task_id1_inter( confl_inter_t *element ) {
   return element->task1;
}

void
set_task_id1_inter( confl_inter_t *element, int id ) {
   element->task1 = id;
}

int
get_task_id2_inter( confl_inter_t *element ) {
   return element->task2;
}

void
set_task_id2_inter( confl_inter_t *element, int id ) {
   element->task2 = id;
}

int
get_path_id1_inter( confl_inter_t *element ) {
   return element->path1;
}

void
set_path_id1_inter( confl_inter_t *element, int id ) {
   element->path1 = id;
}

int
get_path_id2_inter( confl_inter_t *element ) {
   return element->path2;
}

void
set_path_id2_inter( confl_inter_t *element, int id ) {
   element->path2 = id;
}

float
get_confl_inter( confl_inter_t *element ) {
   return element->confl;
}

void
set_confl_inter( confl_inter_t *element, float value ) {
   element->confl = value;
}

/**********************/
/* confl CONSTRUCTORS */
/**********************/

confl_inter_t *
create_inter_confl( int task_id1, int path_id1,
                    int task_id2, int path_id2, float confl ) {
   confl_inter_t *tmp;

   tmp = (confl_inter_t *)malloc(sizeof(confl_inter_t));
   if (tmp==NULL) {
      printf("Error in create_confl\n");
      exit(1);
   }
   else {
      set_task_id1_inter( tmp, task_id1 );
      set_path_id1_inter( tmp, path_id1 );
      set_task_id2_inter( tmp, task_id2 );
      set_path_id2_inter( tmp, path_id2 );
      set_confl_inter( tmp, confl );
   }
   return tmp;
}

void
free_inter_confl( confl_inter_t *segment ) {
   if( segment ) {
      free( segment );
   }
}

/*******************************/
/* confl_inter_table OPERATORS */
/*******************************/

void
alloc_confl_inter_table( ) {
   printf("alloc_confl_inter_table\n");
   confl_inter_table = st_init_table( st_ptrcmp, st_ptrhash );
}

void 
free_confl_inter_table( ) {
   confl_inter_t *tmp;
   confl_inter_t *tmp1;
   st_generator *gen; 

   st_foreach_item( confl_inter_table, gen, (char**)&tmp, (char**)&tmp1 ) {
      free_inter_confl( tmp1 );
   }
   st_free_table( layout_table );
}
 
int
make_confl_key ( int task1, int path1, int task2, int path2 ) {
   return (
   ((task1&KEY_TASK_CONFL_MASK)<<(2*KEY_PATH_CONFL_BITS+KEY_TASK_CONFL_BITS))|
   ((path1&KEY_PATH_CONFL_MASK)<<(KEY_PATH_CONFL_BITS+KEY_TASK_CONFL_BITS))|
   ((task2&KEY_TASK_CONFL_MASK)<<KEY_PATH_CONFL_BITS )|
    (path2&KEY_PATH_MASK));
}

float
lookup_confl_inter_table( int task1, int path1, int task2, int path2 ) {
   int key;
   confl_inter_t *tmp = NULL;

   key = make_confl_key( task1, path1, task2, path2 );
#ifdef DEBUG
   printf("lookup_confl_inter_table: key= %d\n", key);
#endif

   /*
   tmp = get_confl_from_table( key );
   */

    if ( st_lookup( confl_inter_table, (char *) key, (char **) &tmp )) {
#ifdef DEBUG
          printf("lookup_confl_inter_table: task1= %d, get_task_id1= %d\n", 
                 task1, get_task_id1_inter(tmp));
          printf("lookup_confl_inter_table: path1= %d, get_path_id1= %d\n", 
                 path1, get_path_id1_inter(tmp));
          printf("lookup_confl_inter_table: task2= %d, get_task_id2= %d\n", 
                 task2, get_task_id2_inter(tmp));
          printf("lookup_confl_inter_table: path2= %d, get_path_id2= %d\n", 
                 path2, get_path_id2_inter(tmp));
/*
          printf("lookup_confl_inter_table: confl= %.2f\n", get_confl_inter(tmp));
*/
#endif
          if ((get_task_id1_inter(tmp) == task1) &&
              (get_path_id1_inter(tmp) == path1) &&
              (get_task_id2_inter(tmp) == task2) &&
              (get_path_id2_inter(tmp) == path2))
          return get_confl_inter(tmp);  /* found path in layout table */
    }
    return -1.0;
}

void
insert_in_confl_inter_table( int task1, int path1, 
                             int task2, int path2, float confl ) {
   confl_inter_t *confl_inter = NULL;
   int key;

   confl_inter = create_inter_confl( task1, path1, task2, path2, confl );
   key = make_confl_key( task1, path1, task2, path2 );
   /* set_confl_in_table( key, confl_inter ); */
   st_insert( confl_inter_table, (char *) key, (char *) confl_inter ); 
}
