.\" /*
.\"  * Revision Control Information
.\"  *
.\"  * $Source: /projects/hwsw/hwsw/common/src/polis/polis_lib/help/polis.man,v $
.\"  * $Author: luciano $
.\"  * $Revision: 1.8 $
.\"  * $Date: 1996/07/15 16:16:41 $
.\"  *
.\"
.TH POLIS 1CAD "1 August 1994"
.SH NAME
polis \- System for Hardware/Software Codesign
.SH SYNOPSIS
.B polis
[options] [file]
.SH DESCRIPTION
\fBPOLIS\fR is a system for hardware/software codesign of embedded systems.
\fBPOLIS\fR starts from a description of a system of interacting
components in \fIshift\fP format
and produces an optimized implementation consisting of a hardware
part (in \fIblif\fP format) and a software part (in C).
.PP
\fBPOLIS\fR can be run in interactive mode accepting commands
from the user, or in batch mode reading commands
from a file or from the command line.
If no options are given on the command line, \fBPOLIS\fP will
enter interactive mode.
Otherwise, \fBPOLIS\fP will enter batch mode.
When running in batch mode, \fBPOLIS\fP reads its
input from the file given on the command line, or from standard input
if no filename is given;
output is directed to standard output, unless \fB-o\fP is used to
specify an output filename.
.PP
When \fBPOLIS\fP starts-up, it performs an initial source of the
file $POLIS/polis_lib/.polisrc.
Typically this defines a standard
set of aliases for various commands.
Following that the files ~/.polisrc and ./polisrc
are sourced for user-defined aliases at startup.
.SH OPTIONS
.TP 8
.B -c cmdline
Run \fBPOLIS\fR in batch mode, and execute \fBcmdline\fR.
Multiple commands are separated with semicolons.
.TP 8
.B -f script
Run \fBPOLIS\fR in the batch mode, and execute commands from the
file \fBscript\fR.
.TP 8
.B -s
Suppress sourcing the commands from the standard
startup script ($POLIS/polis_lib/.polisrc).
.de XX
.sp 1
.LP
.ti -5
.B
..
.de X1
.br
.ti -5
.B
..
.PP
.SH COMMAND SUMMARY
All commands are summarized below according
to their function : network manipulation (operations
on the component network),
input-output, status,
and miscellaneous.
Many commands from \fBSIS\fP are supported in \fBPOLIS\fP;
the user is referred to the \fBSIS\fP man page for details.
.sp 2
.TS
center box;
c s
l l.
\fBNetwork Manipulation Commands\fP
_
build_sg	create the s-graphs for software synthesis
calc_cost	calculate total software cost
flatten_all	flatten the hierarchy of a component network
flatten_comp	flatten one NET node
gen_os	generate a C-file for the components+os
partition	create the hw/sw partition with interface components
set_impl	sets the component implementations to hardware or software
sg_to_c	creates C-files from the s-graphs.
.TE
.sp 2
.TS
center box;
c s
l l.
\fBInput-Output Commands\fP
_
read_shift	read a component network in SHIFT format
read_cost_param	read cost parameters for software on a particular microprocessor
write_esterel	write the connectivity information in esterel format
write_shift	write a component network in SHIFT format
.TE
.sp 1
.TS
center box;
c s
l l.
\fBNetwork Status Commands\fP
_
list	list the nodes in the SHIFT network
print_cost	assign costs to each node in the s-graph and prints them out
print_polis	print out the nodes of a component network
print_sg	print out the S-graphs
print_sg_stats	print statistics about the size of the s-graphs
.TE
.sp 1
.TS
center box;
c s
l l.
\fBMiscellaneous\fP
_
polis_sim	simulation environment for hardware partition
simcfsm	simulate a SHIFT network of components
.TE
.sp .5i
.SH COMPONENT LIST ARGUMENTS
Most commands which take a node (cfsm or net or function or
instance) as an argument
also take a list of components as an
argument.  This is referred to, for example,
as a \fBcfsm-list\fP (\fBcomp-list\fP)
in the documentation below.
Omitting the argument implies that all the components should be considered.
.PP
Each element of the list is a \fBpathname\fP. A pathname is a list of
\fBlevels\fP, separated by '@'. A level is a list of \fBpatterns\fP, separated
by '&'. A pattern can be:
.br
a name (which matches all the components with the given
name) 
.br
a comma-separated 
list of implementations (NONE, HW, SW, USER, INTERFACE or ALL)
preceded by either 'i=' or 'I='
(which matches all the components with any one of
the given implementations in the first
case, and all the instances of a component with any one of 
the given implementations in
the second case)
.br
a comma-separated 
list of types (CFSM, NET, FUNC, COMP or ALL) 
preceded by either 'i=' or 'I='
(which matches all the components with any one of
the given types in the first
case, and all the instances of a component with any one of 
the given types in
the second case)
.br
a model name preceded by 'M=' (which matches all its instances).
.PP
For example, pathname
 sample@T=CFSM&i=SW
chooses all the elements of net \fIsample\fP with
a CFSM model which have been assigned a software implementation.
.PP
For commands that operate on a corresponding Boolean network and take
nodes as arguments,
\fBi()\fP can be used to specify the primary
inputs of the network and \fBo()\fP to specify the primary
outputs of the network.
\fBi(cfsm)\fP is used to specify the
direct fanin of \fIcfsm\fP, and \fBo(cfsm)\fP for the
direct fanout of \fIcfsm\fP.
.sp .5i
.SH STANDARD ALIASES
When \fBPOLIS\fP starts, it executes commands from a system startup file
($(POLIS)/sis_lib/.polisrc).
This defines a standard set of
aliases, and then sources the files ~/.polisrc
and ./polisrc to allow
users to define their own set of aliases.  The default alias set includes
the following aliases which have proven useful.  Note that many of the aliases
are intended for compatibility with \fBSIS\fP Version #1.2.
.sp 2
.TS
center box;
c s s
c c c
l l l.
\fBStandard Aliases\fP
alias	command	description
_
rsh	read_shift	read a component network in shift format
we	write_esterel	write the connectivity information in esterel format
wsh	write_shift	write a component network in shift format
.TE
.bp
.SH DETAILED COMMAND DESCRIPTIONS
.PP
.so build_sg.1
.so flatten_all.1
.so flatten_comp.1
.so gen_os.1
.so list.1
.so partition.1
.so polis_sim.1
.so read_cost_param.1
.so read_shift.1
.so print_cost.1
.so print_polis.1
.so print_sg.1
.so print_sg_stats.1
.so set_impl.1
.so sg_to_c.1
.so simcfsm.1
.so write_esterel.1
.so write_shift.1
.SH FILES
$POLIS/polis_lib/.polisrc
.br
$POLIS/sis_lib/*
.SH "SEE ALSO"
strl2shift, sis(1CAD)
.SH AUTHORS
Felice Balarin, 
Massimiliano Chiodo, 
Daniel Engels, 
Paolo Giusto, 
Wilsin Gosti,
Harry Hsieh,
Attila Jurecska, 
Luciano Lavagno, 
Claudio Passerone,
Ellen Sentovich, 
Kei Suzuki,
Bassam Tabbara,
Sherman Yee,
Alberto Sangiovanni-Vincentelli
.SH OTHER CONTRIBUTORS
\fBPOLIS\fP is based on \fBSIS\fP, and therefore on includes source
code written by its authors.
.SH BUGS

.SH COMMENTS
