/\./ {
	# MUST BE FIRST
	if (doing_names || doing_const) {
		doing_names = 0
		doing_const = 0
		names[nnames] = names[nnames] "\t\t'0';\n"
		nnames++
	}
	# fallthrough
}
/\.model/ {
	model = $2
	found_model[$2] = 1
	for (s =0; s < nsubckts; s++) {
		printed_subckt[name[s]] = 0
	}
	nsubckts = 0
	nparams = 0
	for (i = 0; i < nsignals; i++) {
		added_signal[signals[i]] = 0
	}
	nsignals = 0
	for (i = 0; i < nnames; i++) {
		names[i] = ""
	}
	nnames = 0
	nlatches = 0
	for (i = 0; i < ninputs; i++) {
		external[inputs[i]] = 0
	}
	for (i = 0; i < noutputs; i++) {
		external[outputs[i]] = 0
	}
	ninputs = 0
	noutputs = 0
	next
}
/\.inputs/ {
	for (i = 2; i <= NF; i++) {
		inputs[ninputs++] = $i
		external[$i] = 1
		is_input[model "%" $i] = 1
	}
	next
}
/\.outputs/ {
	for (i = 2; i <= NF; i++) {
		outputs[noutputs++] = $i
		external[$i] = 1
		is_output[model "%" $i] = 1
	}
	next
}
/\.subckt/ {
	if (! found_model[$2]) {
		printf "error: undefined .subckt %s\n", $2
	}
	name[nsubckts] = $2
	firstparam[nsubckts] = nparams
	lastparam[nsubckts] = nparams + NF/2 - 1
	for (i = 3; i <= NF; i+=2) {
		formals[nparams] = $i
		actuals[nparams] = $(i+1)
		if (! external[$(i+1)] && ! added_signal[$(i+1)]) {
			signals[nsignals++] = $(i+1)
			added_signal[$(i+1)] = 1
		}
		nparams++
	}
	nsubckts++
	next
}
/\.names/ {
	if (! external[$NF] && ! added_signal[$NF]) {
		signals[nsignals++] = $NF
		added_signal[$NF] = 1
	}
	if (NF == 2) {
		names[nnames] = "\t" $NF " <= "
		doing_const = 1
	}
	else {
		names[nnames] = "\t" $NF " <= "
		for (i = 2; i < NF; i++) {
			io[i-2] = $i
		}
		nio = NF - 2
		doing_names = 1
		skiptab = 1
	}
	next
}
/^[ ]*[10-]/ {
	if (doing_names) {
		if ($2 != "1") {
			printf "error: invalid output %s at line %d\n", $2, NR
		}
		if (skiptab) {
			names[nnames] = names[nnames] "("
			skiptab = 0
		}
		else {
			names[nnames] = names[nnames] "\t\t("
		}
		for (i = 0; i < nio; i++) {
			if (substr($1, i + 1, 1) == "1") {
				names[nnames] = names[nnames] " " io[i] " AND "
			}
			else if (substr($1, i + 1, 1) == "0") {
				 names[nnames] = names[nnames] " (NOT " io[i] " ) AND "
			}
		}
		names[nnames] = names[nnames] "'1') OR "
		nnames++
	}
	else if (doing_const) {
		names[nnames] = names[nnames] "'" $1 "' OR "
	}
	else {
		printf "error: PLA outside .names at line %d\n", NR
	}
	next
}
/\.latch/ {
	latches[nlatches++] = "\t\t" $3 " <= " $2 ";"
	if (! external[$3] && ! added_signal[$3]) {
		signals[nsignals++] = $3
		added_signal[$3] = 1
	}
}
/\.end/ {
	# print entity interface
	printf "ENTITY %s IS\n\tPORT(\n", model
	for (i = 0; i < ninputs; i++) {
		printf "\t\t%s\t: in BIT;\n", inputs[i]
	}
	for (i = 0; i < noutputs; i++) {
		printf "\t\t%s\t: buffer BIT;\n", outputs[i]
	}
	printf "\t\tCLOCK\t: in BIT);\nEND %s ;\n\n", model
	printf "ARCHITECTURE POLIS OF %s IS\n", model

	# print components
	for (s =0; s < nsubckts; s++) {
		if (! printed_subckt[name[s]]) {
			printf "\tCOMPONENT %s \n\t\tPORT(\n", name[s]
			for (i = firstparam[s]; i < lastparam[s]; i++) {
				if (is_input[name[s] "%" formals[i]]) {
					printf "\t\t\t%s\t: in BIT;\n", formals[i]
				}
				else if (is_output[name[s] "%" formals[i]]) {
					printf "\t\t\t%s\t: buffer BIT;\n", formals[i]
				}
				else if (found_model[name[s]]) {
				# otherwise already reported
					printf "error: undefined formal %s of %s\n", formals[i], name[s]
				}
			}
			printf "\t\t\tCLOCK\t: in BIT);\n\tEND COMPONENT;\n\n"
			printed_subckt[name[s]] = 1
		}
	}

	# print signals
	for (i = 0; i < nsignals; i++) {
		printf "SIGNAL %s\t: BIT;\n", signals[i]
	}

	# print architecture
	printf "\nBEGIN\n"
	for (s =0; s < nsubckts; s++) {
		printf "\tinst%d: %s PORT MAP(\n", s, name[s]
		for (i = firstparam[s]; i < lastparam[s]; i++) {
			printf "\t\t%s => %s ,\n", formals[i], actuals[i]
		}
		printf "\t\tCLOCK => CLOCK);\n\n"
	}

	# print data flow with logic
	for (i = 0; i < nnames; i++) {
		printf "%s\n", names[i]
	}

	if (nlatches) {
		# print a process with latches
		printf "\tPROCESS BEGIN\n"
		printf "\t\tWAIT UNTIL CLOCK'EVENT AND CLOCK = '1';\n"
		for (i = 0; i < nlatches; i++) {
			printf "%s\n", latches[i]
		}
		printf "\tEND PROCESS;\n"
	}

	printf "END POLIS;\n\n\n"
}
