/*
 *
 * 	vers=1.05
 *
 *	c.ld - command file for C for 68HC11
 *
 *	Configuration is a Motorola MC68HC11EVB evaluation board.
 *	This is designed for standalone programs that are used in place of the
 *	BUFFALO monitor.  The executable code and initialized data will
 *	placed in an 8Kx8 EPROM at 0xe000. There is 8K of RAM at 0xc000.
 */

set H11RAM = 0x00;		/* page containing 68HC11 base page */
set H11REG = 0x01;		/* page containing 68HC11 registers */

set H11VECSIZE = 42;		/* size of 68HC11 vectors */
set H11RAMREG = (H11RAM<<4)|H11REG; /* mask for setting H11INIT in start.s */
set H11REGORG = (H11REG<<12)	/* origin of 68HC11 registers */
set H11VECORG = 0x10000-H11VECSIZE; /* origin of 68HC11 vectors */

section	.text text origin 0x8000;	/* moved from 0xE000 LL */
section .const data = .const, .strings;
section .init;

section .data data origin 0x4000 copiedfrom .init = .data;
section .bss bss comms;
section .heap bss minsize 2048;

section .H11REGS bss origin H11REGORG;
section .H11VEC data origin H11VECORG maxsize H11VECSIZE;
section .base bss origin 0;

check endof(.text) > 0xffff fatal "Code area too large";
check endof(.bss) >= 0x5fff fatal "RAM area too large";
check sizeof(.base) > 256 fatal "Base page too large";

set _ramstart = startof(.bss);
set _ramend = endof(.bss);
set _heapstart = startof(.heap);
set _heapend = endof(.heap);
set _stackstart = endof(.heap);
set _stackend = 0x6000;		/* 0x5fff is the end of the RAM		*/
set _initstart = startof(.init);
set _initend = endof(.init);
set _datastart = startof(.data);
set _dataend = endof(.data);

check _stackend - _stackstart < 256 fatal "Stack too small";

-b 'system_setup.o11' 			/* start up routine			*/

readline;			/* read the command line		*/

-lc				/* use the C library			*/
-lcio				/* C i/o library			*/
-lm				/* math library				*/
-lgen				/* general library			*/
