/* This file should be copied locally and filled in with any user-defined
 * type access, parsing and formatting functions. It is enclosed ONLY
 * in the RTOS and Ptolemy. The parts in the #ifdef UNIX section are used
 * only on a workstation, and not on the target microcontroller.
 * Please, follow the enclosed sample template for complex integer numbers
 * (see also loc_types.h).
 */

#ifdef DUMMY_SAMPLE
#include "loc_types.h"

/* function called at RTOS initialization; v is the value in the .r SHIFT
 * statement, or the first assigned value in Esterel.
 */

void init_complex (c, v)
complex *c;
int v;
{
	c->r = v;
	c->i = v;
}

#ifdef UNIX
/* function called by RTOS and Ptolemy to format the type as a string for
 * debug printout.
 * E.g. if ocmpl is a complex output event, "ocmpl(12,3)"
 * is printed by the RTOS and "12,3" by Ptolemy if the real part has value
 * 12 and the imaginary part has value 3.
 */

char *format_complex (c)
complex c;
{
	static char s[80];

	sprintf (s, "%d,%d", c.r, c.i);
	return s;
}

/* function called by RTOS and Ptolemy to parse an input value.
 * E.g. if icmpl is a complex input event, "icmpl(12,3)"
 * typed at the RTOS prompt or "12,3" used in a Ptolemy initializer string set
 * the real part to 12 and the imaginary part to 3.
 */

void parse_complex (c, s)
complex *c;
char *s;
{
	sscanf (s, "%d,%d", &(c->r), &(c->i));
}
#endif /*UNIX*/
#endif /*DUMMY_SAMPLE*/
