# Tcl/Tk source for a display that shows the values of the star inputs
#
# Author: Edward A. Lee
# Version: @(#)tkShowValues.tcl	1.15	3/3/95
#
# Copyright (c) 1990-1995 The Regents of the University of California.
# All rights reserved.
# 
# Permission is hereby granted, without written agreement and without
# license or royalty fees, to use, copy, modify, and distribute this
# software and its documentation for any purpose, provided that the
# above copyright notice and the following two paragraphs appear in all
# copies of this software.
# 
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
# ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
# THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
# 
# THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
# PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
# CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
# 
# 						PT_COPYRIGHT_VERSION_2
# 						COPYRIGHTENDKEY
# See the file $PTOLEMY/copyright for copyright notice,
# limitation of liability, and disclaimer of warranty provisions.
#

set ${starID}(put_in_control_panel) 0
set ${starID}(wait_between_outputs) 0
set s $ptkControlPanel.label_$starID

# If a window with the right name already exists, we assume it was
# created by a previous run of the very same star, and hence can be
# used for this new run.  Some trickiness occurs, however, because
# parameter values may have changed, including the number of inputs.

if {[winfo exists $s]} {
    set window_previously_existed 1
} {
    set window_previously_existed 0
}

if {!$window_previously_existed} {

    proc PolisEnergyMakeWindow {putInCntrPan win label numInputs univ starID} {
	    if {![winfo exists $win]} {
	        global ptkControlPanel
            if {$putInCntrPan} {
	            frame $win -bd 3
	            pack after $ptkControlPanel.low $win top
            } {
                toplevel $win -bd 3
                wm title $win "Energy"
                wm iconname $win "Energy"
            }

            frame $win.f
            message $win.msg -width 12c -padx 2m -pady 2m \
                    -font *-times-bold-r-normal--*-140-*-*-*-*-*-*
            pack append $win $win.msg {top expand fillx} $win.f top
	    }
	    set stname [lindex $label 0]
	    $win.msg configure -text $stname

        # The following flag is used if the waitBetweenOutputs parameter is set
    	
    	global $starID
	    set ${starID}(waiting) 0
	    catch {destroy $win.f.left}
	    catch {destroy $win.f.right}
	    frame $win.f.left
	    frame $win.f.right
        for {set i 0} {$i < [expr $numInputs-1]} {incr i} {
            set stname [lindex $label [expr $i+1]]
    	    label $win.f.left.label$i -text $stname -padx 2m
    	    label $win.f.right.val$i -width 20 -relief sunken -padx 2m
	        pack $win.f.left.label$i -in $win.f.left -side top -anchor w
	        pack $win.f.right.val$i -in $win.f.right -side top -anchor w
        }
        label $win.f.left.label$i -text Timestamp -padx 2m
    	label $win.f.right.val$i -width 20 -relief sunken -padx 2m
	    pack $win.f.left.label$i -in $win.f.left -side top -anchor w -fill x
	    pack $win.f.right.val$i -in $win.f.right -side top -anchor w -fill x
        pack $win.f.left -in $win.f -side left -fill x
        pack $win.f.right -in $win.f -side right -fill x
        set ${starID}(breakp) 1
        catch {destroy $win.dlg}
        frame $win.dlg
        checkbutton $win.dlg.break -text "Stop at each fire" -variable \
                    ${starID}(breakp) -anchor w -padx 2m -pady 2m
        label $win.dlg.stopped -text "" -fg #f00 -padx 2m -pady 2m
        pack $win.dlg.break $win.dlg.stopped -in $win.dlg -side left -fill x
        pack $win.dlg -in $win -side top -fill x
        catch {destroy $win.b}
        frame $win.b
        set ${starID}(nodismiss) 1
        button $win.b.ok -text "DISMISS" -command \
		    "catch {incr ${starID}(waiting)}
		     catch {set ${starID}(breakp) 0}
		     catch {set ${starID}(nodismiss) 0}
		     destroy $win"
        button $win.b.cont -text "Continue" -command \
            "catch {incr ${starID}(waiting)}"
        pack $win.b.cont $win.b.ok -in $win.b -side left -fill x
        pack append $win $win.b {top fillx}
	    if {[set ${starID}(breakp)]} {
	    # Arrange for the waiting variable
	    # to be updated if the run status changes (for example,
	    # a halt is requested)
	        global ptkRunFlag
	        trace variable ptkRunFlag($univ) w PolisEnergyReleaseWait
	    }
    }
}

proc PolisEnergyReleaseWait {runflag univ op} "
	global ${starID}
	incr ${starID}(waiting)
"

PolisEnergyMakeWindow [set ${starID}(put_in_control_panel)] $s \
    [set ${starID}(label)] [set ${starID}(numInputs)] [curuniverse] $starID

if {!$window_previously_existed} {

    proc PolisEnergySetEnergyValues {starID numInputs win} {
        global $starID
        if {[set ${starID}(nodismiss)]} {
        	set c $win.f
			set inputVals [grabEnergyInfo_$starID]
    		for {set i 0} {$i < $numInputs} {incr i} {
            	set in [lindex $inputVals $i]
            	$win.f.right.val$i configure -text $in
	    	}
	    }
    }

    proc PolisEnergyWait {flag starID numInputs win univ} {
        global $starID
    	if {[set ${starID}(breakp)]} {
	        set ${starID}(waiting) 0
            $win.b.cont configure -state normal
            $win.dlg.stopped configure -text "STOP"
	        tkwait variable ${starID}(waiting)
	        catch {$win.dlg.stopped configure -text ""}
            catch {$win.f.cont configure -state disabled}
        }
    }
}

# In the following definition, the value of $s
# is evaluated when the file is sourced.
proc goTcl_$starID {starID} "
    PolisEnergySetEnergyValues $starID [set ${starID}(numInputs)] $s
	PolisEnergyWait [set ${starID}(breakp)] $starID \
		[set ${starID}(numInputs)] $s [curuniverse]
"

proc destructorTcl_$starID {starID} {
        global $starID
        if {[set ${starID}(put_in_control_panel)]} {
            # Remove the meters from the control panel, if they still exist
            global ptkControlPanel
            destroy $ptkControlPanel.label_$starID
        }
}
unset s

