defstar {
  name { ad_base }
  domain { DE }
  derivedfrom { Polis }
  desc { Star generated by POLIS
68hc11:
    min_time: 288
    max_time: 1488
    tot_size: 2560
  }
  hinclude { "uC_types.h" }
  hinclude { "swlib.h" }
  ccinclude { "InfString.h" }
  hinclude { "DETarget.h" }
  input {
    name { e_AD_START }
    type { int }
  }
  input {
    name { e_ECLK }
    type { int }
  }
  input {
    name { s_PE0 }
    type { int }
    desc { sensor }
  }
  input {
    name { s_PE1 }
    type { int }
    desc { sensor }
  }
  input {
    name { s_PE2 }
    type { int }
    desc { sensor }
  }
  input {
    name { s_PE3 }
    type { int }
    desc { sensor }
  }
  input {
    name { s_PE4 }
    type { int }
    desc { sensor }
  }
  input {
    name { s_PE5 }
    type { int }
    desc { sensor }
  }
  input {
    name { s_PE6 }
    type { int }
    desc { sensor }
  }
  input {
    name { s_PE7 }
    type { int }
    desc { sensor }
  }
  output {
    name { e_AD_END }
    type { int }
  }
  output {
    name { e_ADR1 }
    type { int }
  }
  output {
    name { e_ADR2 }
    type { int }
  }
  output {
    name { e_ADR3 }
    type { int }
  }
  output {
    name { e_ADR4 }
    type { int }
  }
  public {
    double now;
    int resourceId;
    double clkFreq;
    /* Input event flags */
    int e_AD_START_flag;
    int e_ECLK_flag;
    /* Settable constants */
    /* State and output variables */
    int _st;
    int count;
    int pe;
  }
  private {
    double *_time, totnf;
    double _nextf, _latency;
    FILE *fpfire;
    FILE *fpover;
  }
  state {
    name { implem }
    type { string }
    default { "{implem}" }
    desc { Valid values: HW SW }
  }
  state {
    name { resourceName }
    type { string }
    default {"Resource1"}
    desc { resource to be used for simulation }
  }
  state {
    name { Clock_freq }
    type { float }
    default{ Clock_freq }
    desc { CPU Clock frequency }
  }
  state {
    name { policy }
    type { string }
    default { "{SCHEDULER}" }
    desc { Scheduling Policy to be used for simulation }
  }
  state {
    name { Priority }
    type { int }
    default { 0 }
    desc { Priority: 0 -> lower }
  }
  state {
    name { firename }
    type { string }
    default { "{Firingfile}" }
    desc { File to write firing information }
    attributes { A_NONSETTABLE }
  }
  state {
    name { overname }
    type { string }
    default { "{Overflowfile}" }
    desc { File to write overflow information }
    attributes { A_NONSETTABLE }
  }


  constructor {
    delayType = TRUE;
    _delay = 0.0;
  }
  method {
    name { emitEventToIntQ }
    access { public }
    type { int }
    arglist { "( int outputPort, double delay )" }
    code {
      static const int ce_AD_END = 1;
      static const int ce_ADR1 = 2;
      static const int ce_ADR2 = 3;
      static const int ce_ADR3 = 4;
      static const int ce_ADR4 = 5;
      double emitTime;
      OutDEPort* tl;
      if (needsSharedResource) {
           StarLLCell *temp = (StarLLCell*)storeList->getAndRemove();
           if (temp == 0) temp = new StarLLCell(-1, -1);
           emitTime = now + (delay/clkFreq);
           temp->time = emitTime;
           temp->outputPort = outputPort;
           emittedEvents->append(temp);
      } else {
           if (now < 0) emitTime = 0.0;
           else emitTime = now + (_latency/clkFreq);
           switch(outputPort) {
             case ce_AD_END:
                tl = & e_AD_END;
                tl->put(emitTime) << 1;
                break;
             case ce_ADR1:
                tl = & e_ADR1;
                tl->put(emitTime) << v_ADR1;
                break;
             case ce_ADR2:
                tl = & e_ADR2;
                tl->put(emitTime) << v_ADR2;
                break;
             case ce_ADR3:
                tl = & e_ADR3;
                tl->put(emitTime) << v_ADR3;
                break;
             case ce_ADR4:
                tl = & e_ADR4;
                tl->put(emitTime) << v_ADR4;
                break;
            default:
               break;
         }
      }
      return 1;
    } 
  } 
  method {
    name { emitEvent }
    access { public }
    type { void }
    arglist { "(StarLLCell* cell, double emitTime)" }
    code {
      static const int ce_AD_END = 1;
      static const int ce_ADR1 = 2;
      static const int ce_ADR2 = 3;
      static const int ce_ADR3 = 4;
      static const int ce_ADR4 = 5;
      OutDEPort * tl;
      switch( cell->outputPort ) {
            case ce_AD_END:
                tl = & e_AD_END;
                tl->put(emitTime) << 1;
                break;
            case ce_ADR1:
                tl = & e_ADR1;
                tl->put(emitTime) << v_ADR1;
                break;
            case ce_ADR2:
                tl = & e_ADR2;
                tl->put(emitTime) << v_ADR2;
                break;
            case ce_ADR3:
                tl = & e_ADR3;
                tl->put(emitTime) << v_ADR3;
                break;
            case ce_ADR4:
                tl = & e_ADR4;
                tl->put(emitTime) << v_ADR4;
                break;
            default:
               break;
      }
      storeList->append(cell);
    }
  }
  method {
    name { getDelay }
    access { public }
    type { double }
    arglist { "()" }
    code {
      // need to add 1 clock cycle for SW stars so that dummy event is seen
      if (needsSharedResource) return (_delay)/clkFreq;// FIXME: a hack
      else return (totnf);
    }
  }
  method {
    name { CleanUp }
    access { public }
    type { void }
    arglist { "( int emitted )" }
    code {
      if ( emitted ) {
        e_AD_START_flag = 0;
        e_ECLK_flag = 0;
      }
    }
  }
  code {
#define DO_ESTIMATE  (do_estimate=1)
#define DO_NOT_ESTIMATE  (do_estimate=0)
#define illegal_cpu() 0
#define FUZZY(n,d) {if ((ad_base_Star->needsSharedResource)&&(do_estimate)) { \
                     _delay += _cache_delay[n]?_cache_delay[n]:_cache_delay[n]=ad_base_Star->FuzzyDelay(d); \
                 }}
#define ENERGY(e)
#define DELAY(d) {if ((ad_base_Star->needsSharedResource)&&(do_estimate)) {_delay += (d);}}
/* Input event macros */
#define detect_e_AD_START_to_z_ad_base_0 ad_base_Star->e_AD_START_flag
#define detect_e_ECLK_to_z_ad_base_0 ad_base_Star->e_ECLK_flag
/* Output event macros */
#define emit_e_AD_END() ad_base_Star->emitEventToIntQ( e_AD_END, _delay )
#define emit_e_ADR1() ad_base_Star->emitEventToIntQ( e_ADR1, _delay )
#define emit_e_ADR2() ad_base_Star->emitEventToIntQ( e_ADR2, _delay )
#define emit_e_ADR3() ad_base_Star->emitEventToIntQ( e_ADR3, _delay )
#define emit_e_ADR4() ad_base_Star->emitEventToIntQ( e_ADR4, _delay )
#define startup( p )
#define cleanup( p, e ) ad_base_Star->CleanUp( e )
#define always_cleanup( p ) ad_base_Star->CleanUp( 1 )
#define cpu ad_base_Star->resourceId
#define SET_LT(a,b) ad_base_Star->Set_LT((a),(b))
/* Settable constant macros */
/* State and output variable macros */
#define v__st ad_base_Star->_st
#define v_count ad_base_Star->count
#define v_pe ad_base_Star->pe
    DEad_base * ad_base_Star = NULL;
    static int do_estimate = 1;
    static double _delay = 0.0;
    static double _cache_delay[121];
    /* Output event constants */
    static const int e_AD_END = 1;
    static const int e_ADR1 = 2;
    static const int e_ADR2 = 3;
    static const int e_ADR3 = 4;
    static const int e_ADR4 = 5;
    /* Input variable declarations */
    static int v_AD_START;
    static int v_PE0;
    static int v_PE1;
    static int v_PE2;
    static int v_PE3;
    static int v_PE4;
    static int v_PE5;
    static int v_PE6;
    static int v_PE7;
    /* Output variable declarations */
    static int v_ADR1;
    static int v_ADR2;
    static int v_ADR3;
    static int v_ADR4;

    static unsigned int path_id;
#include "z_ad_base_0.c"
  }
  setup {
    DETarget *t;
    char stemp[1024];

    t = (DETarget *) target();
    if (!(t->isResource())) {
      strcpy( stemp, "Please, set target to Resource Contention Scheduler" );
      Error::abortRun( *this, stemp );
    }
    timeOfArrival = -1;
    _nextf = 1.0;
    _latency = 1.0;
    DERCStar::Closeflow();
    strcpy(resource, resourceName);
    strcpy(stemp, implem);
    needsSharedResource = getImplem( stemp );
    if ( needsSharedResource < 2 ) {
    }
  }
  begin {
    char stemp[ 1024 ];
    InfString name;
    int saveimpl;

    if ( needsSharedResource < 2 ) {
      priority = Priority;
      ad_base_Star = this;
      /* set variables which reflect Star parameters*/
      strcpy( stemp, policy);
      schedPolicy = resourcePointer->getSchedPolicy( stemp );
      sprintf( stemp, "%d", priority );
      lastFiredPr = resourcePointer->registerPriority( stemp );
      clkFreq = Clock_freq;
      clkFreq = (1000000.0)*clkFreq;
      resourcePointer->registerClock( clkFreq );
      totnf = 1.0/clkFreq;
      emittedEvents = new SequentialList();
      storeList = new SequentialList();
      _delay = 0.0;
      int i;
      for ( i=0; i<=120; i++ ) {
         _cache_delay[i] = 0;
      }
      /* State and output variables initialization */
      _st = 0;
      /* Settable constants initialization */
      /* Input event flag initialization */
      e_AD_START_flag = 0;
      e_ECLK_flag = 0;
      if ( needsSharedResource ) {
        strcpy( stemp, resource );
        if (strchr(stemp, '.')) *(strchr(stemp,'.')) = '\0';
        if (( resourceId = (int) ReadCostParam( stemp )) == -1 ) {
          strcat( stemp, ": not a valid CPU" );
          Error::abortRun( *this, stemp );
        }
      }
      strcpy( stemp, firename );
      if (strcmp(stemp,"{Firingfile}")) fpfire = DERCStar::Openfiring(stemp);
      strcpy( stemp, overname );
      if (strcmp(stemp,"{Overflowfile}")) fpover = DERCStar::Openoverflow(stemp);
      now = -1.0;
      saveimpl = needsSharedResource;
      needsSharedResource = 0;
      _t_z_ad_base_0(0,0);

      needsSharedResource = saveimpl;
    } else {
      strcpy( stemp, "Implementation not supported!" );
      Error::abortRun( *this, stemp );
    }
  }
  go {
    char stemp[ 1024 ];
    int runflag = 0;
    StringList name;

    now = arrivalTime;
    name = Block::fullName();
    ad_base_Star = this;
    if (feedbackIn->dataNew && (int)((*feedbackIn)%0) == 0) {
        DERCStar* starUsingResource = resourcePointer->getTopCell()->star;
        if (((StarLLCell*)emittedEvents->head())->time != now ) {
            refireAtTime(((StarLLCell*)emittedEvents->head())->time, 0.0);
        } else {
            if ( starUsingResource != this ) {
                Error::abortRun( *this, "Star should not reach this point if preempted!" );
            }
            if(emittedEvents->head() == emittedEvents->tail() ) {
                // This is the dummy event indicating end of resource usage 
                resourcePointer->removeFinishedStar(this); 
                emittedEvents->getAndRemove(); 
                assert(emittedEvents->getAndRemove() == 0); 
                if ( fpfire ) {
                    sprintf( stemp, "%s: %f %d end %s\n", (const char*) name, now, priority, resource );
                    DERCStar::Printfiring( stemp );
                }
            } else { 
                StarLLCell *temp; 
                while((temp = (StarLLCell*)emittedEvents->head())->time == now && (emittedEvents->head() != emittedEvents->tail()) ) { 
                    StarLLCell *temp = (StarLLCell*)emittedEvents->getAndRemove(); 
                    assert(now == temp->time); 
                    emitEvent(temp, now); 
                } 
                refireAtTime(((StarLLCell*)emittedEvents->head())->time, 0.0); 
            }
            feedbackIn->dataNew = FALSE;
        }
    } else { // This star is firing for the first time, and has resource if req 
        if ( e_AD_START.dataNew ) {
          e_AD_START_flag = 1;
          e_AD_START.dataNew = FALSE;
          runflag = 1;
        }
        if ( e_ECLK.dataNew ) {
          e_ECLK_flag = 1;
          e_ECLK.dataNew = FALSE;
          runflag = 1;
        }

      // Read the input values 
      v_AD_START = (e_AD_START)%0;
      v_PE0 = (s_PE0)%0;
      v_PE1 = (s_PE1)%0;
      v_PE2 = (s_PE2)%0;
      v_PE3 = (s_PE3)%0;
      v_PE4 = (s_PE4)%0;
      v_PE5 = (s_PE5)%0;
      v_PE6 = (s_PE6)%0;
      v_PE7 = (s_PE7)%0;
      _delay = 0.0;
      if (runflag == 1) {
        _t_z_ad_base_0(0,0);

        // write to firing file 
        if ( fpfire ) {
            sprintf( stemp, "%s: %f %d start %s\n", (const char*) name, now, priority, resource );
            DERCStar::Printfiring( stemp );
        }
        if (needsSharedResource) emitEventToIntQ( 0, _delay );// emit dummy event 
        timeOfArrival = -1;// reset this as have finished this firing 
        if (needsSharedResource) refireAtTime(((StarLLCell*)emittedEvents->head())->time, 0.0);
      }
    }
  }
}
