defstar {
  name { oc_prog_rel }
  domain { DE }
  derivedfrom { Polis }
  desc { 
    Programmable Relative Output Compare
  }
  hinclude { "uC_types.h" }
  hinclude { "swlib.h" }
  ccinclude { "InfString.h" }
  hinclude { "DETarget.h" }
  input {
    name { e_RESET }
    type { int }
  }
  input {
    name { e_OC_START }
    type { int }
  }
  input {
    name { e_TCLK }
    type { int }
  }
  output {
    name { e_OC_END }
    type { int }
  }
  public {
    double now;
    int resourceId;
    double clkFreq;
    /* Input event flags */
    int e_RESET_flag;
    int e_OC_START_flag;
    int e_TCLK_flag;
    /* Settable constants */
    int c_MAXCOUNT;
    int c_N;
    int c_SIMSCALE;
    /* State and output variables */
    int _st;
    int COUNT_OC;
  }
  private {
    double *_time, totnf;
    double _nextf, _latency;
    FILE *fpfire;
    FILE *fpover;
    /* Variables for behavioural simulation */
    double nextf;
    double ckfr;
    int ncycle;
  }
  state {
    name { implem }
    type { string }
    default { "{implem}" }
    desc { Valid values: HW SW BEHAV }
  }
  state {
    name { resourceName }
    type { string }
    default {"Resource1"}
    desc { resource to be used for simulation }
  }
  state {
    name { Clock_freq }
    type { float }
    default{ Clock_freq }
    desc { CPU Clock frequency }
  }
  state {
    name { policy }
    type { string }
    default { "{SCHEDULER}" }
    desc { Scheduling Policy to be used for simulation }
  }
  state {
    name { Priority }
    type { int }
    default { 0 }
    desc { Priority: 0 -> lower }
  }
  state {
    name { firename }
    type { string }
    default { "{Firingfile}" }
    desc { File to write firing information }
    attributes { A_NONSETTABLE }
  }
  state {
    name { overname }
    type { string }
    default { "{Overflowfile}" }
    desc { File to write overflow information }
    attributes { A_NONSETTABLE }
  }
  state {
    name { MAXCOUNT }
    type { int }
    default { "MAXCOUNT" }
    desc { Maximum value of the FRC counter }
  }
  state {
    name { N }
    type { int }
    default { "1" }
    desc { Parameter N (range 1 to 5) identifies one of the 5 output compares }
  }
  state {
    name { SIMSCALE }
    type { int }
    default { "1" }
    desc { Scale down simulation (only for HW and SW) }
  }


  constructor {
    delayType = TRUE;
    _delay = 0.0;
  }
  method {
    name { emitEventToIntQ }
    access { public }
    type { int }
    arglist { "( int outputPort, double delay )" }
    code {
      static const int ce_OC_END = 1;
      double emitTime;
      OutDEPort* tl;
      if (needsSharedResource) {
           StarLLCell *temp = (StarLLCell*)storeList->getAndRemove();
           if (temp == 0) temp = new StarLLCell(-1, -1);
           emitTime = now + (delay/clkFreq);
           temp->time = emitTime;
           temp->outputPort = outputPort;
           emittedEvents->append(temp);
      } else {
           if (now < 0) emitTime = 0.0;
           else emitTime = now + (_latency/clkFreq);
           switch(outputPort) {
             case ce_OC_END:
                tl = & e_OC_END;
                tl->put(emitTime) << 1;
                break;
            default:
               break;
         }
      }
      return 1;
    } 
  } 
  method {
    name { emitEvent }
    access { public }
    type { void }
    arglist { "(StarLLCell* cell, double emitTime)" }
    code {
      static const int ce_OC_END = 1;
      OutDEPort * tl;
      switch( cell->outputPort ) {
            case ce_OC_END:
                tl = & e_OC_END;
                tl->put(emitTime) << 1;
                break;
            default:
               break;
      }
      storeList->append(cell);
    }
  }
  method {
    name { getDelay }
    access { public }
    type { double }
    arglist { "()" }
    code {
      // need to add 1 clock cycle for SW stars so that dummy event is seen
      if (needsSharedResource) return (_delay)/clkFreq;// FIXME: a hack
      else return (totnf);
    }
  }

  method {
    name { CleanUp }
    access { public }
    type { void }
    arglist { "( int emitted )" }
    code {
      if ( emitted ) {
        e_RESET_flag = 0;
        e_OC_START_flag = 0;
        e_TCLK_flag = 0;
      }
    }
  }
  code {
#define DO_ESTIMATE  (do_estimate=1)
#define DO_NOT_ESTIMATE  (do_estimate=0)
#define illegal_cpu() 0
#define FUZZY(n,d) {if ((oc_prog_rel_Star->needsSharedResource)&&(do_estimate)) { \
                   _delay += _cache_delay[n]?_cache_delay[n]:_cache_delay[n]=oc_prog_rel_Star->FuzzyDelay(d); \
                 }}
#define ENERGY(e)
#define DELAY(d) {if ((oc_prog_rel_Star->needsSharedResource)&&(do_estimate)) {_delay += (d);}}
/* Input event macros */
#define detect_e_RESET_to_z_oc_prog_rel_0 oc_prog_rel_Star->e_RESET_flag
#define detect_e_OC_START_to_z_oc_prog_rel_0 oc_prog_rel_Star->e_OC_START_flag
#define detect_e_TCLK_to_z_oc_prog_rel_0 oc_prog_rel_Star->e_TCLK_flag
/* Output event macros */
#define emit_e_OC_END() oc_prog_rel_Star->emitEventToIntQ( e_OC_END, _delay )
#define startup( p )
#define cleanup( p, e ) oc_prog_rel_Star->CleanUp( e )
#define always_cleanup( p ) oc_prog_rel_Star->CleanUp( 1 )
#define cpu oc_prog_rel_Star->resourceId
#define SET_LT(a,b) oc_prog_rel_Star->Set_LT((a),(b))
/* Settable constant macros */
#define v_MAXCOUNT oc_prog_rel_Star->c_MAXCOUNT
#define v_N oc_prog_rel_Star->c_N
#define v_SIMSCALE oc_prog_rel_Star->c_SIMSCALE
/* State and output variable macros */
#define v__st oc_prog_rel_Star->_st
#define v_COUNT_OC oc_prog_rel_Star->COUNT_OC
    DEoc_prog_rel * oc_prog_rel_Star = NULL;
    static int do_estimate = 1;
    static double _delay = 0.0;
    static double _cache_delay[26];
    /* Output event constants */
    static const int e_OC_END = 1;
    /* Input variable declarations */
    static int v_OC_START;
    static int v_TCLK;
    /* Output variable declarations */

    static unsigned int path_id;
#include "z_oc_prog_rel_0.c"
  }
  setup {
    DETarget *t;
    char stemp[1024];

    t = (DETarget *) target();
    if (!(t->isResource())) {
      strcpy( stemp, "Please, set target to Resource Contention Scheduler" );
      Error::abortRun( *this, stemp );
    }
    timeOfArrival = -1;
    _nextf = 1.0;
    _latency = 1.0;
    DERCStar::Closeflow();
    strcpy(resource, resourceName);
    strcpy(stemp, implem);
    needsSharedResource = getImplem( stemp );
    if ( needsSharedResource < 2 ) {
    }
  }
  begin {
    char stemp[ 1024 ];
    InfString name;
    int saveimpl;

    if ( needsSharedResource < 2 ) {
      priority = Priority;
      oc_prog_rel_Star = this;
      /* set variables which reflect Star parameters*/ 
      strcpy( stemp, policy); 
      schedPolicy = resourcePointer->getSchedPolicy( stemp );
      sprintf( stemp, "%d", priority );
      lastFiredPr = resourcePointer->registerPriority( stemp );
      clkFreq = Clock_freq;
      clkFreq = (1000000.0)*clkFreq;
      resourcePointer->registerClock( clkFreq );
      totnf = 1.0/clkFreq;
      emittedEvents = new SequentialList();
      storeList = new SequentialList();
      _delay = 0.0;
      int i;
      for ( i=0; i<=25; i++ ) {
         _cache_delay[i] = 0;
      }
      /* State and output variables initialization */
      _st = 0;
      /* Settable constants initialization */
      c_MAXCOUNT = MAXCOUNT;
      c_N = N;
      c_SIMSCALE = SIMSCALE;
      /* Input event flag initialization */
      e_RESET_flag = 0;
      e_OC_START_flag = 0;
      e_TCLK_flag = 0;
      if ( needsSharedResource ) {
        strcpy( stemp, resource );
        if (strchr(stemp, '.')) *(strchr(stemp,'.')) = '\0';
        if (( resourceId = (int) ReadCostParam( stemp )) == -1 ) {
          strcat( stemp, ": not a valid CPU" );
          Error::abortRun( *this, stemp );
        }
      }
      strcpy( stemp, firename );
      if (strcmp(stemp,"{Firingfile}")) fpfire = DERCStar::Openfiring(stemp);
      strcpy( stemp, overname );
      if (strcmp(stemp,"{Overflowfile}")) fpover = DERCStar::Openoverflow(stemp);
      now = -1.0;
      saveimpl = needsSharedResource;
      needsSharedResource = 0;
      _t_z_oc_prog_rel_0(0,0);

      needsSharedResource = saveimpl;
    } else if ( needsSharedResource == 2 ) {
      nextf = 0;
      ncycle = 0;
      ckfr = (double) Clock_freq;
    } else {
      strcpy( stemp, "Implementation not supported!" );
      Error::abortRun( *this, stemp );
    }
  }
  go {
    char stemp[ 1024 ];
    int runflag = 0;
    StringList name;
    int refire;

    now = arrivalTime;
    if ( needsSharedResource < 2 ) {
      name = Block::fullName();
      oc_prog_rel_Star = this;
      if (feedbackIn->dataNew && (int)((*feedbackIn)%0) == 0) {
          DERCStar* starUsingResource = resourcePointer->getTopCell()->star;
          if (((StarLLCell*)emittedEvents->head())->time != now ) {
              refireAtTime(((StarLLCell*)emittedEvents->head())->time, 0.0);
          } else {
              if ( starUsingResource != this ) {
                  Error::abortRun( *this, "Star should not reach this point if preempted!" );
              }
              if(emittedEvents->head() == emittedEvents->tail() ) {
                  // This is the dummy event indicating end of resource usage 
                  resourcePointer->removeFinishedStar(this); 
                  emittedEvents->getAndRemove(); 
                  assert(emittedEvents->getAndRemove() == 0); 
                  if ( fpfire ) {
                      sprintf( stemp, "%s: %f %d end %s\n", (const char*) name, now, priority, resource );
                      DERCStar::Printfiring( stemp );
                  }
              } else { 
                  StarLLCell *temp; 
                  while((temp = (StarLLCell*)emittedEvents->head())->time == now && (emittedEvents->head() != emittedEvents->tail()) ) { 
                      StarLLCell *temp = (StarLLCell*)emittedEvents->getAndRemove(); 
                      assert(now == temp->time); 
                      emitEvent(temp, now); 
                  } 
                  refireAtTime(((StarLLCell*)emittedEvents->head())->time, 0.0); 
              }
              feedbackIn->dataNew = FALSE;
          }
      } else { // This star is firing for the first time, and has resource if req 
          if ( e_RESET.dataNew ) {
            e_RESET_flag = 1;
            e_RESET.dataNew = FALSE;
            runflag = 1;
          }
          if ( e_OC_START.dataNew ) {
            e_OC_START_flag = 1;
            e_OC_START.dataNew = FALSE;
            runflag = 1;
          }
          if ( e_TCLK.dataNew ) {
            e_TCLK_flag = 1;
            e_TCLK.dataNew = FALSE;
            runflag = 1;
          }

        // Read the input values 
        v_OC_START = (e_OC_START)%0;
        v_TCLK = (e_TCLK)%0;
        _delay = 0.0;
        if (runflag == 1) {
          _t_z_oc_prog_rel_0(0,0);

          // write to firing file 
          if ( fpfire ) {
              sprintf( stemp, "%s: %f %d start %s\n", (const char*) name, now, priority, resource );
              DERCStar::Printfiring( stemp );
          }
          if (needsSharedResource) emitEventToIntQ( 0, _delay );// emit dummy event 
          timeOfArrival = -1;// reset this as have finished this firing 
          if (needsSharedResource) refireAtTime(((StarLLCell*)emittedEvents->head())->time, 0.0); 
        }
      }
    } else if ( needsSharedResource == 2 ) {    
      refire = ( canGetFired() ? 1 : 0 );
      if ( e_RESET.dataNew ) {
        nextf = 0.0;
        while ( e_RESET.dataNew ) e_RESET.getSimulEvent();
      }
      if ( e_TCLK.dataNew ) {
        ncycle = (e_TCLK)%0;
        while ( e_TCLK.dataNew ) e_TCLK.getSimulEvent();
      }
      if ( e_OC_START.dataNew ) {
        if ( ncycle > 0 && ((e_OC_START)%0 % MAXCOUNT) > 0 ) {
          nextf = ((e_OC_START)%0 % MAXCOUNT ) * ncycle / (ckfr*1000000.0) + now;
          refireAtTime( nextf );
        } else {
          nextf = 0.0;
        }
        while ( e_OC_START.dataNew ) e_OC_START.getSimulEvent();
      }
      if ( refire ) {
        refire = 0;
        if ( nextf && nextf == now ) {
          e_OC_END.put( nextf ) << (int) 1;
          nextf = 0.0;
        }
      }
    }
  }
}
