defstar {
  name { pwm }
  domain { DE }
  derivedfrom { Polis }
  desc {
    PWM driver
  }
  hinclude { "uC_types.h" }
  hinclude { "swlib.h" }
  ccinclude { "InfString.h" }
  hinclude { "DETarget.h" }
  input {
    name { e_PWMCLK }
    type { int }
  }
  input {
    name { e_PWM_DC }
    type { int }
  }
  output {
    name { e_PWM_OUT }
    type { int }
  }
  public {
    double now;
    int resourceId;
    double clkFreq;
    /* Input event flags */
    int e_PWMCLK_flag;
    int e_PWM_DC_flag;
    /* Settable constants */
    int c_PWMMAXCOUNT;
    int c_SIMSCALE;
    int c_PWM_POLARITY;
    int c_N;
    /* State and output variables */
    int _st;
    int value;
  }
  private {
    double *_time, totnf;
    double _nextf, _latency;
    FILE *fpfire;
    FILE *fpover;
    /* Variables for behavioural simulation */
    double ckfr;
    int pwmcount;
    int pwmint;
    int datarec;
    int newval;
    int datum;
    int pwmstart;
    int pwmpol;
    int npwmpol;
  }
  state {
    name { implem }
    type { string }
    default { "{implem}" }
    desc { Valid values: HW SW BEHAV }
  }
  state {
    name { resourceName }
    type { string }
    default {"Resource1"}
    desc { resource to be used for simulation }
  }
  state {
    name { Clock_freq }
    type { float }
    default{ Clock_freq }
    desc { CPU Clock frequency }
  }
  state {
    name { policy }
    type { string }
    default { "{SCHEDULER}" }
    desc { Scheduling Policy to be used for simulation }
  }
  state {
    name { Priority }
    type { int }
    default { 0 }
    desc { Priority: 0 -> lower }
  }
  state {
    name { firename }
    type { string }
    default { "{Firingfile}" }
    desc { File to write firing information }
    attributes { A_NONSETTABLE }
  }
  state {
    name { overname }
    type { string }
    default { "{Overflowfile}" }
    desc { File to write overflow information }
    attributes { A_NONSETTABLE }
  }
  state {
    name { PWMMAXCOUNT }
    type { int }
    default { "PWMMAXCOUNT" }
    desc { Number of clock cycles for each PWM period }
  }
  state {
    name { SIMSCALE }
    type { int }
    default { "1" }
    desc { Scale down simulation (only for HW and SW) }
  }
  state {
    name { PWM_POLARITY }
    type { int }
    default { "1" }
    desc { Logical value of the active part of the PWM signal }
  }
  state {
    name { N }
    type { int }
    default { "0" }
    desc { Parameter N (range 0 to 7) identifies one of the 8 PWM generators }
  }


  constructor {
    delayType = TRUE;
    _delay = 0.0;
  }
  method {
    name { emitEventToIntQ }
    access { public }
    type { int }
    arglist { "( int outputPort, double delay )" }
    code {
      static const int ce_PWM_OUT = 1;
      double emitTime;
      OutDEPort* tl;
      if (needsSharedResource) {
           StarLLCell *temp = (StarLLCell*)storeList->getAndRemove();
           if (temp == 0) temp = new StarLLCell(-1, -1);
           emitTime = now + (delay/clkFreq);
           temp->time = emitTime;
           temp->outputPort = outputPort;
           emittedEvents->append(temp);
      } else {
           if (now < 0) emitTime = 0.0;
           else emitTime = now + (_latency/clkFreq);
           switch(outputPort) {
             case ce_PWM_OUT:
                tl = & e_PWM_OUT;
                tl->put(emitTime) << v_PWM_OUT;
                break;
            default:
               break;
         }
      }
      return 1;
    } 
  } 
  method {
    name { emitEvent }
    access { public }
    type { void }
    arglist { "(StarLLCell* cell, double emitTime)" }
    code {
      static const int ce_PWM_OUT = 1;
      OutDEPort * tl;
      switch( cell->outputPort ) {
            case ce_PWM_OUT:
                tl = & e_PWM_OUT;
                tl->put(emitTime) << v_PWM_OUT;
                break;
            default:
               break;
      }
      storeList->append(cell);
    }
  }
  method {
    name { getDelay }
    access { public }
    type { double }
    arglist { "()" }
    code {
      // need to add 1 clock cycle for SW stars so that dummy event is seen
      if (needsSharedResource) return (_delay)/clkFreq;// FIXME: a hack
      else return (totnf);
    }
  }
  method {
    name { CleanUp }
    access { public }
    type { void }
    arglist { "( int emitted )" }
    code {
      if ( emitted ) {
        e_PWMCLK_flag = 0;
        e_PWM_DC_flag = 0;
      }
    }
  }
  code {
#define DO_ESTIMATE  (do_estimate=1)
#define DO_NOT_ESTIMATE  (do_estimate=0)
#define illegal_cpu() 0
#define FUZZY(n,d) {if ((pwm_Star->needsSharedResource)&&(do_estimate)) { \
                     _delay += _cache_delay[n]?_cache_delay[n]:_cache_delay[n]=pwm_Star->FuzzyDelay(d); \
                 }}
#define ENERGY(e)
#define DELAY(d) {if ((pwm_Star->needsSharedResource)&&(do_estimate)) {_delay += (d);}}
/* Input event macros */
#define detect_e_PWMCLK_to_z_pwm_0 pwm_Star->e_PWMCLK_flag
#define detect_e_PWM_DC_to_z_pwm_0 pwm_Star->e_PWM_DC_flag
/* Output event macros */
#define emit_e_PWM_OUT() pwm_Star->emitEventToIntQ( e_PWM_OUT, _delay )
#define startup( p )
#define cleanup( p, e ) pwm_Star->CleanUp( e )
#define always_cleanup( p ) pwm_Star->CleanUp( 1 )
#define cpu pwm_Star->resourceId
#define SET_LT(a,b) pwm_Star->Set_LT((a),(b))
/* Settable constant macros */
#define v_PWMMAXCOUNT pwm_Star->c_PWMMAXCOUNT
#define v_SIMSCALE pwm_Star->c_SIMSCALE
#define v_PWM_POLARITY pwm_Star->c_PWM_POLARITY
#define v_N pwm_Star->c_N
/* State and output variable macros */
#define v__st pwm_Star->_st
#define v_value pwm_Star->value
    DEpwm * pwm_Star = NULL;
    static int do_estimate = 1;
    static double _delay = 0.0;
    static double _cache_delay[24];
    /* Output event constants */
    static const int e_PWM_OUT = 1;
    /* Input variable declarations */
    static int v_PWMCLK;
    static int v_PWM_DC;
    /* Output variable declarations */
    static int v_PWM_OUT;

    static unsigned int path_id;
#include "z_pwm_0.c"
  }
  setup {
    DETarget *t;
    char stemp[1024];

    t = (DETarget *) target();
    if (!(t->isResource())) {
      strcpy( stemp, "Please, set target to Resource Contention Scheduler" );
      Error::abortRun( *this, stemp );
    }
    timeOfArrival = -1;
    _nextf = 1.0;
    _latency = 1.0;
    DERCStar::Closeflow();
    strcpy(resource, resourceName);
    strcpy(stemp, implem);
    needsSharedResource = getImplem( stemp );
    if ( needsSharedResource < 2 ) {
    }
  }
  begin {
    char stemp[ 1024 ];
    InfString name;
    int saveimpl;

    if ( needsSharedResource < 2 ) {
      priority = Priority;
      pwm_Star = this;
      /* set variables which reflect Star parameters*/
      strcpy( stemp, policy);
      schedPolicy = resourcePointer->getSchedPolicy( stemp );
      sprintf( stemp, "%d", priority );
      lastFiredPr = resourcePointer->registerPriority( stemp );
      clkFreq = Clock_freq;
      clkFreq = (1000000.0)*clkFreq;
      resourcePointer->registerClock( clkFreq );
      totnf = 1.0/clkFreq;
      emittedEvents = new SequentialList();
      storeList = new SequentialList();
      _delay = 0.0;
      int i;
      for ( i=0; i<=23; i++ ) {
         _cache_delay[i] = 0;
      }
      /* State and output variables initialization */
      _st = 0;
      /* Settable constants initialization */
      c_PWMMAXCOUNT = PWMMAXCOUNT;
      c_SIMSCALE = SIMSCALE;
      c_PWM_POLARITY = PWM_POLARITY;
      c_N = N;
      /* Input event flag initialization */
      e_PWMCLK_flag = 0;
      e_PWM_DC_flag = 0;
      if ( needsSharedResource ) {
        strcpy( stemp, resource );
        if (strchr(stemp, '.')) *(strchr(stemp,'.')) = '\0';
        if (( resourceId = (int) ReadCostParam( stemp )) == -1 ) {
          strcat( stemp, ": not a valid CPU" );
          Error::abortRun( *this, stemp );
        }
      }
      strcpy( stemp, firename );
      if (strcmp(stemp,"{Firingfile}")) fpfire = DERCStar::Openfiring(stemp);
      strcpy( stemp, overname );
      if (strcmp(stemp,"{Overflowfile}")) fpover = DERCStar::Openoverflow(stemp);
      now = -1.0;
      saveimpl = needsSharedResource;
      needsSharedResource = 0;
      _t_z_pwm_0(0,0);

      needsSharedResource = saveimpl;
    } else if ( needsSharedResource == 2 ) {
      ckfr = (double) Clock_freq;
      datarec = 0;
      pwmcount = 0;
      pwmint = 0;
      pwmstart = 0;
      pwmpol = ( PWM_POLARITY ? 1 : 0 );
      npwmpol = ( PWM_POLARITY ? 0 : 1 );
      e_PWM_OUT.put( 0 ) << pwmpol;
    } else {
      strcpy( stemp, "Implementation not supported!" );
      Error::abortRun( *this, stemp );
    }
  }
  go {
    char stemp[ 1024 ];
    int runflag = 0, refire;
    StringList name;

    now = arrivalTime;
    if (needsSharedResource < 2) {
      name = Block::fullName();
      pwm_Star = this;
      if (feedbackIn->dataNew && (int)((*feedbackIn)%0) == 0) {
          DERCStar* starUsingResource = resourcePointer->getTopCell()->star;
          if (((StarLLCell*)emittedEvents->head())->time != now ) {
              refireAtTime(((StarLLCell*)emittedEvents->head())->time, 0.0);
          } else {
              if ( starUsingResource != this ) {
                  Error::abortRun( *this, "Star should not reach this point if preempted!" );
              }
              if(emittedEvents->head() == emittedEvents->tail() ) {
                  // This is the dummy event indicating end of resource usage 
                  resourcePointer->removeFinishedStar(this); 
                  emittedEvents->getAndRemove(); 
                  assert(emittedEvents->getAndRemove() == 0); 
                  if ( fpfire ) {
                      sprintf( stemp, "%s: %f %d end %s\n", (const char*) name, now, priority, resource );
                      DERCStar::Printfiring( stemp );
                  }
              } else { 
                  StarLLCell *temp; 
                  while((temp = (StarLLCell*)emittedEvents->head())->time == now && (emittedEvents->head() != emittedEvents->tail()) ) { 
                      StarLLCell *temp = (StarLLCell*)emittedEvents->getAndRemove(); 
                      assert(now == temp->time); 
                      emitEvent(temp, now); 
                  } 
                  refireAtTime(((StarLLCell*)emittedEvents->head())->time, 0.0); 
              }
              feedbackIn->dataNew = FALSE;
          }
      } else { // This star is firing for the first time, and has resource if req 
          if ( e_PWMCLK.dataNew ) {
            e_PWMCLK_flag = 1;
            e_PWMCLK.dataNew = FALSE;
            runflag = 1;
          }
          if ( e_PWM_DC.dataNew ) {
            e_PWM_DC_flag = 1;
            e_PWM_DC.dataNew = FALSE;
            runflag = 1;
          }

        // Read the input values 
        v_PWMCLK = (e_PWMCLK)%0;
        v_PWM_DC = (e_PWM_DC)%0;
        _delay = 0.0;
        if (runflag == 1) {
          _t_z_pwm_0(0,0);

          // write to firing file 
          if ( fpfire ) {
              sprintf( stemp, "%s: %f %d start %s\n", (const char*) name, now, priority, resource );
              DERCStar::Printfiring( stemp );
          }
          if (needsSharedResource) emitEventToIntQ( 0, _delay );// emit dummy event 
          timeOfArrival = -1;// reset this as have finished this firing 
          if (needsSharedResource) refireAtTime(((StarLLCell*)emittedEvents->head())->time, 0.0);
        }
      }
    } else if (needsSharedResource == 2 ) {
      refire = ( canGetFired() ? 1 : 0 );
      if ( e_PWM_DC.dataNew ) {
        datarec = 1;
        newval = (e_PWM_DC)%0 % (int) PWMMAXCOUNT;
        while ( e_PWM_DC.dataNew ) e_PWM_DC.getSimulEvent();
      }
      if ( e_PWMCLK.dataNew ) {
        pwmint = (e_PWMCLK)%0;
        while ( e_PWMCLK.dataNew ) e_PWMCLK.getSimulEvent();
      }
      if ( datarec && pwmint ) {
        if ( refire ) {
          if ( pwmcount == 0 ) {
            datum = newval;
            if ( datum == 0 ) {
              e_PWM_OUT.put( now + 1 / (ckfr*1000000.0)) << npwmpol;
              refireAtTime( now + ((double)( pwmint * (double) PWMMAXCOUNT )) / (ckfr*1000000.0));
            } else {
              e_PWM_OUT.put( now + 1 / (ckfr*1000000.0)) << pwmpol;
              pwmcount = datum;
              refireAtTime( now + ((double) (pwmcount * pwmint)) / (ckfr*1000000.0));
            }
          } else {
            e_PWM_OUT.put( now + 1 / (ckfr*1000000.0)) << npwmpol;
            pwmcount = 0;
            refireAtTime( now + (double) (((double) PWMMAXCOUNT - datum ) * pwmint) / (ckfr*1000000));
          }
        } else if ( !pwmstart ) {
          pwmstart = 1;
          pwmcount = 0;
          refireAtTime( now + ((double) pwmint) / (ckfr*1000000.0));
        }
      }
    }
  }
}
