defstar {
  name { pwmfrc }
  domain { DE }
  derivedfrom { Polis }
  desc { 
    Free Running Counter for PWM
  }
  hinclude { "uC_types.h" }
  hinclude { "swlib.h" }
  ccinclude { "InfString.h" }
  hinclude { "DETarget.h" }
  input {
    name { e_ECLK }
    type { float }
  }
  output {
    name { e_PWMCLK }
    type { int }
  }
  public {
    double now;
    int resourceId;
    double clkFreq;
    /* Input event flags */
    int e_ECLK_flag;
    /* Settable constants */
    int c_PWMMAXCOUNT;
    int c_SIMSCALE;
    int c_PWMPRESCALER;
    /* State and output variables */
    int _st;
    int COUNT;
    int SCALE;
  }
  private {
    double *_time, totnf;
    double _nextf, _latency;
    FILE *fpfire;
    FILE *fpover;
    /* Variables for behavioural simulation */
    int emitted;
    int scaler;
    double valuein;
  }
  state {
    name { implem }
    type { string }
    default { "{implem}" }
    desc { Valid values: HW SW BEHAV }
  }
  state {
    name { resourceName }
    type { string }
    default {"Resource1"}
    desc { resource to be used for simulation }
  }
  state {
    name { Clock_freq }
    type { float }
    default{ Clock_freq }
    desc { CPU Clock frequency }
  }
  state {
    name { policy }
    type { string }
    default { "{SCHEDULER}" }
    desc { Scheduling Policy to be used for simulation }
  }
  state {
    name { Priority }
    type { int }
    default { 0 }
    desc { Priority: 0 -> lower }
  }
  state {
    name { firename }
    type { string }
    default { "{Firingfile}" }
    desc { File to write firing information }
    attributes { A_NONSETTABLE }
  }
  state {
    name { overname }
    type { string }
    default { "{Overflowfile}" }
    desc { File to write overflow information }
    attributes { A_NONSETTABLE }
  }
  state {
    name { PWMMAXCOUNT }
    type { int }
    default { "PWMMAXCOUNT" }
    desc { Number of clock cycles for each PWM period }
  }
  state {
    name { SIMSCALE }
    type { int }
    default { "1" }
    desc { Scale down simulation (only for HW and SW) }
  }
  state {
    name { PWMPRESCALER }
    type { int }
    default { "1" }
    desc { Prescaler of the PWM FRC }
  }


  constructor {
    delayType = TRUE;
    _delay = 0.0;
  }
  method {
    name { emitEventToIntQ }
    access { public }
    type { int }
    arglist { "( int outputPort, double delay )" }
    code {
      static const int ce_PWMCLK = 1;
      double emitTime;
      OutDEPort* tl;
      if (needsSharedResource) {
           StarLLCell *temp = (StarLLCell*)storeList->getAndRemove();
           if (temp == 0) temp = new StarLLCell(-1, -1);
           emitTime = now + (delay/clkFreq);
           temp->time = emitTime;
           temp->outputPort = outputPort;
           emittedEvents->append(temp);
      } else {
           if (now < 0) emitTime = 0.0;
           else emitTime = now + (_latency/clkFreq);
           switch(outputPort) {
             case ce_PWMCLK:
                tl = & e_PWMCLK;
                tl->put(emitTime) << v_PWMCLK;
                break;
            default:
               break;
         }
      }
      return 1;
    } 
  } 
  method {
    name { emitEvent }
    access { public }
    type { void }
    arglist { "(StarLLCell* cell, double emitTime)" }
    code {
      static const int ce_PWMCLK = 1;
      OutDEPort * tl;
      switch( cell->outputPort ) {
            case ce_PWMCLK:
                tl = & e_PWMCLK;
                tl->put(emitTime) << v_PWMCLK;
                break;
            default:
               break;
      }
      storeList->append(cell);
    }
  }
  method {
    name { getDelay }
    access { public }
    type { double }
    arglist { "()" }
    code {
      // need to add 1 clock cycle for SW stars so that dummy event is seen
      if (needsSharedResource) return (_delay)/clkFreq;// FIXME: a hack
      else return (totnf);
    }
  }
  method {
    name { CleanUp }
    access { public }
    type { void }
    arglist { "( int emitted )" }
    code {
      if ( emitted ) {
        e_ECLK_flag = 0;
      }
    }
  }
  code {
#define DO_ESTIMATE  (do_estimate=1)
#define DO_NOT_ESTIMATE  (do_estimate=0)
#define illegal_cpu() 0
#define FUZZY(n,d) {if ((pwmfrc_Star->needsSharedResource)&&(do_estimate)) { \
                     _delay += _cache_delay[n]?_cache_delay[n]:_cache_delay[n]=pwmfrc_Star->FuzzyDelay(d); \
                 }}
#define ENERGY(e)
#define DELAY(d) {if ((pwmfrc_Star->needsSharedResource)&&(do_estimate)) {_delay += (d);}}
/* Input event macros */
#define detect_e_ECLK_to_z_pwmfrc_0 pwmfrc_Star->e_ECLK_flag
/* Output event macros */
#define emit_e_PWMCLK() pwmfrc_Star->emitEventToIntQ( e_PWMCLK, _delay )
#define startup( p )
#define cleanup( p, e ) pwmfrc_Star->CleanUp( e )
#define always_cleanup( p ) pwmfrc_Star->CleanUp( 1 )
#define cpu pwmfrc_Star->resourceId
#define SET_LT(a,b) pwmfrc_Star->Set_LT((a),(b))
/* Settable constant macros */
#define v_PWMMAXCOUNT pwmfrc_Star->c_PWMMAXCOUNT
#define v_SIMSCALE pwmfrc_Star->c_SIMSCALE
#define v_PWMPRESCALER pwmfrc_Star->c_PWMPRESCALER
/* State and output variable macros */
#define v__st pwmfrc_Star->_st
#define v_COUNT pwmfrc_Star->COUNT
#define v_SCALE pwmfrc_Star->SCALE
    DEpwmfrc * pwmfrc_Star = NULL;
    static int do_estimate = 1;
    static double _delay = 0.0;
    static double _cache_delay[14];
    /* Output event constants */
    static const int e_PWMCLK = 1;
    /* Input variable declarations */
    /* Output variable declarations */
    static int v_PWMCLK;

    static unsigned int path_id;
#include "z_pwmfrc_0.c"
  }
  setup {
    DETarget *t;
    char stemp[1024];

    t = (DETarget *) target();
    if (!(t->isResource())) {
      strcpy( stemp, "Please, set target to Resource Contention Scheduler" );
      Error::abortRun( *this, stemp );
    }
    timeOfArrival = -1;
    _nextf = 1.0;
    _latency = 1.0;
    DERCStar::Closeflow();
    strcpy(resource, resourceName);
    strcpy(stemp, implem);
    needsSharedResource = getImplem( stemp );
    if ( needsSharedResource < 2 ) {
    }
  }
  begin {
    char stemp[ 1024 ];
    InfString name;
    int saveimpl;

    if ( needsSharedResource < 2 ) {
      priority = Priority;
      pwmfrc_Star = this;
      /* set variables which reflect Star parameters*/
      strcpy( stemp, policy);
      schedPolicy = resourcePointer->getSchedPolicy( stemp );
      sprintf( stemp, "%d", priority );
      lastFiredPr = resourcePointer->registerPriority( stemp );
      clkFreq = Clock_freq;
      clkFreq = (1000000.0)*clkFreq;
      resourcePointer->registerClock( clkFreq );
      totnf = 1.0/clkFreq;
      emittedEvents = new SequentialList();
      storeList = new SequentialList();
      _delay = 0.0;
      int i;
      for ( i=0; i<=13; i++ ) {
         _cache_delay[i] = 0;
      }
      /* State and output variables initialization */
      _st = 0;
      /* Settable constants initialization */
      c_PWMMAXCOUNT = PWMMAXCOUNT;
      c_SIMSCALE = SIMSCALE;
      c_PWMPRESCALER = PWMPRESCALER;
      /* Input event flag initialization */
      e_ECLK_flag = 0;
      if ( needsSharedResource ) {
        strcpy( stemp, resource );
        if (strchr(stemp, '.')) *(strchr(stemp,'.')) = '\0';
        if (( resourceId = (int) ReadCostParam( stemp )) == -1 ) {
          strcat( stemp, ": not a valid CPU" );
          Error::abortRun( *this, stemp );
        }
      }
      strcpy( stemp, firename );
      if (strcmp(stemp,"{Firingfile}")) fpfire = DERCStar::Openfiring(stemp);
      strcpy( stemp, overname );
      if (strcmp(stemp,"{Overflowfile}")) fpover = DERCStar::Openoverflow(stemp);
      now = -1.0;
      saveimpl = needsSharedResource;
      needsSharedResource = 0;
      _t_z_pwmfrc_0(0,0);

      needsSharedResource = saveimpl;
    } else if ( needsSharedResource == 2 ) {
      emitted = 0;
    } else {
      strcpy( stemp, "Implementation not supported!" );
      Error::abortRun( *this, stemp );
    }
  }
  go {
    char stemp[ 1024 ];
    int runflag = 0;
    StringList name;

    now = arrivalTime;
    if ( needsSharedResource < 2 ) {
      name = Block::fullName();
      pwmfrc_Star = this;
      if (feedbackIn->dataNew && (int)((*feedbackIn)%0) == 0) {
          DERCStar* starUsingResource = resourcePointer->getTopCell()->star;
          if (((StarLLCell*)emittedEvents->head())->time != now ) {
              refireAtTime(((StarLLCell*)emittedEvents->head())->time, 0.0);
          } else {
              if ( starUsingResource != this ) {
                  Error::abortRun( *this, "Star should not reach this point if preempted!" );
              }
              if(emittedEvents->head() == emittedEvents->tail() ) {
                  // This is the dummy event indicating end of resource usage 
                  resourcePointer->removeFinishedStar(this); 
                  emittedEvents->getAndRemove(); 
                  assert(emittedEvents->getAndRemove() == 0); 
                  if ( fpfire ) {
                      sprintf( stemp, "%s: %f %d end %s\n", (const char*) name, now, priority, resource );
                      DERCStar::Printfiring( stemp );
                  }
              } else { 
                  StarLLCell *temp; 
                  while((temp = (StarLLCell*)emittedEvents->head())->time == now && (emittedEvents->head() != emittedEvents->tail()) ) { 
                      StarLLCell *temp = (StarLLCell*)emittedEvents->getAndRemove(); 
                      assert(now == temp->time); 
                      emitEvent(temp, now); 
                  } 
                  refireAtTime(((StarLLCell*)emittedEvents->head())->time, 0.0); 
              }
              feedbackIn->dataNew = FALSE;
          }
      } else { // This star is firing for the first time, and has resource if req 
          if ( e_ECLK.dataNew ) {
            e_ECLK_flag = 1;
            e_ECLK.dataNew = FALSE;
            runflag = 1;
          }

        // Read the input values 
        _delay = 0.0;
        if (runflag == 1) {
          _t_z_pwmfrc_0(0,0);

          // write to firing file 
          if ( fpfire ) {
              sprintf( stemp, "%s: %f %d start %s\n", (const char*) name, now, priority, resource );
              DERCStar::Printfiring( stemp );
          }
          if (needsSharedResource) emitEventToIntQ( 0, _delay );// emit dummy event 
          timeOfArrival = -1;// reset this as have finished this firing 
          if (needsSharedResource) refireAtTime(((StarLLCell*)emittedEvents->head())->time, 0.0);
        }
      }
    } else if ( needsSharedResource == 2 ) {
      if ( !emitted ) {
        valuein = (e_ECLK)%0;
        scaler = ( PWMPRESCALER / 2 ) + 1;
        e_PWMCLK.put(now) << (int) (( valuein * 1000.0 * Clock_freq ) * (int) scaler);
        emitted = 1;
      }
      while ( e_ECLK.dataNew ) e_ECLK.getSimulEvent();
    }
  }
}
