defstar {
  name { Polis }
  domain { DE }
  derivedfrom { RCStar }
  desc { Common Polis Software Scheduler }
  hinclude { "XYPlot.h" }
  hinclude { "time.h" }
  hinclude { "polis_sim.h" }
  hinclude { "cache.h" }

  protected {
    static int  counter;
    static int  priority_pol;
    static int  priority_next;
    static double  next_curr;
    static double  next_glob;
    static int  star_ack;
    static int  star_run;
    static FILE *fpoverflow;
    static FILE *fpfiring;

    static FILE *fromCommand;
    static FILE *toCommand;
    static FILE *breakname;
    static FILE *isslog;
    static FILE *fptraces;
    static int iss_pid;
    static int break_inserted;
    static int prev_cycles;
    static int ptl_cycles;
    static double prev_energy;
    static double ptl_energy;
    static int delay;
    static double energy;
    static double MinTime;
    static double MaxTime;
    static double MinEnergy;
    static double MaxEnergy;
    static double xMinPower;
    static double xMaxPower;
    static double MinPower;
    static double MaxPower;
    static int xyplotEnergy_created;
    static XYPlot xyplotEnergy;
    static int xyplotPower_created;
    static XYPlot xyplotPower;
    static int num_break_inserted;
    static clock_t simCpuTime;     // in order to measure elapsed delays
    static int cache_initialized;
  }
  code {
    #ifdef BUFSIZE
    #undef BUFSIZE
    #endif
    #define BUFSIZE 1024

    #ifndef __cplusplus
    #define __cplusplus
    #endif 

    #include "pipefork.h"
    #include <sys/wait.h>
    #include <unistd.h>
    #include "swlib.h"
#ifdef GCC_BASED_SIMUL
    #include "mips-vars.h"
#endif

    int DEPolis::counter;
    int DEPolis::priority_pol;
    int DEPolis::priority_next;
    double DEPolis::next_curr;
    double DEPolis::next_glob;
    int DEPolis::star_ack;
    int DEPolis::star_run;
    FILE * DEPolis::fpoverflow;
    FILE * DEPolis::fpfiring;

    FILE * DEPolis::toCommand;
    FILE * DEPolis::fromCommand;
    FILE * DEPolis::breakname;
    FILE * DEPolis::isslog;
    FILE * DEPolis::fptraces;
    int DEPolis::iss_pid;
    int DEPolis::break_inserted;
    int DEPolis::prev_cycles;
    int DEPolis::ptl_cycles;
    double DEPolis::prev_energy;
    double DEPolis::ptl_energy;
    int DEPolis::delay;
    double DEPolis::energy;
    double DEPolis::MinTime;
    double DEPolis::MaxTime;
    double DEPolis::MinEnergy;
    double DEPolis::MaxEnergy;
    double DEPolis::MinPower;
    double DEPolis::MaxPower;
    int DEPolis::xyplotEnergy_created;
    XYPlot DEPolis::xyplotEnergy;
    int DEPolis::xyplotPower_created;
    XYPlot DEPolis::xyplotPower;
    int DEPolis::num_break_inserted;
    clock_t DEPolis::simCpuTime;
    int DEPolis::cache_initialized;
  }

  method {
    name { resetAll }
    access { public }
    type { void }
    code {
      DEPolis::simCpuTime = clock();

      DEPolis::counter = 0;
      DEPolis::priority_pol = -1;
      DEPolis::priority_next = -1;
      DEPolis::next_curr = 0;
      DEPolis::next_glob = 0;
      DEPolis::star_ack = -1;
      DEPolis::star_run = -1;
    }
  }

  method {
    name { getImplem }
    access { public }
    type { "int" }
    arglist { "( char *stimpl )" }
    code {
      if ( !strcmp( stimpl, "HW" )) {
        return 0;
      } else if ( !strcmp( stimpl, "SW" )) {
        return 1;
      } else if ( !strcmp( stimpl, "BEHAV" )) {
        return 2;
      } else {
        char stemp[ 1024 ];
        strcpy( stemp, stimpl );
        strcat( stemp, ": not a valid implementation" );
        Error::abortRun( stemp );
      }
      return (-1);
    }
  }

  method {
    name { Openoverflow }
    access { public }
    type { "FILE *" }
    arglist { "( char *name )" }
    code {
      if ( DEPolis::fpoverflow == NULL ) {
        DEPolis::fpoverflow = fopen( name, "w" );
        if ( DEPolis::fpoverflow ) {
          fprintf( DEPolis::fpoverflow, "Overflow report\n\n" );
          fflush( DEPolis::fpoverflow );
        }
      }
      return( fpoverflow );
    }
  }

  method {
    name { Printoverflow }
    access { public }
    type { void }
    arglist { "( char *st )" }
    code {
      if ( DEPolis::fpoverflow ) {
        fprintf( DEPolis::fpoverflow, st );
        fflush( DEPolis::fpoverflow );
      }
    }
  }

  method {
    name { Openfiring }
    access { public }
    type { "FILE *" }
    arglist { "( char *name )" }
    code {
      if ( DEPolis::fpfiring == NULL ) {
        DEPolis::fpfiring = fopen( name, "w" );
        if ( DEPolis::fpfiring ) {
          fprintf( DEPolis::fpfiring, "Firing report\n\n" );
          fflush( DEPolis::fpfiring );
        }
      }
      return( fpfiring );
    }
  }

  method {
    name { Printfiring }
    access { public }
    type { void }
    arglist { "( char *st )" }
    code {
      if ( DEPolis::fpfiring ) {
        fprintf( DEPolis::fpfiring, st );
        fflush( DEPolis::fpfiring );
      }
    }
  }

  method {
    name { Closeflow }
    access { public }
    type { void }
    code {
      if ( DEPolis::fpoverflow ) {
        fclose( DEPolis::fpoverflow );
        DEPolis::fpoverflow = NULL;
      }
      if ( DEPolis::fpfiring ) {
        fclose( DEPolis::fpfiring );
        DEPolis::fpfiring = NULL;
      }
      if ( DEPolis::toCommand ) {
        fclose( DEPolis::toCommand );
        DEPolis::toCommand = NULL;
      }
      if ( DEPolis::fromCommand ) {
        fclose( DEPolis::toCommand );
        DEPolis::toCommand = NULL;
      }
      if ( DEPolis::isslog ) {
        fclose( DEPolis::isslog );
        DEPolis::isslog = NULL;
      }
      if ( DEPolis::fptraces ) {
        fclose( DEPolis::fptraces );
        DEPolis::fptraces = NULL;
      }
    }
  }

  method {
    name { getincCounter }
    access { public }
    type { int }
    code {
      DEPolis::counter++;
      return( DEPolis::counter - 1 );
    }
  }

  method {
    name { readPri }
    access { public }
    type { int }
    code {
      return( DEPolis::priority_next );
    }
  }

  method {
    name { readNext }
    access { public }
    type { double }
    code {
      return( DEPolis::next_curr );
    }
  }
  
  // Test if star can get the CPU. There are three cases:
  // 1. If now>=next then the processor is idle, so we must test if
  //    any other star of higher priority has requested the processor.
  // 2. Even if the processor is not idle, we can get it if the priority
  //    is higher than the process running and any other process which
  //    has made a request. This is clearly an interrupt.
  // 3. If the processor is not idle we must check that our priority
  //    is higher than the process running. There can be other processes
  //    of the same priority who made a similar request. This is an
  //    interrupt.
  method {
    name { testCPU }
    access { public }
    type { int }
    arglist { "( double now, int pri )" }
    code {
      return(( now >= DEPolis::next_curr && 
               pri >= DEPolis::priority_next ) ||
               pri > DEPolis::priority_next ||
               ( pri == DEPolis::priority_next && 
               pri > DEPolis::priority_pol ));
    }
  }

  method {
    name { canRun }
    access { public }
    type { int }
    arglist { "( int star_id )" }
    code {
      return( DEPolis::star_ack == star_id );
    }
  }

  method {
    name { saveContext }
    access { public }
    type { void }
    arglist { "( int *opri, double *onxc, int *osrn, int pri )" }
    code {
      *opri = DEPolis::priority_pol;
      *onxc = DEPolis::next_curr;
      *osrn = DEPolis::star_run;
      DEPolis::priority_pol = pri;
      DEPolis::priority_next = pri;
      DEPolis::star_run = DEPolis::star_ack;
    }
  }

  method {
    name { setNewDelay }
    access { public }
    type { void }
    arglist { "( double delay, int now, double *onxg )" }
    code {
      DEPolis::next_curr = now + delay;
      if ( now < DEPolis::next_glob ) {
        DEPolis::next_glob += delay;
      } else {
        DEPolis::next_glob = now + delay;
      }
      *onxg = DEPolis::next_glob;
    }
  }

  method {
    name { interrupt }
    access { public }
    type { double }
    arglist { "( double oldnextg )" }
    code {
        return( DEPolis::next_glob - oldnextg );
    }
  }

  method {
    name { restoreContext }
    access { public }
    type { void }
    arglist { "( double newnextc, int oldpri, int oldstarrun )" }
    code {
      DEPolis::next_curr = newnextc;
      DEPolis::priority_pol = oldpri;
      DEPolis::priority_next = oldpri;
      DEPolis::star_ack = -1;
      if ( oldpri >= 0 ) {
        DEPolis::star_run = oldstarrun;
      }
    }
  }

  method {
    name { scheduling }
    access { public }
    type { int }
    arglist { "( int star_id, int star_pri, char *sched )" }
    code {
      int retval;
      
      retval = 0;
      if ( !strcasecmp( sched, "_OldRoundRobin" )) {

        /* Old Round robin scheduler */
        int dist1, dist2;

        if ( DEPolis::star_ack == -1 ) {
          DEPolis::star_ack = star_id;
          retval = 1;
        } else {
          dist1 = DEPolis::star_ack - DEPolis::star_run;
          if ( DEPolis::star_ack <= DEPolis::star_run ) {
            dist1 = DEPolis::counter + dist1;
          }
          dist2 = star_id - DEPolis::star_run;
          if ( star_id <= DEPolis::star_run ) {
            dist2 = DEPolis::counter + dist2;
          }
          if ( dist2 < dist1 ) {
            DEPolis::star_ack = star_id;
            retval = 1;
          } else {
            retval = 0;
          }
        }

      } else if ( !strcasecmp( sched, "RoundRobin" )) {
        /* Round Robin Scheduler */
        int dist1, dist2;
        
        if ( DEPolis::priority_pol > -1 ) {
          retval = 0;
        } else if ( DEPolis::star_ack == -1 ) {
          DEPolis::star_ack = star_id;
          retval = 1;
        } else {
          dist1 = DEPolis::star_ack - DEPolis::star_run;
          if ( DEPolis::star_ack <= DEPolis::star_run ) {
            dist1 = DEPolis::counter + dist1;
          }
          dist2 = star_id - DEPolis::star_run;
          if ( star_id <= DEPolis::star_run ) {
            dist2 = DEPolis::counter + dist2;
          }
          if ( dist2 < dist1 ) {
            DEPolis::star_ack = star_id;
            retval = 1;
          } else {
            retval = 0;
          }
        }

      } else if ( !strcasecmp( sched, "Priority" )) {
        /* Priority Preemptive Scheduler */
        int dist1, dist2;
        
        if ( DEPolis::star_ack == -1 || star_pri > DEPolis::priority_next ) {
          DEPolis::star_ack = star_id;
          DEPolis::priority_next = star_pri;
          retval = 1;
        } else if ( star_pri < DEPolis::priority_next ) {
          retval = 0;
        } else {
          dist1 = DEPolis::star_ack - DEPolis::star_run;
          if ( DEPolis::star_ack <= DEPolis::star_run ) {
            dist1 = DEPolis::counter + dist1;
          }
          dist2 = star_id - DEPolis::star_run;
          if ( star_id <= DEPolis::star_run ) {
            dist2 = DEPolis::counter + dist2;
          }
          if ( dist2 < dist1 ) {
            DEPolis::star_ack = star_id;
            DEPolis::priority_next = star_pri;
            retval = 1;
          } else {
            retval = 0;
          }
        }

      } else if ( !strcasecmp( sched, "NonPreemptive" )) {
        /* Priority Non Preemptive Scheduler */
        int dist1, dist2;
        
        if ( DEPolis::priority_pol > -1 ) {
          retval = 0;
        } else if (DEPolis::star_ack == -1 || star_pri > DEPolis::priority_next) {
          DEPolis::star_ack = star_id;
          DEPolis::priority_next = star_pri;
          retval = 1;
        } else if ( star_pri < DEPolis::priority_next ) {
          retval = 0;
        } else {
          dist1 = DEPolis::star_ack - DEPolis::star_run;
          if ( DEPolis::star_ack <= DEPolis::star_run ) {
            dist1 = DEPolis::counter + dist1;
          }
          dist2 = star_id - DEPolis::star_run;
          if ( star_id <= DEPolis::star_run ) {
            dist2 = DEPolis::counter + dist2;
          }
          if ( dist2 < dist1 ) {
            DEPolis::star_ack = star_id;
            DEPolis::priority_next = star_pri;
            retval = 1;
          } else {
            retval = 0;
          }
        }
      } else {
        /* Not a valid scheduler */
        char stemp[ 1024 ];
        
        strcpy( stemp, sched );
        strcat( stemp, ": not a valid scheduler" );
        Error::abortRun( stemp );
      }
      return( retval );
    }
  }

//
//  Methods for the link Ptolemy/ISS
//

  method {
    name { resetIss }
    access { public }
    type { void }
    code {
      DEPolis::simCpuTime = clock();

      DEPolis::prev_cycles = 0;
      DEPolis::ptl_cycles = 0;
      DEPolis::prev_energy = 0.0;
      DEPolis::ptl_energy = 0.0;
      DEPolis::delay = 0;
      DEPolis::break_inserted = 0;
      fprintf( stderr, "DEPolis::iss_pid= %d\n", DEPolis::iss_pid);
//      if (DEPolis::iss_pid) {
//         kill(DEPolis::iss_pid,9);
//      }
      DEPolis::iss_pid = 0;
    }
  }

  method {
    name { Pipefork }
    access { public }
    type { int }
    arglist { "( char **argv, FILE **toCommand, FILE **fromCommand, \
                 int *pid, char *execname, char *scpu )" }
    code {

    int forkpid, waitpid;
    int topipe[2], frompipe[2];
    char buffer[1024];

#if (defined hpux) || (defined __osf__) || (defined _IBMR2) || defined(__SVR4)
    int status;
#else
    union wait status;
#endif

    /* create the PIPES...
     * fildes[0] for reading from command
     * fildes[1] for writing to command
     */
    (void) pipe(topipe);
    (void) pipe(frompipe);

    if ((forkpid = vfork()) == 0) {
       /* child here, connect the pipes */

       setbuf(stdout,NULL);

       (void) dup2(topipe[0], fileno(stdin));
       (void) dup2(frompipe[1], fileno(stdout));
       (void) dup2(frompipe[1], fileno(stderr));

       (void) close(topipe[0]);
       (void) close(topipe[1]);
       (void) close(frompipe[0]);
       (void) close(frompipe[1]);

       if ( !strcmp(scpu,"Chameleon") ) {
          (void) execlp( "pty", "-v", "chsim", "-z", "stat", "-t", "test",
                         (char *) 0);
       }
       else if ( !strcmp(scpu,"SPARC") ) {
//          (void) execlp( "ptydebug", "-v", "sparcsim", "-t", execname, (char *) 0);
          (void) execlp( "pty", "-v", "sparcsim", "-t", execname, (char *) 0);
       }
       (void) sprintf(buffer, "util_pipefork: can not exec %s", argv[0]);
       perror(buffer);
       (void) _exit(1);
    }

    if (pid) {
       *pid = forkpid;
    }

    waitpid = wait3(&status, WNOHANG, 0);

    /* parent here, use slimey vfork() semantics to get return status */
#ifndef sun
    if (waitpid == forkpid && WIFEXITED(status)) {
        return 0;
    }
#endif
    if ((*toCommand = fdopen(topipe[1], "w")) == NULL) {
        return 0;
    }
    if ((*fromCommand = fdopen(frompipe[0], "r")) == NULL) {
        return 0;
    }
    (void) close(topipe[0]);
    (void) close(frompipe[1]);
    return 1;
    }
  }

  method {
    name { initIss }
    access { public }
    type { void }
    arglist { "( char *brkname, char *execname, char *scpu, \
                 char *root_dir, int ncpu, int iss_debug )" }
    code {
       char stemp[80];
       char model_name[80];
       char brk_name[80];

       if ( !strcmp(scpu,"SPARC") ) {
          if (!DEPolis::break_inserted) {

             if ( iss_debug ) {
                sprintf( stemp, "%s/isslog.txt", root_dir ); 
                DEPolis::isslog = fopen( stemp, "w" );
                setbuf(DEPolis::isslog,NULL);
             }

             // Set "debug information" mode
             StringToIss("d\n", iss_debug);

             // Passing cpu value
             StringToIss("w\n", iss_debug);
             StringToIss("s\n", iss_debug);
             StringToIss("_cpu\n", iss_debug);
             sprintf( stemp, "%x\n", ncpu );
             StringToIss( stemp, iss_debug );
             DEPolis::break_inserted = 1;

             // Set breakpoint in ptl_sync
             fprintf(stderr, "DEPolis: breakname opened\n");
             DEPolis::breakname = fopen( brkname, "r" );
             DEPolis::num_break_inserted=0;
             StringToIss("s\n", iss_debug);
             sprintf( stemp, "%d\n", DEPolis::num_break_inserted );
             StringToIss( stemp, iss_debug );
             StringToIss("_ptl_sync\n", iss_debug);
             DEPolis::num_break_inserted++;
             fclose(DEPolis::breakname);
          }

          // Set breakpoints in the current module
          DEPolis::breakname = fopen( brkname, "r" );
          while (fgets(stemp, BUFSIZE, DEPolis::breakname) != NULL) {
             sscanf(stemp,"%s\t%s", model_name, brk_name);
             StringToIss("s\n", iss_debug);
             sprintf( stemp, "%d\n", DEPolis::num_break_inserted );
             StringToIss( stemp, iss_debug );
             sprintf( stemp, "_%s\n", brk_name );
             StringToIss( stemp, iss_debug );
             DEPolis::num_break_inserted++;
          }
          fclose(DEPolis::breakname);
       }
    }
  }

  method {
    name { StringToIss }
    access { public }
    type { void }
    arglist { "( char *command, int iss_debug )" }
    code {
       fprintf(DEPolis::toCommand,"%s", command); 
       fflush(DEPolis::toCommand);
       if ( iss_debug ) {
          fprintf(DEPolis::isslog,"%s", command); 
          fflush(DEPolis::isslog);
       }
    }
  }

  method {
    name { StringFromIss }
    access { public }
    type { void }
    arglist { "( char *fromcommand )" }
    code {
       fromcommand[0] = '\0';
       fgets(fromcommand,BUFSIZE,fromCommand);
    }
  }

  method {
    name { StringToIsslog }
    access { public }
    type { void }
    arglist { "( char *command )" }
    code {
       fprintf(DEPolis::isslog,"%s", command); 
       fflush(DEPolis::isslog);
    }
  }

  method {
    name { StartupIssCycles }
    access { public }
    type { void }
    arglist { "( int cycles)" }
    code {
       fprintf( stderr, "StartupIssCycles begin\n" );
       DEPolis::prev_cycles = cycles;
       fprintf( stderr, "StartupIssCycles end\n" );
    }
  }

  method {
    name { StoreIssCycles }
    access { public }
    type { void }
    arglist { "( int cycles, int cached_delay )" }
    code {
       if (!cached_delay) {
          DEPolis::delay = cycles - DEPolis::prev_cycles;
          DEPolis::prev_cycles = cycles;
       } else {
          DEPolis::delay = cycles;
       }
       DEPolis::ptl_cycles += DEPolis::delay;
    }
  }

  method {
    name { StartupIssEnergy }
    access { public }
    type { void }
    arglist { "( double energy )" }
    code {
       DEPolis::prev_energy = energy;
    }
  }

  method {
    name { StoreIssEnergy }
    access { public }
    type { void }
    arglist { "( double energy, int cached_energy )" }
    code {
       if (!cached_energy) {
          DEPolis::energy = energy - DEPolis::prev_energy;
       }
       else {
          DEPolis::energy = energy;
       }
       DEPolis::ptl_energy += DEPolis::energy;
       DEPolis::prev_energy = energy;
    }
  }

  method {
    name { GetIssCycles }
    access { public }
    type { int }
    code {
       return (DEPolis::delay);
    }
  }

  method {
    name { GetIssEnergy }
    access { public }
    type { double }
    code {
       return (DEPolis::energy);
    }
  }

  method {
    name { AddEnergy }
    access { public }
    type { void }
    arglist { "( double value )" }
    code {
       DEPolis::energy += value;
    }
  }

  method {
    name { runIss }
    access { public }
    type { void }
    arglist { "( char *execname, char *cpuname )" }
    code {
       char *argv[]={(char *) 0};

       if ( DEPolis::toCommand==NULL ) {
          if (!Pipefork(argv,&DEPolis::toCommand,&DEPolis::fromCommand,&DEPolis::iss_pid,execname,cpuname)) {
             fprintf(stderr, "Error in pipefork\n");
          }
          setbuf(toCommand,NULL);
          setbuf(stdout,NULL);
       }
    }
  }

//////////////
// SIS link
//////////////

  method {
    name { InitSis }
    access { public }
    type { void }
    code {
#ifdef LINK_SIS
        cur_vector = LS_NIL; 
        input_names = LS_NIL; 
        vector_number = 0; 
        init_polis();
        fprintf( stderr, "InitSis\n" );
#endif /* LINK_SIS */
    }
  }

  method {
    name { EndSis }
    access { public }
    type { void }
    code {
#ifdef LINK_SIS
        end_polis();
        fprintf( stderr, "EndSis\n" );
#endif /* LINK_SIS */
    }
  }

  method {
    name { ReadBlif }
    access { public }
    type { void }
    arglist { "( char *blifname, network_t **ntwk )" }
    code {
#ifdef LINK_SIS
        FILE *bliffile;

        bliffile = fopen( blifname, "r" );
        if (( bliffile = fopen( blifname, "r" )) != NULL) { 
           read_blif( bliffile, ntwk );
        } 
        else {
          char stemp[ 1024 ];
          strcpy( stemp, blifname );
          strcat( stemp, ": file not existing" );
          Error::abortRun( stemp );
        }
//        write_blif( stdout, *ntwk, 0, 0 );
        fclose( bliffile );
#endif /* LINK_SIS */
    }
  }

  method {
    name { PowerCommandLineInterface }
    access { public }
    type { void }
    arglist { "( network_t *ntwk, char *vecfile, float clkFreq )" }
    code {
#ifdef LINK_SIS
        char *power_argv[6];

        cur_vector = LS_NIL;
        input_names = LS_NIL;
        vector_number = 0;

        power_argv[0] = strdup( "power_estimate" );
        power_argv[1] = strdup( "-m" );
        power_argv[2] = strdup( "SAMPLING" );
        power_argv[3] = strdup( vecfile );
        power_argv[4] = strdup( "-c" );
        // power_argv[5] = strdup( "20.0" );
        fprintf( stderr, "PowerCommandLineInterface: clkFreq= %f\n", clkFreq ); 
        power_argv[5] = ALLOC( char, 20 );
        sprintf( power_argv[5], "%f", clkFreq ); 
        power_command_line_interface( ntwk, 6, power_argv );
        for ( int i=0; i<6; i++ ) {
           FREE( power_argv[i] );
        } 
#endif /* LINK_SIS */
    }
  }

  method {
    name { ConvBitVector }
    access { public }
    type { void }
    arglist { "( char *buf, int arg, int size )" }
    code {
       int i, temp;

       temp = arg;
       for (i=0; i<size; i++) {
          if ((temp%2) == 1) {
             strcat( buf, "1" );
          }
          else {
             strcat( buf, "0" );
          }
          if (temp>0) {
             temp = temp/2;
          }
          else {
             temp = (temp-1)/2; /* simulate ASR */
          }
       }
       strcat( buf, "\0" );
    }
  }

/////////////////////
// Cache simulation
/////////////////////

  method {
    name { InitCache }
    access { public }
    type { void }
    arglist { "( int task_id, c_sg_node_t **sgraph, \
                 int cachesize, int linesize, int associativity, \
                 int historylength, float threshold, int csimulstyle, \
                 int i_fetch_size, int miss_penalty, int tracegenflag, \
                 char *module_name, int *numnodes )" }
    code {
#ifdef CACHE_SIMUL
       char root_dir[ 256 ];
       char stemp[1024]; 

       if ( ! DEPolis::cache_initialized ) {
          init_cache( cachesize, linesize, associativity, historylength, 
                      threshold, csimulstyle, 2 /* I_FETCHSIZE */, miss_penalty );
          DEPolis::cache_initialized = 1;
       }
       strcpy( root_dir, getenv("PTPWD") );
       if ( ! *sgraph ) {
          /* first simulation run */
          sprintf( stemp, "%s/z_%s_0", root_dir, module_name );
          build_sgraph( task_id, 0, 0, stemp );
          *sgraph = get_task_sgraph( get_task( task_id ));
          *numnodes = get_task_num_c_sg_nodes( get_task( task_id ));
       }
       else {
          /* following simulation runs */
          *numnodes = get_task_num_c_sg_nodes( get_task( task_id ));
          clean_hit_rate( *sgraph, *numnodes-1 );
          update_addresses( *sgraph, *numnodes, 0 );
       }
       if ( tracegenflag ) {
          sprintf( stemp, "%s/traces.txt", root_dir );
          OpenTraces( stemp );
       }
#endif /* CACHE_SIMUL */
    }
  } 

  method {
    name { Compute_Abstract_Cache_State }
    access { public }
    type { void }
    arglist { "( int task_id, int path_id, \
                 c_sg_node_t *sgraph, header_path_t **header_path, \
                 path_cache_layout_t **curr_layout, int is_SW, \
                 int csimulstyle, int csimulflag, int tracegenflag )" }
    code {
#ifdef CACHE_SIMUL
        compute_abstract_cache_state( task_id, path_id, sgraph, header_path, curr_layout, is_SW, csimulstyle, csimulflag, tracegenflag, DEPolis::fptraces );
#endif /* CACHE_SIMUL */
    }
  }

  method {
    name { Cache_Plot_Hit_Rate }
    access { public }
    type { void }
    arglist { "( int task_id, c_sg_node_t *sgraph, \
                 header_path_t *header_path, int max_label, \
                 int is_SW, int csimulflag, int csimulstyle, \
                 XYPlot *xyplot, int plotgraph )" }
    code {
#ifdef CACHE_SIMUL

        // cache_plot_hit_rate( task_id, sgraph, header_path, max_label, is_SW, csimulflag, csimulstyle );

        int curr_node_id;
        c_sg_node_t *curr_node;

        if ( is_SW && csimulflag ) {
           if ( csimulstyle==FAST ) {
              curr_node_id = 0;
#ifdef DEBUG
              printf("CACHE_PLOT_HIT_RATE:\n");
              printf("\ttask_id= %d\n", task_id);
#endif
              while( curr_node_id <= max_label ) {
                 curr_node = get_node_sgraph( sgraph, curr_node_id );
#ifdef DEBUG
                 printf("\thit_rate(%d)= %.2f\tpath_vect(%d)= %d\n",
                        curr_node_id, get_hit_rate( curr_node ), curr_node_id,
                        get_path_vect_pos(header_path, curr_node_id));
#endif
#ifdef PTOLEMY
                 if ( plotgraph ) {
                    xyplot->addPoint( curr_node_id, get_hit_rate(curr_node), 1 );
                 }
#endif
                 curr_node_id++;
              }
           }
        }
#endif /* CACHE_SIMUL */
    }
  }

  method {
    name { Alloc_Header_Path }
    access { public }
    type { int }
    arglist { "( int numnodes )" }
    code {
#ifdef CACHE_SIMUL
        return( (int) alloc_header_path( numnodes ));
#endif /* CACHE_SIMUL */
    }
  }

  method {
    name { Add_Node_Path }
    access { public }
    type { void }
    arglist { "( header_path_t *header_path, int node, \
                 int is_SW, int csimulflag )" }
    code {
#ifdef CACHE_SIMUL
        add_node_in_path( header_path, node, is_SW, csimulflag );
#endif /* CACHE_SIMUL */
    }
  }

  method {
    name { OpenTraces }
    access { public }
    type { "FILE *" }
    arglist { "( char *name )" }
    code {
#ifdef CACHE_SIMUL
      if ( DEPolis::fptraces == NULL ) {
        DEPolis::fptraces = fopen( name, "w" );
        if ( DEPolis::fptraces ) {
          fprintf( DEPolis::fptraces, "Traces report\n\n" );
          fflush( DEPolis::fptraces );
        }
      }
      return( DEPolis::fptraces );
#endif /* CACHE_SIMUL */
    }
  }

  method {
    name { UpdateDelay_with_CacheInfo }
    access { public }
    type { double }
    code {
#ifdef CACHE_SIMUL
        double result;

        result = (double) ((n_instr_fetch_path - (int) n_instr_hit_path) * MISS_PENALTY);
        fprintf( stderr, "UpdateDelay_with_CacheInfo: _delay_cache= %f\n", result );
        return result;
#endif /* CACHE_SIMUL */
    }
  }

/////////////////
// Fuzzy delays
/////////////////

  method {
    name { FuzzyDelay }
    access { public }
    type { double }
    arglist { "( char *fuzzy_expr )" }
    code {
#ifdef USE_FUZZY
        return( (double) fuzzy_delay( fuzzy_expr )); 
#endif /* USE_FUZZY */
    }
  }

  method {
    name { FuzzyEnergy }
    access { public }
    type { double }
    arglist { "( char *fuzzy_expr )" }
    code {
#ifdef USE_FUZZY
        return( (double) fuzzy_energy( fuzzy_expr )); 
#endif /* USE_FUZZY */
    }
  }

  method {
    name { ReadCostParam }
    access { public }
    type { int }
    arglist { "( char *cpu_name )" }
    code {
#ifdef USE_FUZZY
        FILE *fp;
        char fname[ 1024 ];

        strcpy( fname, getenv("POLIS") );
        strcat( fname, "/polis_lib/estm/" );
        strcat( fname, cpu_name );
        strcat( fname, ".param" );
        if (( fp = fopen( fname, "r" )) != NULL) { 
           estm_free_estm_fncp(); 
           estm_read_param_file( fp );
           fclose( fp );
           return 0;
        }
        return -1;
#endif /* USE_FUZZY */
    }
  }

}
