/*************************************************************************

  Copyright   [Copyright (c) 1994-1996 The Regents of the Univ. of California.
  All rights reserved.

  Permission is hereby granted, without written agreement and without license
  or royalty fees, to use, copy, modify, and distribute this software and its
  documentation for any purpose, provided that the above copyright notice and
  the following two paragraphs appear in all copies of this software.

  IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
  DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
  OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
  CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN
  "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE
  MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.]

******************************************************************************/
/* Uses oc to implement a fixed or programmable timer.
 * If -DOC_SELFSTART is present in the Makefile, it restarts itself after
 * expiring. Otherwise, it is restarted by OC_START.
 * Programmability is defined by -DOC_PROG in the Makefile, when 
 * pre-processing the simulation esterel source.
 * In this file, this is also confirmed by the presence of constant OCPERIOD in
 * os.c. If it is defined, it is used as period and the timer
 * is not programmable.
 * Otherwise, it is programmable. It elapses when the counter reaches the
 * value of OC_START if -DOC_ABS is defined in the Makefile (absolute
 * counting), or if when the counter is incremented by the
 * value of OC_START if -DOC_ABS is not defined in the Makefile (relative
 * counting).
 * It assumes that OC_START and OC_END are top-level events in the system 
 * netlist (i.e., they keep their names...)
 */
$if (v_N == 1)
$define local_init
$define intr_enable _r_h11tmsk1.bit.b7
$define out_affect1 _r_h11oc1m.byte
$define out_affect2 _r_h11oc1m.byte
$define oc_reg _r_h11toc1
$define oc_flag _r_h11tflg1.bit.b7
$define OC_VEC OC1_VEC
$define oc_started oc1_started
$else
$if (v_N == 2)
$define local_init
$define intr_enable _r_h11tmsk1.bit.b6
$define out_affect1 _r_h11tctl1.bit.b7
$define out_affect2 _r_h11tctl1.bit.b6
$define oc_reg _r_h11toc2
$define oc_flag _r_h11tflg1.bit.b6
$define OC_VEC OC2_VEC
$define oc_started oc2_started
$else
$if (v_N == 3)
$define local_init
$define intr_enable _r_h11tmsk1.bit.b5
$define out_affect1 _r_h11tctl1.bit.b5
$define out_affect2 _r_h11tctl1.bit.b4
$define oc_reg _r_h11toc3
$define oc_flag _r_h11tflg1.bit.b5
$define OC_VEC OC3_VEC
$define oc_started oc3_started
$else
$if (v_N == 4)
$define local_init
$define intr_enable _r_h11tmsk1.bit.b4
$define out_affect1 _r_h11tctl1.bit.b3
$define out_affect2 _r_h11tctl1.bit.b2
$define oc_reg _r_h11toc4
$define oc_flag _r_h11tflg1.bit.b4
$define OC_VEC OC4_VEC
$define oc_started oc4_started
$else
$if (v_N == 5)
$define local_init _r_h11pactl.bit.b2 = 0;
$define intr_enable _r_h11tmsk1.bit.b3
$define out_affect1 _r_h11tctl1.bit.b1
$define out_affect2 _r_h11tctl1.bit.b0
$define oc_flag _r_h11tflg1.bit.b3
$define OC_VEC OC5_VEC
$define oc_started oc5_started
$else
invalid_oc_number_selected cause_syntax_error;
$endif
$endif
$endif
$endif
$endif

$if (v_SIMSCALE != 0)
    simscale_must_be_0 otherwise_syntax_error;
$endif

/****************************************************************************
 * State variable section
 */

#ifndef OC_SELFSTART
static char oc_started;
#endif

/****************************************************************************
 * Initialization section
 */
void init_oc@ ()
{
#ifdef OC_SELFSTART
$ifdef e_o_OC_END_POLLED
	/* intr disable */
	intr_enable = 0;
$else
	extern __IT_type IT_handle_e_o_OC_END ();
$ifdef EMULATOR
	extern void _VECTOR(void (* func)(void), int _VECTOR);
	_VECTOR (IT_handle_e_o_OC_END, OC_VEC);
$endif

	oc_reg = v_OCPERIOD + _r_h11tcnt;
$endif /*e_o_OC_END_POLLED*/
#else
$ifndef e_o_OC_END_POLLED
	extern __IT_type IT_handle_e_o_OC_END ();
$ifdef EMULATOR
	_VECTOR (IT_handle_e_o_OC_END, OC_VEC);
$endif
$endif /*e_o_OC_END_POLLED*/

	/* initialize the state */
	oc_started = 0;

	/* intr disable */
	intr_enable = 0;
#endif /*OC_SELFSTART*/

	/* don't affect output pins */
	out_affect1 = 0;  
	/* the same is done twice for oc1.... */
	out_affect2 = 0;  

	/* set oc5/ic4 to oc5 */
	local_init

$ifndef e_o_OC_END_POLLED
	/* intr enable */
	intr_enable = 1; 
$endif /*e_o_OC_END_POLLED*/
}

/****************************************************************************
 * Event detection section
 */

#ifndef OC_SELFSTART
void deliver_e_OC_START_to_oc@ ()
{
	oc_started = 1;
#ifdef OC_PROG
#ifdef OC_ABS
	oc_reg = v_OC_START;
#else
	oc_reg = v_OC_START + _r_h11tcnt;
#endif /*OC_ABS*/
#else
	oc_reg = v_OCPERIOD + _r_h11tcnt;
#endif /*OC_PROG*/

$ifndef e_o_OC_END_POLLED
	/* intr enable */
	intr_enable = 1; 
$endif /*e_o_OC_END_POLLED*/
}
#endif /*OC_SELFSTART*/

void deliver_e_RESET_to_oc@() { init_oc@ (); }

/****************************************************************************
 * Dummy event emission section
 * (should be inlined once we find out how to rename them according to
 * parameter passing)
 *void deliver_e_TCLK_to_oc@() { }
 */

/****************************************************************************
 * Event emission section
 */

$ifdef e_o_OC_END_POLLED
int got_inp_event_e_o_OC_END ()
$else
__IT_type IT_handle_e_o_OC_END ()
$endif /*e_o_OC_END_POLLED*/
{
$ifndef e_o_OC_END_POLLED
    /* intr disable */
    intr_enable = 0;
$endif /*e_o_OC_END_POLLED*/
#ifdef OC_SELFSTART
	oc_reg = v_OCPERIOD + _r_h11tcnt;
#endif /*OC_SELFSTART*/

$ifdef e_o_OC_END_POLLED
#ifdef OC_SELFSTART
	if (oc_flag) {
#else
	if (oc_flag && oc_started) {
#endif /*OC_SELFSTART*/
$endif /*e_o_OC_END_POLLED*/
	/* funny way to clear a flag... */
		oc_flag = 1;
	/* the test is done just to be safe */
#ifdef OC_SELFSTART
$ifndef e_o_OC_END_POLLED
		emit_e_o_OC_END ();
$endif /*e_o_OC_END_POLLED*/
#else
$ifndef e_o_OC_END_POLLED
		if (oc_started) emit_e_o_OC_END ();
$endif /*e_o_OC_END_POLLED*/
		oc_started = 0;
#endif /*OC_SELFSTART*/
$ifdef e_o_OC_END_POLLED
		return 1;
	}
	return 0;
$else
    /* intr enable */
    intr_enable = 1;
$endif /*e_o_OC_END_POLLED*/
}

$undef local_init
$undef intr_enable
$undef out_affect1
$undef out_affect2
$undef oc_reg
$undef oc_flag
$undef OC_VEC
$undef oc_started 
$undef init_oc@ 
/****************************************************************************
 * end
 */
