/*************************************************************************

  Copyright   [Copyright (c) 1994-1996 The Regents of the Univ. of California.
  All rights reserved.

  Permission is hereby granted, without written agreement and without license
  or royalty fees, to use, copy, modify, and distribute this software and its
  documentation for any purpose, provided that the above copyright notice and
  the following two paragraphs appear in all copies of this software.

  IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
  DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
  OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
  CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN
  "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE
  MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.]

******************************************************************************/
/* Uses pwm to implement a programmable pulse-width modulated output
 * If -DPWM_ENABLED is defined in the Makefile when
 * pre-processing the simulation esterel source, then two inputs (PWM_ENABLE
 * and PWM_DISABLE) start and stop it respectively.
 * The duty cycle is defined by the value of event
 * PWM_DC (10 bit integer).
 * Constant PWM_POLARITY defines the value of the output at the beginning
 * of the cycle.
 * It assumes that its inputs and outputs are top-level events in the system 
 * netlist (i.e., they keep their names...)
 */

$if  (v_N == 0)
$define pwm_enable _r_h11pwen.bit.b0
$define pwm_pol _r_h11pwpol.bit.b0
$define pwm_reg _r_h11pwdty0
$else
$if  (v_N == 1)
$define pwm_enable _r_h11pwen.bit.b1
$define pwm_pol _r_h11pwpol.bit.b1
$define pwm_reg _r_h11pwdty1
$else
$if  (v_N == 2)
$define pwm_enable _r_h11pwen.bit.b2
$define pwm_pol _r_h11pwpol.bit.b2
$define pwm_reg _r_h11pwdty2
$else
$if  (v_N == 3)
$define pwm_enable _r_h11pwen.bit.b3
$define pwm_pol _r_h11pwpol.bit.b3
$define pwm_reg _r_h11pwdty3
$else
$if  (v_N == 4)
$define pwm_enable _r_h11pwen.bit.b4
$define pwm_pol _r_h11pwpol.bit.b4
$define pwm_reg _r_h11pwdty4
$else
$if  (v_N == 5)
$define pwm_enable _r_h11pwen.bit.b5
$define pwm_pol _r_h11pwpol.bit.b5
$define pwm_reg _r_h11pwdty5
$else
$if  (v_N == 6)
$define pwm_enable _r_h11pwen.bit.b6
$define pwm_pol _r_h11pwpol.bit.b6
$define pwm_reg _r_h11pwdty6
$else
$if  (v_N == 7)
$define pwm_enable _r_h11pwen.bit.b7
$define pwm_pol _r_h11pwpol.bit.b7
$define pwm_reg _r_h11pwdty7
$else
invalid_pwm_number_selected cause_syntax_error;
$endif
$endif
$endif
$endif
$endif
$endif
$endif
$endif

$if (v_SIMSCALE != 0)
	simscale_must_be_0 otherwise_syntax_error;
$endif

/****************************************************************************
 * State variable section
 */

/****************************************************************************
 * Initialization section
 */
void init_pwm@ ()
{
#ifdef PWM_ENABLED
	pwm_enable = 0;
#else /*PWM_ENABLED*/
	pwm_enable = 1;
#endif /*PWM_ENABLED*/
	pwm_pol = v_PWM_POLARITY;
}

/****************************************************************************
 * Event detection section
 */

void deliver_e_PWM_DC_to_pwm@ ()
{
	pwm_reg = v_PWM_DC;
}

#ifdef PWM_ENABLED
void deliver_e_PWM_ENABLE_to_pwm@ ()
{
	pwm_enable = 1;
}

void deliver_e_PWM_DISABLE_to_pwm@ ()
{
	pwm_enable = 0;
}
#endif /*PWM_ENABLED*/

/****************************************************************************
 * Dummy event section
 * (should be inlined once we find out how to rename them according to
 * parameter passing)
 *void deliver_e_PWMCLK_to_pwm@() { }
 *int got_inp_event_e_PWM_OUT() { return 1; }
 */

$undef pwm_enable
$undef pwm_pol
$undef pwm_reg
/****************************************************************************
 * end
 */

